/***********************************************
classes.h

Purpose

Define classes which will not ordinarily be redefined
in an application.

Overview


Definitions of Variables

CP_Filename_Namelist = Namelist Data File Name.
CP_Filename_FITS_Image1 = First Input Image Data File Name.
CP_Filename_FITS_Image2 = Second Input Image Data File Name.
CP_Filename_FITS_Out = Output Image Data File Name.
CP_Filename_Log = Logfile Name.

I_status = If error occurs, function and error
  code is written here. Nominally is zero. Most functions
  check this variable and will skip processing if it is nonzero.
  Computation loop sets the flag to zero and proceeds
  regardless, so that bad pixels will not prevent the processing of good ones.
I_NaNFlag = Flag to indicate that a non-number was produced
  in the computation loop, since that process does not
  convey error information into I_status.
I_NaNCount = A counter that keeps track of the number of NaNs produced.
I_Debug = A flag to control program for debugging purposes.
I_Verbose = A flag to turn on print statements to show program progress.
I_SuperVerbose = A flag to print results of computations.

I_Length_X = Length of first dimension of data.
I_Length_Y = Length of second dimension of data.
I_Num_Frames  = Number of Frames of data.
I_Operation = Which operation to perform.

DP_Output_Array = Array for storing results to be output.

***********************************************/

typedef struct {
/* Flag and value (as string) */
char	CP_Flag[MAX_FILENAME_LENGTH];
char	CP_FlagValue[MAX_FILENAME_LENGTH];
} M_Flag;

class M_CmdParse	{
public:
/* Status used for parse_args only */
int	I_status;
/* The input */
int 	argc;
char 	**argv;
/* Number of flags found */
int	I_Numflags;
/* Pointer to array of flags and values */
M_Flag	*flags;           

public:
M_CmdParse(int argc_in, char **argv_in);
~M_CmdParse();
int GetArgc();
int GetStatus();
char** GetArgv();
void SetArgc(int argc_in);
void SetArgv(char **argv_in);
/* Write the arguments and their values as an array */		     
int parse_args();
int get_flag_value(const char* CP_Flag, unsigned short& US_val);
int get_flag_value(const char* CP_Flag, short& S_val);
int get_flag_value(const char* CP_Flag, unsigned int& UI_val);
int get_flag_value(const char* CP_Flag, int& I_val);
int get_flag_value(const char* CP_Flag, long& L_val);
int get_flag_value(const char* CP_Flag, float& F_val);
int get_flag_value(const char* CP_Flag, double& D_val);
int get_flag_value(const char* CP_Flag, char* CP_val);
int get_flag_value(char* CP_Flag, char* CP_val);
};

class  M_Status {
public:
int	I_status;
int	I_NaNFlag;
int	I_NaNCount;
int	I_Debug;
int	I_Verbose;
int	I_SuperVerbose;
M_Status(M_CmdParse *GP_Cmd);
~M_Status();
};

typedef struct {
/* Flag and value (as string) */
char	CP_Name[MAX_FILENAME_LENGTH];
char	CP_Value[MAX_FILENAME_LENGTH];
} M_NameVal;


class  M_Namelist {

/* Namelist filename and sectionname */
public:
char	CP_Filename[MAX_FILENAME_LENGTH];
char	CP_SectionName[MAX_FILENAME_LENGTH];


/* Number of namelist entries found */
int	I_Numnames;

/* Array of names and values */
M_NameVal	*names;


public:
int get_values(M_CmdParse *GP_Cmd, M_Status *GP_Stat);
int parse_namelist(M_Status *GP_Stat);
int get_namelist_value(const char* CP_Name, unsigned short& US_val);
int get_namelist_value(const char* CP_Name, short& S_val);
int get_namelist_value(const char* CP_Name, unsigned int& UI_val);
int get_namelist_value(const char* CP_Name, int& I_val);
int get_namelist_value(const char* CP_Name, long& L_val);
int get_namelist_value(const char* CP_Name, float& F_val);
int get_namelist_value(const char* CP_Name, double& D_val);
int get_namelist_value(const char* CP_Name, char* CP_val);

M_Namelist(const char	*NLName, const char *NLSection);
~M_Namelist();	

};


class M_Logfile{
public:
/* Module Name */
char	CP_ModName[MAX_FILENAME_LENGTH];

/* Software Version */
char	CP_Version[MAX_FILENAME_LENGTH];

/* Log filename */
char	CP_Filename[MAX_FILENAME_LENGTH];
public:
int	I_Overwrite;
int	I_Time;
int	I_applayer_message;
int	I_applayer_root;
int	I_cfitsio_message;

char	CP_Message[I_MESSAGE_LENGTH];
char	CP_Function[I_FCNNAME_LENGTH];
int log_writer(M_CmdParse *GP_Cmd, M_Namelist *GP_Name, M_Status *GP_Stat);
int	decode_message(int I_message, char *CP_Message);
int	decode_fcnname(int I_root, char *CP_Function);
M_Logfile(const char*, const char*, const char*, M_CmdParse*, M_Namelist*);
~M_Logfile();
};


typedef struct	{

int	type;
char	keyword[FLEN_KEYWORD];
char	value[FLEN_VALUE];
char	comment[FLEN_COMMENT];

} FITS_card;


class M_FITSinfo {
public:
/* FITS filename and pointer*/
char	CP_Filename[MAX_FILENAME_LENGTH];
fitsfile 	*ffp_FITS;

/* Name of FITS file to copy header from and pointer*/
char	CP_HdrFile[MAX_FILENAME_LENGTH];
fitsfile 	*ffp_FITS_Hdr;

/* Keyword list */
const char		**CPP_Keywords;
int		I_NumKWs;
FITS_card	*KWlines;

/* Dimensions */
int	I_Bitpix;
int	I_Naxis;
int	I_Length_X;
int	I_Length_Y;
int	I_Num_Frames;

/* Pointer to data */
double*	data;

public:
int read_data(M_Status *GP_Stat);
int read_keywords(M_Status *GP_Stat);
int get_keyword_value(const char* CP_Keyword, unsigned short& US_val);
int get_keyword_value(const char* CP_Keyword, short& S_val);
int get_keyword_value(const char* CP_Keyword, unsigned int& UI_val);
int get_keyword_value(const char* CP_Keyword, int& I_val);
int get_keyword_value(const char* CP_Keyword, long& L_val);
int get_keyword_value(const char* CP_Keyword, float& F_val);
int get_keyword_value(const char* CP_Keyword, double& D_val);
int get_keyword_value(const char* CP_Keyword, char* CP_val);
int load_keyword(const char* keyword,short value,const char* comment,M_Status *GP_Stat);
int load_keyword(const char* keyword,int value,const char* comment,M_Status *GP_Stat);
int load_keyword(const char* keyword,long value,const char* comment,M_Status *GP_Stat);
int load_keyword(const char* keyword,float value,const char* comment,M_Status *GP_Stat);
int load_keyword(const char* keyword,double value,const char* comment,M_Status *GP_Stat);
int load_keyword(const char* keyword,const char* value,const char* comment,M_Status *GP_Stat);
int load_keyword(const char* comment,M_Status *GP_Stat);
int load_keyword(char* comment,M_Status *GP_Stat);
int output_keywords(M_Status *GP_Stat);
int write_data(M_Status *GP_Stat);
int write_output(int,M_CmdParse*,M_Logfile*,M_Namelist*,M_FITSinfo*,void*,void*,void*,M_FITSinfo*,M_Status*);
int load_keyword(const char* keyword,const char* comment,M_Status *GP_Stat);
/*int load_output_keywords(int I_index, M_Status *GP_Stat);*/
int load_std_keywords(const char*, const char*, M_Status*);
int load_extra_keywords(int 		I_index,
			M_Logfile	*GP_Log,
			M_Namelist	*GP_Name,
			M_FITSinfo	*GP_Infiles,
			void		*GP_Const,
			void		*GP_Init,
			void		*GP_Comp,
			M_FITSinfo	*GP_Outfiles,
			M_Status	*GP_Stat);
int set_output_data(	int 		I_index,
			M_FITSinfo	*GP_Infiles,
			void		*GP_ConstIn,
			void		*GP_InitIn,
			void		*GP_CompIn,
			M_FITSinfo	*GP_Outfiles,
			M_Status	*GP_Stat);
M_FITSinfo(const char	*Infilename,const char **CPP_KeywordsIn);
M_FITSinfo(const char	*Infilename);
M_FITSinfo();
~M_FITSinfo();	

};



class M_ReadWrite {
public:
int		I_NumInfiles;
int		I_NumOutfiles;

public:
int read_inputs(M_CmdParse*,
		M_Namelist*,
		M_FITSinfo*,
		M_Status*);
int write_outputs(M_CmdParse*,
		  M_Logfile*,
		  M_Namelist*,
		  M_FITSinfo*,
		  void*,
		  void*,
		  void*,
		  M_FITSinfo*,
		  M_Status*);
M_ReadWrite(const int, const int);
~M_ReadWrite();

};




