/*********************************************** 
fitsinfo.c

Purpose

Read the image data from a FITS file and save into an array.

Overview

Open the FITS file, allocate memory, read the data into an array, 

Definitions of Variables

External:
See generic.h and generic_exec.cpp

 
Internal:
I_fits_return_status = Resturn value of fits function.
I_ndims_found = Number of data dimensions found.
I_anynull = Whether nulls were found in the data.
LP_naxes = Array to store lengths of axes of data.
D_nullval = Set to zero, indicating not to replace nulls.

***********************************************/

#include <string.h>
#include <stdlib.h>

#include "dlmodule_includes.h"

extern const char*	CPP_NamelistNamesOut[MAX_FILENAME_LENGTH];

M_FITSinfo::M_FITSinfo(const char *Infilename, const char **CPP_KeywordsIn)
{
  /* Constructor */
  sprintf(CP_Filename, "%s", Infilename);
  ffp_FITS = NULL;
  sprintf(CP_HdrFile, "%s", "NONE");
  ffp_FITS_Hdr = NULL;
  CPP_Keywords = CPP_KeywordsIn;
  I_NumKWs=0;
  if(CPP_Keywords)
    while(CPP_Keywords[I_NumKWs]) I_NumKWs++;
  KWlines = NULL;
  I_Bitpix = 0; 
  I_Length_X = 0;
  I_Length_Y = 0;
  I_Num_Frames = 0;
  data = NULL;

}

M_FITSinfo::M_FITSinfo(const char	*Infilename)
{
  /* Constructor */
  sprintf(CP_Filename, "%s", Infilename);
  ffp_FITS = NULL;
  sprintf(CP_HdrFile, "%s", "NONE");
  ffp_FITS_Hdr = NULL;
  I_NumKWs=0;
  KWlines = NULL;
  I_Bitpix = 0; 
  I_Length_X = 0;
  I_Length_Y = 0;
  I_Num_Frames = 0;
  data = NULL;
}

M_FITSinfo::M_FITSinfo()
{
  /* Constructor */
  sprintf(CP_Filename, "%s", "NONE");
  ffp_FITS = NULL;
  sprintf(CP_HdrFile, "%s", "NONE");
  ffp_FITS_Hdr = NULL;
  I_NumKWs=0;
  KWlines = NULL;
  I_Bitpix = 0; 
  I_Length_X = 0;
  I_Length_Y = 0;
  I_Num_Frames = 0;
  data = NULL;
}


M_FITSinfo::~M_FITSinfo()
{
  if(data!=NULL) free(data);
  if(KWlines!=NULL) free(KWlines);
}

int M_FITSinfo::read_data(M_Status *GP_Stat)
{
  int	I_fits_return_status=0, I_anynull, I_nullval=0;
  long	LP_naxes[3];
  double	D_nullval=0;


  if(GP_Stat->I_status) return(GP_Stat->I_status);

  /* Get the data and keyword values from the input image file */

  fits_open_file(	&(ffp_FITS),
			CP_Filename,
			READONLY,
			&I_fits_return_status);

		
  if(I_fits_return_status)
    {GP_Stat->I_status = READ_IMAGE|FITS_MSG_MASK|I_fits_return_status;
    return(GP_Stat->I_status);}
 
  /* Read the keywords */		
  read_keywords(GP_Stat);
  if(GP_Stat->I_status) return(GP_Stat->I_status);

  LP_naxes[0] = I_Length_X;
  LP_naxes[1] = I_Length_Y;
  LP_naxes[2] = I_Num_Frames;

  if(1 || I_Bitpix < 0)	{
    /* Allocate memory for the data */
    data = (double *) calloc(LP_naxes[0]*LP_naxes[1]*LP_naxes[2], sizeof(double));
    /* Use I_fits_return_status to store status, 
       so if there was a problem with calloc
       subsequent fits functions can detect it */
    if(data==NULL) I_fits_return_status = MALLOC_FAILED;

    /* Read the data */
    fits_read_img(	ffp_FITS,
			TDOUBLE,
			1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
			&D_nullval,
			(double *) data,
			&I_anynull,
			&I_fits_return_status);
  }
  else	{
    /* Allocate memory for the data */
    data = (double *) calloc(LP_naxes[0]*LP_naxes[1]*LP_naxes[2], sizeof(int));
    /* Use I_fits_return_status to store status, 
       so if there was a problem with calloc
       subsequent fits functions can detect it */
    if(data==NULL) I_fits_return_status = MALLOC_FAILED;

    /* Read the data */
    fits_read_img(	ffp_FITS,
			TINT,
			1, LP_naxes[0]*LP_naxes[1]*LP_naxes[2],
			&I_nullval,
			(int *) data,
			&I_anynull,
			&I_fits_return_status);
  }


  fits_close_file(ffp_FITS,
		  &I_fits_return_status);

  /* Return the fits status as applayer status, with mask to say so */		
  if(I_fits_return_status)
    GP_Stat->I_status = READ_IMAGE|FITS_MSG_MASK|I_fits_return_status;


  if(GP_Stat->I_Verbose)
    if(GP_Stat->I_status==0)	{
      printf("read_data: Data read from file:\n%s\n", CP_Filename);
      printf("read_data: I_Bitpix = %d\n", I_Bitpix);
      printf("read_data: I_Length_X = %d\n", I_Length_X);
      printf("read_data: I_Length_Y = %d\n", I_Length_Y);
      printf("read_data: I_Num_Frames = %d\n", I_Num_Frames);
    }
    else
      printf("read_data: Could not read from file:\n%s\n", CP_Filename);
	
  return(GP_Stat->I_status);
}



/*********************************************** 
read_keywords()

Purpose

Read keywords from a FITS file and save into an array.

Overview

Read the keywords and their values using cfitsio
functions. Dimensional keywords have their own variables.

Definitions of Variables

External:
See generic_defs.h and generic_exec.c

 
Internal:
I_fits_return_status = Return value of fits function.
I_ndims_found = Number of data dimensions found.
I_anynull = Whether nulls were found in the data.
LP_naxes = Array to store lengths of axes of data.
D_nullval = Set to zero, indicating not to replace nulls.

***********************************************/


int M_FITSinfo::read_keywords(M_Status *GP_Stat)
{
  int	i;
  int	I_fits_return_status=0, I_ndims_found;
  long	LP_naxes[3];

  char	CP_Keyword[FLEN_KEYWORD], CP_Value[FLEN_VALUE], CP_Comment[FLEN_COMMENT];

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  strncpy(CP_Keyword, "BITPIX", FLEN_KEYWORD);
  fits_read_key(	ffp_FITS,
			TINT,
			CP_Keyword,
			&I_Bitpix, 
			CP_Comment, 
			&I_fits_return_status);

  strncpy(CP_Keyword, "NAXIS", FLEN_KEYWORD);
  fits_read_key(	ffp_FITS,
			TINT,
			CP_Keyword,
			&I_Naxis, 
			CP_Comment, 
			&I_fits_return_status);
		
  if(GP_Stat->I_Debug)
    printf("read_keywords: Keyword NAXIS = %d\n", I_Naxis);

  /* Set default axis lengths */
  LP_naxes[1] = 1;
  LP_naxes[2] = 1;

  /* Read the keywords that tell the dimensions of the data */
  strncpy(CP_Keyword, "NAXIS", FLEN_KEYWORD);
  fits_read_keys_lng(	ffp_FITS,
			CP_Keyword,
			1, I_Naxis,
			LP_naxes,
			&I_ndims_found,
			&I_fits_return_status);
	
  I_Length_X = LP_naxes[0];
  I_Length_Y = LP_naxes[1];
  I_Num_Frames = LP_naxes[2];


  /* Read other requested keywords */
  /* Allocate the memory to store the strings */
  KWlines = (FITS_card*) malloc(I_NumKWs*sizeof(FITS_card));
  if(KWlines == NULL)
    {GP_Stat->I_status = READ_IMAGE|MALLOC_FAILED; return(GP_Stat->I_status);}

  /* Read the keyword and store in the keyword line structure */
  for(i=0;i<I_NumKWs;i++)	{
    KWlines[i].type = TSTRING;
    strncpy(CP_Keyword, CPP_Keywords[i], FLEN_KEYWORD);
    strncpy(KWlines[i].keyword, CP_Keyword, FLEN_KEYWORD);
    fits_read_key(	ffp_FITS,
			TSTRING,
			CP_Keyword,
			CP_Value, 
			CP_Comment, 
			&I_fits_return_status);		
    strncpy(KWlines[i].value, CP_Value, FLEN_VALUE);
    strncpy(KWlines[i].comment, CP_Comment, FLEN_COMMENT);

    if(GP_Stat->I_Debug)
      printf("read_keywords: Keyword = %s\t%s\n", CP_Keyword, CP_Value);

  }

  /* Return the fits status, with mask to say so */		
  if(I_fits_return_status)
    GP_Stat->I_status = READ_IMAGE|FITS_MSG_MASK|I_fits_return_status;

  return(GP_Stat->I_status);
}


/*********************************************** 
get_keyword_value

Purpose
Look up the value of the named variable in the 
keyword table information.

Overview
Loop through the table of values produced by
parse_namelist, seeking the designatated name, and
overwrite the variable's value if the name is found.

Definitions of Variables

External:
See generic.h


Internal:


***********************************************/


int M_FITSinfo::get_keyword_value(const char* CP_Keyword, unsigned short& US_val)
{
  int		i;
  if(ffp_FITS == NULL) return(1);
  for(i=0;i<I_NumKWs;i++)
    if(strcmp(CP_Keyword, KWlines[i].keyword) == 0)	{
      sscanf(KWlines[i].value, "%hd", &US_val);
      printf("get_keyword_value: Keyword = %s\t New Value = %d\n", CP_Keyword, US_val);
    }
  return(0);
}


int M_FITSinfo::get_keyword_value(const char* CP_Keyword, short& S_val)
{
  int		i;
  if(ffp_FITS == NULL) return(1);
  for(i=0;i<I_NumKWs;i++)
    if(strcmp(CP_Keyword, KWlines[i].keyword) == 0)	{
      sscanf(KWlines[i].value, "%hd", &S_val);
      printf("get_keyword_value: Keyword = %s\t New Value = %d\n", CP_Keyword, S_val);
    }
  return(0);
}


int M_FITSinfo::get_keyword_value(const char* CP_Keyword, unsigned int& UI_val)
{
  int		i;
  if(ffp_FITS == NULL) return(1);
  for(i=0;i<I_NumKWs;i++)
    if(strcmp(CP_Keyword, KWlines[i].keyword) == 0)	{
      if(strcmp(KWlines[i].value, "") == 0)
	UI_val = 1 - UI_val;
      else
	sscanf(KWlines[i].value, "%d", &UI_val);
      printf("get_keyword_value: Keyword = %s\t New Value = %d\n", CP_Keyword, UI_val);
    }
  return(0);
}


int M_FITSinfo::get_keyword_value(const char* CP_Keyword, int& I_val)
{
  int		i;
  if(ffp_FITS == NULL) return(1);
  for(i=0;i<I_NumKWs;i++)
    if(strcmp(CP_Keyword, KWlines[i].keyword) == 0)	{
      if(strcmp(KWlines[i].value, "") == 0)
	I_val = 1 - I_val;
      else
	sscanf(KWlines[i].value, "%d", &I_val);
      printf("get_keyword_value: Keyword = %s\t New Value = %d\n", CP_Keyword, I_val);
    }
  return(0);
}


int M_FITSinfo::get_keyword_value(const char* CP_Keyword, long& L_val)
{
  int		i;
  if(ffp_FITS == NULL) return(1);
  for(i=0;i<I_NumKWs;i++)
    if(strcmp(CP_Keyword, KWlines[i].keyword) == 0)	{
      if(strcmp(KWlines[i].value, "") == 0)
	L_val = 1 - L_val;
      else
	sscanf(KWlines[i].value, "%d", &L_val);
      printf("get_keyword_value: Keyword = %s\t New Value = %d\n", CP_Keyword, L_val);
    }
  return(0);
}


int M_FITSinfo::get_keyword_value(const char* CP_Keyword, float& F_val)
{
  int		i;
  if(ffp_FITS == NULL) return(1);
  for(i=0;i<I_NumKWs;i++)
    if(strcmp(CP_Keyword, KWlines[i].keyword) == 0)	{
      sscanf(KWlines[i].value, "%f", &F_val);
      printf("get_keyword_value: Keyword = %s\t New Value = %f\n", CP_Keyword, F_val);
    }
  return(0);
}


int M_FITSinfo::get_keyword_value(const char* CP_Keyword, double& D_val)
{
  int		i;
  if(ffp_FITS == NULL) return(1);
  for(i=0;i<I_NumKWs;i++)
    if(strcmp(CP_Keyword, KWlines[i].keyword) == 0)	{
      sscanf(KWlines[i].value, "%lf", &D_val);
      printf("get_keyword_value: Keyword = %s\t New Value = %f\n", CP_Keyword, D_val);
    }
  return(0);
}


int M_FITSinfo::get_keyword_value(const char* CP_Keyword, char* CP_val)
{
  int		i;
  if(ffp_FITS == NULL) return(1);
  for(i=0;i<I_NumKWs;i++)
    if(strcmp(CP_Keyword, KWlines[i].keyword) == 0)	{
      sprintf(CP_val, "%s", KWlines[i].value);
      printf("get_keyword_value: Keyword = %s\t New Value = %s\n", CP_Keyword, CP_val);
    }
  return(0);
}



/*********************************************** 
load_keyword()

Purpose

Write a keyword value to the keyword array for later output.

Overview

Reallocate for sufficient size, add strings to the new last
element of the array.

Definitions of Variables

External:
See generic_defs.h and generic_exec.c

Internal:
I_fits_return_status = Return value of fits function.
CP_Keyname = Name of Key to be written.
CP_Comment = String value to be written in comment field of header.

***********************************************/

int M_FITSinfo::load_keyword(	const char*	keyword,
				short		value,
				const char*	comment,
				M_Status	*GP_Stat)
{
  int	Index;

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  Index = I_NumKWs;

  /* Increment the counter and allocate the memory to store the strings */
  I_NumKWs++;
  KWlines = (FITS_card*)
    realloc(KWlines, I_NumKWs*sizeof(FITS_card));
  if(KWlines == NULL)
    {GP_Stat->I_status = LD_KEYWORD|MALLOC_FAILED; return(GP_Stat->I_status);}

  /* Write the keyword value into the array */
  KWlines[Index].type = TSHORT;
  snprintf(KWlines[Index].keyword, sizeof(KWlines[Index].keyword), "%s", keyword);
  snprintf(KWlines[Index].value, sizeof(KWlines[Index].value), "%d", value);
  snprintf(KWlines[Index].comment, sizeof(KWlines[Index].comment), "%s", comment);
				
  return(GP_Stat->I_status);
}


int M_FITSinfo::load_keyword(	const char*	keyword,
				int		value,
				const char*	comment,
				M_Status	*GP_Stat)
{
  int	Index;

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  Index = I_NumKWs;

  /* Increment the counter and allocate the memory to store the strings */
  I_NumKWs++;
  KWlines = (FITS_card*)
    realloc(KWlines, I_NumKWs*sizeof(FITS_card));
  if(KWlines == NULL)
    {GP_Stat->I_status = LD_KEYWORD|MALLOC_FAILED; return(GP_Stat->I_status);}

  /* Write the keyword value into the array */
  KWlines[Index].type = TINT;
  snprintf(KWlines[Index].keyword, sizeof(KWlines[Index].keyword), "%s", keyword);
  snprintf(KWlines[Index].value, sizeof(KWlines[Index].value), "%d", value);
  snprintf(KWlines[Index].comment, sizeof(KWlines[Index].comment), "%s", comment);
				
  return(GP_Stat->I_status);
}


int M_FITSinfo::load_keyword(	const char*	keyword,
				long		value,
				const char*	comment,
				M_Status	*GP_Stat)
{
  int	Index;

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  Index = I_NumKWs;

  /* Increment the counter and allocate the memory to store the strings */
  I_NumKWs++;
  KWlines = (FITS_card*)
    realloc(KWlines, I_NumKWs*sizeof(FITS_card));
  if(KWlines == NULL)
    {GP_Stat->I_status = LD_KEYWORD|MALLOC_FAILED; return(GP_Stat->I_status);}

  /* Write the keyword value into the array */
  KWlines[Index].type = TLONG;
  snprintf(KWlines[Index].keyword, sizeof(KWlines[Index].keyword), "%s", keyword);
  snprintf(KWlines[Index].value, sizeof(KWlines[Index].value), "%d", value);
  snprintf(KWlines[Index].comment, sizeof(KWlines[Index].comment), "%s", comment);
				
  return(GP_Stat->I_status);
}


int M_FITSinfo::load_keyword(	const char*	keyword,
				float		value,
				const char*	comment,
				M_Status	*GP_Stat)
{
  int	Index;

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  Index = I_NumKWs;

  /* Increment the counter and allocate the memory to store the strings */
  I_NumKWs++;
  KWlines = (FITS_card*)
    realloc(KWlines, I_NumKWs*sizeof(FITS_card));
  if(KWlines == NULL)
    {GP_Stat->I_status = LD_KEYWORD|MALLOC_FAILED; return(GP_Stat->I_status);}

  /* Write the keyword value into the array */
  KWlines[Index].type = TFLOAT;
  snprintf(KWlines[Index].keyword, sizeof(KWlines[Index].keyword), "%s", keyword);
  snprintf(KWlines[Index].value, sizeof(KWlines[Index].value), "%f", value);
  snprintf(KWlines[Index].comment, sizeof(KWlines[Index].comment), "%s", comment);
				
  return(GP_Stat->I_status);
}

int M_FITSinfo::load_keyword(	const char*	keyword,
				double		value,
				const char*	comment,
				M_Status	*GP_Stat)
{
  int	Index;

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  Index = I_NumKWs;

  /* Increment the counter and allocate the memory to store the strings */
  I_NumKWs++;
  KWlines = (FITS_card*)
    realloc(KWlines, I_NumKWs*sizeof(FITS_card));
  if(KWlines == NULL)
    {GP_Stat->I_status = LD_KEYWORD|MALLOC_FAILED; return(GP_Stat->I_status);}

  /* Write the keyword value into the array */
  KWlines[Index].type = TDOUBLE;
  snprintf(KWlines[Index].keyword, sizeof(KWlines[Index].keyword), "%s", keyword);
  snprintf(KWlines[Index].value, sizeof(KWlines[Index].value), "%f", value);
  snprintf(KWlines[Index].comment, sizeof(KWlines[Index].comment), "%s", comment);
				
  return(GP_Stat->I_status);
}

int M_FITSinfo::load_keyword(	const char*	keyword,
				const char*	value,
				const char*	comment,
				M_Status	*GP_Stat)
{
  int	Index;

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  Index = I_NumKWs;

  /* Increment the counter and allocate the memory to store the strings */
  I_NumKWs++;
  KWlines = (FITS_card*)
    realloc(KWlines, I_NumKWs*sizeof(FITS_card));
  if(KWlines == NULL)
    {GP_Stat->I_status = LD_KEYWORD|MALLOC_FAILED; return(GP_Stat->I_status);}

  /* Write the keyword value into the array */
  KWlines[Index].type = TSTRING;
  snprintf(KWlines[Index].keyword, sizeof(KWlines[Index].keyword), "%s", keyword);
  snprintf(KWlines[Index].value, sizeof(KWlines[Index].value), "%s", value);
  snprintf(KWlines[Index].comment, sizeof(KWlines[Index].comment), "%s", comment);

  /* Accomodate writing TLOGICAL as TSTRING "T" or "F" */
  if(strcmp(value, "T") == 0)	{
    KWlines[Index].type = TLOGICAL;
    sprintf(KWlines[Index].value, "%d", 1);
  }
  if(strcmp(value, "F") == 0)	{
    KWlines[Index].type = TLOGICAL;
    sprintf(KWlines[Index].value, "%d", 0);
  }


				
  return(GP_Stat->I_status);
}

int M_FITSinfo::load_keyword(	const char*	comment,
				M_Status	*GP_Stat)
{
  int	Index;

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  Index = I_NumKWs;


  /* Increment the counter and allocate the memory to store the strings */
  I_NumKWs++;
  KWlines = (FITS_card*)
    realloc(KWlines, I_NumKWs*sizeof(FITS_card));
  if(KWlines == NULL)
    {GP_Stat->I_status = LD_KEYWORD|MALLOC_FAILED; return(GP_Stat->I_status);}


  /* Write the keyword value into the array */
  KWlines[Index].type = FLEN_COMMENT;
  snprintf(KWlines[Index].keyword, sizeof(KWlines[Index].keyword), "%s", "");
  snprintf(KWlines[Index].value, sizeof(KWlines[Index].value), "%s", "");
  snprintf(KWlines[Index].comment, sizeof(KWlines[Index].comment), "%s", comment);

				
  return(GP_Stat->I_status);
}

int M_FITSinfo::load_keyword(	char*	comment,
				M_Status	*GP_Stat)
{
  int	Index;

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  Index = I_NumKWs;


  /* Increment the counter and allocate the memory to store the strings */
  I_NumKWs++;
  KWlines = (FITS_card*)
    realloc(KWlines, I_NumKWs*sizeof(FITS_card));
  if(KWlines == NULL)
    {GP_Stat->I_status = LD_KEYWORD|MALLOC_FAILED; return(GP_Stat->I_status);}


  /* Write the keyword value into the array */
  KWlines[Index].type = FLEN_COMMENT;
  snprintf(KWlines[Index].keyword, sizeof(KWlines[Index].keyword), "%s", "");
  snprintf(KWlines[Index].value, sizeof(KWlines[Index].value), "%s", "");
  snprintf(KWlines[Index].comment, sizeof(KWlines[Index].comment), "%s", comment);

				
  return(GP_Stat->I_status);
}

int M_FITSinfo::load_keyword(	const char*	keyword,
				const char*	comment,
				M_Status	*GP_Stat)
{
  int	Index;

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  Index = I_NumKWs;

  /* Increment the counter and allocate the memory to store the strings */
  I_NumKWs++;
  KWlines = (FITS_card*)
    realloc(KWlines, I_NumKWs*sizeof(FITS_card));
  if(KWlines == NULL)
    {GP_Stat->I_status = LD_KEYWORD|MALLOC_FAILED; return(GP_Stat->I_status);}


  /* Write the keyword value into the array */
  KWlines[Index].type = FLEN_VALUE;
  snprintf(KWlines[Index].keyword, sizeof(KWlines[Index].keyword), "%s", keyword);
  snprintf(KWlines[Index].value, sizeof(KWlines[Index].value), "%s", "");
  snprintf(KWlines[Index].comment, sizeof(KWlines[Index].comment), "%s", comment);
				
  return(GP_Stat->I_status);
}



/*********************************************** 
load_std_keywords()

Purpose

Write keyword values to an array for later output.

Overview
Write the standard keyword values, making a call
to load_keyword for each output keyword. 

Definitions of Variables

External:
See generic_defs.h and generic_exec.c

Internal:
I_fits_return_status = Return value of fits function.
CP_Keyname = Name of Key to be written.
CP_Comment = String value to be written in comment field of header.

***********************************************/

int M_FITSinfo::load_std_keywords(	const char*	CP_ModuleName,
					const char*	CP_Version,
					M_Status 	*GP_Stat)
{
  const	char	STANDARD[]="STANDARD FITSFILE FORMAT";
  const	char	FOURBYTE[]="FOUR_BYTE FLOATING_POINT";

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  load_keyword("SIMPLE  ", "T", STANDARD, GP_Stat);
  if(I_Bitpix==0) load_keyword("BITPIX  ", -32, FOURBYTE, GP_Stat);
  else load_keyword("BITPIX  ", I_Bitpix, "DATA TYPE", GP_Stat);
  if(I_Num_Frames>1) {
    load_keyword("NAXIS", 3, STANDARD, GP_Stat);
    load_keyword("NAXIS1", I_Length_X, STANDARD, GP_Stat);
    load_keyword("NAXIS2", I_Length_Y, STANDARD, GP_Stat);
    load_keyword("NAXIS3", I_Num_Frames, STANDARD, GP_Stat);
  }
  else {
    // NAXIS3 is required by some MIPS modules
    load_keyword("NAXIS", 2, STANDARD, GP_Stat);
    load_keyword("NAXIS1", I_Length_X, STANDARD, GP_Stat);
    load_keyword("NAXIS2", I_Length_Y, STANDARD, GP_Stat);
    load_keyword("NAXIS3", 1, STANDARD, GP_Stat);
  }
  load_keyword("MODULE", CP_ModuleName, "PRODUCED BY MODULE", GP_Stat);
  load_keyword("VERSION", CP_Version, "MODULE VERSION", GP_Stat);

  return(GP_Stat->I_status);
}



/*********************************************** 
output_keywords()

Purpose

Write keyword values to an output fits file.

Overview


Definitions of Variables

External:
See generic.h

Internal:
I_fits_return_status = Return value of fits function.
CP_Keyname = Name of Key to be written.
CP_Comment = String value to be written in comment field of header.

***********************************************/

int M_FITSinfo::output_keywords(M_Status *GP_Stat)
{
  char	STANDARD[]="STANDARD FITSFILE FORMAT";
  char	CP_Keyword[FLEN_KEYWORD];

  int	i=0, I_fits_return_status=0;
  char	CP_Keyname[FLEN_KEYWORD], CP_Comment[FLEN_COMMENT];
  short int	S_Num_In;
  int		I_Num_In, I_Naxis2;
  long		LP_naxes[3];	
  float		F_Num_In;
  double		D_Num_In;


  if(GP_Stat->I_status) return(GP_Stat->I_status);

  if(GP_Stat->I_Debug)
    printf("output_keywords: I_status = %d\n", GP_Stat->I_status);		

  /* Loop through the loaded keywords */
  for(i=0;i<I_NumKWs;i++)	{
    sprintf(CP_Keyname, "%s", KWlines[i].keyword);
    sprintf(CP_Comment, "%s", KWlines[i].comment);
    if(KWlines[i].type==TLOGICAL)	{
      sscanf(KWlines[i].value, "%d", &I_Num_In);
      fits_update_key(ffp_FITS,
		      KWlines[i].type,
		      CP_Keyname,
		      &I_Num_In,
		      CP_Comment,
		      &I_fits_return_status);
    }
    if(KWlines[i].type==TSHORT)	{
      sscanf(KWlines[i].value, "%hd", &S_Num_In);
      fits_update_key(ffp_FITS,
		      KWlines[i].type,
		      CP_Keyname,
		      &S_Num_In,
		      CP_Comment,
		      &I_fits_return_status);
    }
    if(KWlines[i].type==TINT)	{
      sscanf(KWlines[i].value, "%d", &I_Num_In);
      fits_update_key(ffp_FITS,
		      KWlines[i].type,
		      CP_Keyname,
		      &I_Num_In,
		      CP_Comment,
		      &I_fits_return_status);
    }
    if(KWlines[i].type==TFLOAT)	{
      sscanf(KWlines[i].value, "%f", &F_Num_In);
      fits_update_key(ffp_FITS,
		      KWlines[i].type,
		      CP_Keyname,
		      &F_Num_In,
		      CP_Comment,
		      &I_fits_return_status);
    }
    if(KWlines[i].type==TDOUBLE)	{
      sscanf(KWlines[i].value, "%lf", &D_Num_In);
      fits_update_key(ffp_FITS,
		      KWlines[i].type,
		      CP_Keyname,
		      &D_Num_In,
		      CP_Comment,
		      &I_fits_return_status);
    }
    if(KWlines[i].type==TSTRING)	{
      fits_update_key(ffp_FITS,
		      KWlines[i].type,
		      CP_Keyname,
		      KWlines[i].value,
		      CP_Comment,
		      &I_fits_return_status);
    }
    if(KWlines[i].type==FLEN_VALUE)	{
      fits_update_key_null(	ffp_FITS,
				CP_Keyname,
				CP_Comment,
				&I_fits_return_status);
    }
    if(KWlines[i].type==FLEN_COMMENT)	{			
      fits_write_comment(	ffp_FITS,
				CP_Comment,
				&I_fits_return_status);
    }

    if(GP_Stat->I_Debug)	{
      printf("output_keywords: Keyword = %s\t", CP_Keyname);
      /*printf("output_keywords: Type = %d\n", KWlines[i].type);*/
      if(KWlines[i].type==TLOGICAL)
	printf("%d\n", I_Num_In);
      if(KWlines[i].type==TSHORT)
	printf("%d\n", S_Num_In);
      if(KWlines[i].type==TINT)
	printf("%d\n", I_Num_In);
      if(KWlines[i].type==TFLOAT)
	printf("%f\n", F_Num_In);
      if(KWlines[i].type==TDOUBLE)
	printf("%f\n", D_Num_In);
      if(KWlines[i].type==TSTRING)
	printf("%s\n", KWlines[i].value);
      if(KWlines[i].type==FLEN_VALUE)
	printf("%s\n", CP_Comment);
      if(KWlines[i].type==FLEN_COMMENT)
	printf("%s\n", CP_Comment);
      //printf("output_keywords: Status = %d\n", I_fits_return_status);
    }
	
  }	
			
  fits_write_date(ffp_FITS, &I_fits_return_status);		 


  if(I_fits_return_status)
    GP_Stat->I_status = KEYWORDS|FITS_MSG_MASK|I_fits_return_status;		
						
  /* Write the FITS required keywords */
  //if(I_Bitpix==0) I_Bitpix = -32;
  //LP_naxes[0] = I_Length_X;
  //LP_naxes[1] = I_Length_Y;
  //LP_naxes[2] = I_Num_Frames;
  //fits_write_imghdr(ffp_FITS,
  //		  I_Bitpix,
  //		  3,
  //		  LP_naxes,
  //		  &I_fits_return_status);
		  

  /* Delete NAXIS3 keyword and rewrite to make sure it's in the right place */
  sprintf(CP_Keyname, "%s", "NAXIS3  ");
  fits_delete_key(ffp_FITS,
		  CP_Keyname,
		  &I_fits_return_status);

  /* Position to be after NAXIS2 */
  strncpy(CP_Keyname, "NAXIS2", FLEN_KEYWORD);
  fits_read_key(	ffp_FITS,
			TINT,
			CP_Keyname,
			&I_Naxis2, 
			CP_Comment, 
			&I_fits_return_status);

  /* Write the keyword */
  sprintf(CP_Keyname, "%s", "NAXIS3  ");		
  fits_insert_key_lng(	ffp_FITS,
			CP_Keyname,
			I_Num_Frames,
			STANDARD,
			&I_fits_return_status);			

				  
  /* Delete NAXIS3 keyword in case it does not need to be loaded */
  //if(0)
  if(I_Num_Frames==1)	{
    sprintf(CP_Keyname, "%s", "NAXIS   ");
    sprintf(CP_Comment, "%s", "ONE_LAYER FITS FILE");
    I_Num_In = 2;
    fits_update_key(ffp_FITS,
		    TINT,
		    CP_Keyname,
		    &I_Num_In,
		    CP_Comment,
		    &I_fits_return_status);
		


    /* Cheat here and ignore return status -usually when keyword is not there */		
    //I_fits_return_status = 0;
  }

  if(I_fits_return_status)
    GP_Stat->I_status = KEYWORDS|FITS_MSG_MASK|I_fits_return_status;

  return(GP_Stat->I_status);
}


/*********************************************** 
write_data()

Purpose

Write the output fits file.

Overview
Call cfitsio functions.

Definitions of Variables

External:
See generic.h

Internal:
I_Tot_Pixels = Total number of pixel frames in observational image.
I_fits_return_status = Return value of fits function.
ffp_FITS = Handle to output FITS file.


***********************************************/


int M_FITSinfo::write_data(	M_Status	*GP_Stat)
{
  int	I_Tot_Pixels;
  int	I_fits_return_status=0;

  char	CP_Outname[MAX_FILENAME_LENGTH];

  if(GP_Stat->I_status) return(GP_Stat->I_status);

  /*D_Const = SLCP_Const->D_Background;*/
  I_Tot_Pixels = I_Length_X*I_Length_Y*I_Num_Frames;
		 			
  if(GP_Stat->I_Debug)	{
    printf("write_data: Opening for writing \n%s\n", CP_Filename);
    printf("write_data: I_status = %d\n", GP_Stat->I_status);
  }

  /* Add ! to beginning of filename to clobber any existing file */
  sscanf("!", "%c", CP_Outname);  
  sscanf(CP_Filename, "%s", &(CP_Outname[1]));


  /* Optionally copy the header from a specified file to the output file */
  if(strcmp(CP_HdrFile, "NONE") != 0)	{
    /* Open before creating output file, in case the two are the same */
    fits_open_file(	&(ffp_FITS_Hdr), CP_HdrFile, READONLY, &I_fits_return_status);	
		
    if(GP_Stat->I_Debug)	{
      printf("write_data: Opened header copy file for reading \n%s\n", CP_HdrFile);
      printf("write_data: I_fits_return_status = %d\n", I_fits_return_status);
    }

    if(I_fits_return_status)
      {GP_Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status;
      return(GP_Stat->I_status);}

  } /* Conditional on filename = "NONE" */

  /* Open the FITS file for writing */
  fits_create_file(&(ffp_FITS),
		   CP_Outname,
		   &I_fits_return_status);
  if(I_fits_return_status) {
    GP_Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status;
    return(GP_Stat->I_status);}		 

  if(GP_Stat->I_Debug)	{
    printf("write_data: Opened for writing \n%s\n", CP_Filename);
    printf("write_data: I_fits_return_status = %d\n", I_fits_return_status);
  }
			
  if(I_fits_return_status)
    GP_Stat->I_status = KEYWORDS|FITS_MSG_MASK|I_fits_return_status;	


  if(strcmp(CP_HdrFile, "NONE") != 0)	{ 
    /* Copy the keywords from the specified opened fits file */
    fits_copy_header(	ffp_FITS_Hdr,
			ffp_FITS,
			&I_fits_return_status);
			
    /* Close the fits file from which the keywords were read */
    fits_close_file(ffp_FITS_Hdr, &I_fits_return_status);

    if(I_fits_return_status)
      {GP_Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status;
      return(GP_Stat->I_status);}
 
  } /* Conditional on filename = "NONE" */

  /* Overwrite and add to the keywords for this output file */
  output_keywords(GP_Stat);
  if(GP_Stat->I_status) return(GP_Stat->I_status);

  /* Write the results to the output FITS file */
  fits_write_img(	ffp_FITS,
			TDOUBLE,
			1, I_Tot_Pixels,
			(double*) data,
			&I_fits_return_status);
			

  if(I_fits_return_status) 
    GP_Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status;		 

  fits_flush_file(ffp_FITS,
		  &I_fits_return_status);
		
  fits_close_file(ffp_FITS,
		  &I_fits_return_status);

  if(I_fits_return_status)
    GP_Stat->I_status = OUTPUT_MAGS|FITS_MSG_MASK|I_fits_return_status;

  if(GP_Stat->I_Verbose)
    if(GP_Stat->I_status==0)
      printf("write_data: Result written to file:\n%s\n", CP_Filename);
    else
      printf("write_data: Error writing to file:\n%s\n", CP_Filename);

  return(GP_Stat->I_status);
}


/*********************************************** 
write_output()

Purpose

Save output to a fits file.

Overview
Call cfitsio functions which load keywords and set pointer to data.

Definitions of Variables

External:
See generic.h

Internal:
I_Tot_Pixels = Total number of pixel frames in observational image.
I_fits_return_status = Return value of fits function.
ffp_FITS = Handle to output FITS file.


***********************************************/

int M_FITSinfo::write_output(	int		I_OutFileNum,
				M_CmdParse 	*GP_Cmd,
				M_Logfile	*GP_Log,
				M_Namelist	*GP_Name,
				M_FITSinfo	*GP_Infiles,
				void		*GP_Const,
				void		*GP_Init,
		 		void		*GP_Comp,
				M_FITSinfo	*GP_Outfiles,
				M_Status	*GP_Stat)
{				
  char	CP_Flag[MAX_FILENAME_LENGTH];

	
  /* Set the output data file name and pointer to data, etc */
  GP_Name->get_namelist_value(CPP_NamelistNamesOut[I_OutFileNum], CP_Filename);
  sprintf(CP_Flag, "%s%d", "-o", I_OutFileNum+1);
  GP_Cmd->get_flag_value(CP_Flag, GP_Outfiles[I_OutFileNum].CP_Filename);
  /* Function set_output_data is defined in the module's code */
  set_output_data(I_OutFileNum,
		  GP_Infiles,
		  GP_Const,
		  GP_Init,
		  GP_Comp,
		  GP_Outfiles,
		  GP_Stat);
  /* Load the keywords and write the data */	
  load_std_keywords(GP_Log->CP_ModName, GP_Log->CP_Version, GP_Stat);
  /* Function load_extra_keywords is defined in the module's code */
  load_extra_keywords(	I_OutFileNum,
			GP_Log,
			GP_Name,
			GP_Infiles,
			GP_Const,
			GP_Init,
			GP_Comp,
			GP_Outfiles,
			GP_Stat);
  if(strcmp(CP_Filename, "NONE") != 0) write_data(GP_Stat);

  /* Prevent array allocation from being freed; is freed in computation */
  data = NULL;
	
  return(GP_Stat->I_status);
}


