/*********************************************** 
logfile.c

Purpose

Decode error messages and report.

Overview
Append Error Messages to log file.

Definitions of Variables

External:
See generic.h

Internal:
I_Time = Running time of program.
I_applayer_message = Code of error that occurred.
I_applayer_root = Code of function in which error occurred.
I_cfitsio_message = Code of cfitsio error returned.

CP_Message = String interpretation of error code part of status word.
CP_Function = String interpretation of function part of status word.

fp_Logfile = File pointer of logfile.

***********************************************/

#include <string.h>
#include <time.h>

#include "dlmodule_includes.h"

M_Logfile::M_Logfile(	const char*	CP_ModNameIn,
			const char*	CP_VersionIn,
			const char*	CP_FilenameIn,
			M_CmdParse	*GP_Cmd,
			M_Namelist	*GP_Name)
{
char	CP_FilenameL[MAX_FILENAME_LENGTH];

	/* Constructor */
	I_Overwrite = 0;
	I_Time = 0;
	I_applayer_message = 0;
	I_applayer_root = 0;
	I_cfitsio_message = 0;
	sprintf(CP_ModName, "%s", CP_ModNameIn);
	sprintf(CP_Version, "%s", CP_VersionIn);
	sprintf(CP_Filename, "%s", CP_FilenameIn);
	GP_Name->get_namelist_value("Log_Filename", CP_Filename);
	GP_Cmd->get_flag_value("-l", CP_Filename);
	
	/* See whether -L option was used */
	sprintf(CP_FilenameL, "%s", "quizyxx");
	GP_Cmd->get_flag_value("-L", CP_FilenameL);
	if(strcmp(CP_FilenameL, "quizyxx") != 0)	{
	sprintf(CP_Filename, "%s", CP_FilenameL);
	I_Overwrite = 1;
	}
	
}


M_Logfile::~M_Logfile()
{

}

int M_Logfile::log_writer(	M_CmdParse	*GP_Cmd,
				M_Namelist	*GP_Name,
				M_Status	*GP_Stat)
{
int	i;
FILE	*fp_Logfile;

/* Runs without checking status, so can report nonzero status */

if(strcmp(CP_Filename, "stdout") == 0)	{
fprintf(stdout, "log_writer: Printing Log output to stdout\n");
fp_Logfile = stdout;
}
else	{

if(I_Overwrite==0)
fp_Logfile = fopen(CP_Filename, "a");
else
fp_Logfile = fopen(CP_Filename, "w");

if(fp_Logfile==NULL)	{
fprintf(stdout, "log_writer: Could not open Log File, using stdout\n");
fp_Logfile = stdout;
}

}

fprintf(fp_Logfile, "\n");
fprintf(fp_Logfile, "Logfile: %s\n", CP_Filename);
fprintf(fp_Logfile, "Module %s\n", CP_ModName);
fprintf(fp_Logfile, "Version %s\n", CP_Version);

fprintf(fp_Logfile, "Command and flags were \n");
fprintf(fp_Logfile, "%s\n", GP_Cmd->argv[0]);
for(i=0;i<GP_Cmd->I_Numflags;i++)
fprintf(fp_Logfile, "\t\t\t%s\t%s\n", GP_Cmd->flags[i].CP_Flag,
GP_Cmd->flags[i].CP_FlagValue);

fprintf(fp_Logfile, "Namelist: %s\n", GP_Name->CP_Filename);
fprintf(fp_Logfile, "Namelist Values were\n");
for(i=0;i<GP_Name->I_Numnames;i++)
fprintf(fp_Logfile, "\t\t\t%s\t%s\n",
GP_Name->names[i].CP_Name, GP_Name->names[i].CP_Value);

I_applayer_message = RIGHT_MASK&(GP_Stat->I_status);
I_applayer_root = (LEFT_MASK&(GP_Stat->I_status))>>16;

if(I_applayer_message&FITS_MSG_MASK)	{
I_applayer_message -= FITS_MSG_MASK;
ffrprt(fp_Logfile, I_applayer_message);
}

decode_message(I_applayer_message, CP_Message);
decode_fcnname(I_applayer_root, CP_Function);

fprintf(fp_Logfile, "Status Message      0x%04x: %s from Function 0x%04x: %s.\n",
I_applayer_message, CP_Message,  I_applayer_root, CP_Function);

if(GP_Stat->I_NaNFlag)
fprintf(fp_Logfile,
"Warning: %8d   NaN's were produced in the results.\n", GP_Stat->I_NaNCount);

I_Time = clock();
fprintf(fp_Logfile, "Time = %5.3f seconds\n\n", (float)I_Time/CLOCKS_PER_SEC);
//fprintf(fp_Logfile, "\n");

if(fp_Logfile != stdout) fclose(fp_Logfile);

if(GP_Stat->I_Verbose)
printf("log_writer: Log Writer finished.\n");

  return(GP_Stat->I_status);
}



/*********************************************** 
decode_message()

Purpose

Given an error code, write the appropriate error message
into a string.

Overview
Use a switch statement on the error code.

Definitions of Variables

Input:
I_message = Code for error message.


Output:
CP_MessageIn = Error message string.

Internal:


***********************************************/

int M_Logfile::decode_message(int I_message, char *CP_MessageIn)
{

switch(I_message)	{
	case 0:
	sprintf(CP_MessageIn, "%s", "None"); break;
	case NOT_A_NUMBER:
	sprintf(CP_MessageIn, "%s", "NOT_A_NUMBER"); break;
	case MEM_INSUFF:
	sprintf(CP_MessageIn, "%s", "MEM_INSUFF"); break;
	case MAT_SINGULAR:
	sprintf(CP_MessageIn, "%s", "MAT_SINGULAR"); break;
	case POS_EXPECTED:
	sprintf(CP_MessageIn, "%s", "POS_EXPECTED"); break;
	case DIVIDE_ZERO:
	sprintf(CP_MessageIn, "%s", "DIVIDE_ZERO"); break;
	case NNEG_EXPECTED:
	sprintf(CP_MessageIn, "%s", "NNEG_EXPECTED"); break;
	case LOG_NEG:
	sprintf(CP_MessageIn, "%s", "LOG_NEG"); break;
	case ROOT_NEG:
	sprintf(CP_MessageIn, "%s", "ROOT_NEG"); break;
	case FOPEN_FAILED:
	sprintf(CP_MessageIn, "%s", "FOPEN_FAILED"); break;
	case COULDNOTREAD:
	sprintf(CP_MessageIn, "%s", "COULDNOTREAD"); break;
	case DIMMISMATCH:
	sprintf(CP_MessageIn, "%s", "DIMMISMATCH"); break;
	case OUTOFRANGE:
	sprintf(CP_MessageIn, "%s", "OUTOFRANGE"); break;
	case NO_ARGS:
	sprintf(CP_MessageIn, "%s", "NO_ARGS"); break;
	case SYNTAXERR:
	sprintf(CP_MessageIn, "%s", "SYNTAXERR"); break;
	case MALLOC_FAILED:
	sprintf(CP_MessageIn, "%s", "MALLOC_FAILED"); break;
	default:
	sprintf(CP_MessageIn, "%s", "Unknown error");
}	

  return(0);
}


/*********************************************** 
decode_fcnname()

Purpose

Given an function name code, write the appropriate
function name into a string.

Overview
Use a switch statement on the function name code.

Definitions of Variables

Input:
I_root = Code for function name.

Output:
CP_FunctionIn = Name of function.

Internal:


***********************************************/

int	M_Logfile::decode_fcnname(int I_root, char *CP_FunctionIn)
{

switch(I_root)	{
	case 0:
	sprintf(CP_FunctionIn, "%s", "No function"); break;
	case (LEFT_MASK&EXEC)>>16:
	sprintf(CP_FunctionIn, "%s", "EXEC"); break;
	case (LEFT_MASK&INIT_CONSTANTS)>>16:
	sprintf(CP_FunctionIn, "%s", "INIT_CONSTANTS"); break;
	case (LEFT_MASK&INIT)>>16:
	sprintf(CP_FunctionIn, "%s", "INIT"); break;
	case (LEFT_MASK&MKBASIS)>>16:
	sprintf(CP_FunctionIn, "%s", "MKBASIS"); break;
	case (LEFT_MASK&TRANSPMULT)>>16:
	sprintf(CP_FunctionIn, "%s", "TRANSPMULT"); break;
	case (LEFT_MASK&DETCOMP)>>16:
	sprintf(CP_FunctionIn, "%s", "DETCOMP"); break;
	case (LEFT_MASK&PSEUDOCOMP)>>16:
	sprintf(CP_FunctionIn, "%s", "PSEUDOCOMP"); break;
	case (LEFT_MASK&LOG_LIKELYHOOD)>>16:
	sprintf(CP_FunctionIn, "%s", "VEC_MULT"); break;
	case (LEFT_MASK&INNER_PROD)>>16:
	sprintf(CP_FunctionIn, "%s", "INNER_PROD"); break;
	case (LEFT_MASK&LL_COMP)>>16:
	sprintf(CP_FunctionIn, "%s", "LL_COMP"); break;
	case (LEFT_MASK&MAX_LLIKELYHOOD)>>16:
	sprintf(CP_FunctionIn, "%s", "MAX_LLIKELYHOOD"); break;
	case (LEFT_MASK&MAXLL_FINDMAX)>>16:
	sprintf(CP_FunctionIn, "%s", "MAXLL_FINDMAX"); break;
	case (LEFT_MASK&SSORTP)>>16:
	sprintf(CP_FunctionIn, "%s", "SSORTP"); break;
	case (LEFT_MASK&CHARGE_ADDITION)>>16:
	sprintf(CP_FunctionIn, "%s", "CHARGE_ADDITION"); break;
	case (LEFT_MASK&ENDPOINTS)>>16:
	sprintf(CP_FunctionIn, "%s", "ENDPOINTS"); break;
	case (LEFT_MASK&ENDPOINT_FAR)>>16:
	sprintf(CP_FunctionIn, "%s", "ENDPOINT_FAR"); break;
	case (LEFT_MASK&ENDPOINT_NEAR)>>16:
	sprintf(CP_FunctionIn, "%s", "ENDPOINT_NEAR"); break;
	case (LEFT_MASK&PROB_RADHIT)>>16:
	sprintf(CP_FunctionIn, "%s", "PROB_RADHIT"); break;
	case (LEFT_MASK&READ_DATA)>>16:
	sprintf(CP_FunctionIn, "%s", "READ_DATA"); break;
	case (LEFT_MASK&OUTPUT_PROBS)>>16:
	sprintf(CP_FunctionIn, "%s", "OUTPUT_PROBS"); break;
	case (LEFT_MASK&INIT_ALLOCATE)>>16:
	sprintf(CP_FunctionIn, "%s", "INIT_ALLOCATE"); break;
	case (LEFT_MASK&MAT_INVERT)>>16:
	sprintf(CP_FunctionIn, "%s", "MAT_INVERT"); break;
	case (LEFT_MASK&KEYWORDS)>>16:
	sprintf(CP_FunctionIn, "%s", "KEYWORDS"); break;
	case (LEFT_MASK&STD_KEYWORDS)>>16:
	sprintf(CP_FunctionIn, "%s", "STD_KEYWORDS"); break;
	case (LEFT_MASK&CONST_KEYWORDS)>>16:
	sprintf(CP_FunctionIn, "%s", "CONST_KEYWORDS"); break;
	case (LEFT_MASK&FILEN_KEYWORDS)>>16:
	sprintf(CP_FunctionIn, "%s", "FILEN_KEYWORDS"); break;
	case (LEFT_MASK&NAMELIST)>>16:
	sprintf(CP_FunctionIn, "%s", "NAMELIST"); break;
	case (LEFT_MASK&OUTPUT_FLUXES)>>16:
	sprintf(CP_FunctionIn, "%s", "OUTPUT_FLUXES"); break;
	case (LEFT_MASK&OUTPUT_MAGS)>>16:
	sprintf(CP_FunctionIn, "%s", "OUTPUT_MAGS"); break;
	case (LEFT_MASK&OUTPUT_SAMPS)>>16:
	sprintf(CP_FunctionIn, "%s", "OUTPUT_SAMPS"); break;
	case (LEFT_MASK&READ_IMAGE)>>16:
	sprintf(CP_FunctionIn, "%s", "READ_IMAGE"); break;
	case (LEFT_MASK&READ_SLOPES)>>16:
	sprintf(CP_FunctionIn, "%s", "READ_SLOPES"); break;
	case (LEFT_MASK&READ_INTERPS)>>16:
	sprintf(CP_FunctionIn, "%s", "READ_INTERPS"); break;
	case (LEFT_MASK&COMPUTE_RES)>>16:
	sprintf(CP_FunctionIn, "%s", "COMPUTE_RES"); break;
	case (LEFT_MASK&PARSEARGS)>>16:
	sprintf(CP_FunctionIn, "%s", "PARSEARGS"); break;
	case (LEFT_MASK&LD_OUT_KEYWORDS)>>16:
	sprintf(CP_FunctionIn, "%s", "LD_OUT_KEYWORDS"); break;
	case (LEFT_MASK&LD_STD_KEYWORDS)>>16:
	sprintf(CP_FunctionIn, "%s", "LD_STD_KEYWORDS"); break;
	case (LEFT_MASK&LD_KEYWORD)>>16:
	sprintf(CP_FunctionIn, "%s", "LD_KEYWORD"); break;
	default: sprintf(CP_FunctionIn, "%s", "Unspecified function.");
}

  return(0);
}








