/*********************************************** 
namelist.cpp

Purpose
Read and parse a parameter file to get input and output file
names constants, and processing names.

Overview
Use string functions to search for separator characters,
then a series of conditionals to find what variables are
being defined.

Expects a namelist file of the form:

 &GENERIC
  Comment = 'Generic namelist file with default values.',
  FITS_In_Filename= 'img.test.in.fits',
  FITS_Out_Filename='img.test.out.fits',
  Log_Filename 		= 'stdout',
  Layer =1,
 &end

Definitions of Variables
 
External:
See generic.h
 
Local:
fp_namelist = Pointer to open namelist file.
CP_buf = Holds contents of namelist file.
CP_NextWord = First entry in parsed namelist data segment.
Must be name of namelist variable.
CP_Equals = Must be the "=" sign.
CP_Value = Value given to namelist variable.
CP_Comment = Holds comment in namelist file, if any.
CP_Pointer = Keeps track of parsing of namelist data.

I_Got_Namelist = names whether namelist was specified on command line.
I_SectionsFound = Counts how many namelist sections have been found. Not used.

I_fildes_nl = File descriptor for opening namelist file.
I_str_length = Holds length of namelist file.
I_past_end = names if end of namelist file reached.
I_found = names if desired section has been found.
STAT_nlbuf = Holds files status return information.
STATP_nlbuf = Pointer to STAT_nlbuf.

***********************************************/

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>

#ifdef WIN32
#include <io.h>
#endif

#include "dlmodule_includes.h"

#define	STRING_BUFFER_SIZE	 256


M_Namelist::M_Namelist(const char *NLName, const char *NLSection)
{
	/* Constructor */
	sprintf(CP_Filename, "%s", NLName);
	sprintf(CP_SectionName, "%s", NLSection);
	I_Numnames = 0; 
	names = NULL;

}

M_Namelist::~M_Namelist()
{
if(names!=NULL) free(names);
}


int M_Namelist::get_values( M_CmdParse 	*GP_Cmd,
			M_Status	*GP_Stat)
{
	GP_Cmd->get_flag_value("-n", CP_Filename);
	GP_Cmd->get_flag_value("-s", CP_SectionName);

  return(GP_Stat->I_status);
}


int M_Namelist::parse_namelist(M_Status	*GP_Stat)
{


  FILE	*fp_namelist;
  char	*CP_buf=NULL;
  char	CP_NextWord[STRING_BUFFER_SIZE];
  char  CP_Value[STRING_BUFFER_SIZE];
  char	*CP_Pointer, *CP_Comma, *CP_Equals;
     
  int	i=1, I_Got_Namelist=0, I_SectionsFound=0;
  int	I_fildes_nl=0, I_str_length=0;
  int	I_past_end=0, I_found=0;  
  int	I_NameLength=0;
  
  struct stat  STAT_nlbuf, *STATP_nlbuf;  

if(GP_Stat->I_status) return(GP_Stat->I_status); 

/* Skip if no namelist specified */
if(strcmp(CP_Filename, "NONE") == 0) return(GP_Stat->I_status);

/* Open the namelist file to get its length */
#ifdef HAVEOPEN
if ((I_fildes_nl = _open(CP_Filename,O_RDONLY)) == -1)
#else
if ((I_fildes_nl = open(CP_Filename,O_RDONLY)) == -1)
#endif
   {
   printf("parse_namelist: Error opening file %s\n", CP_Filename);
   GP_Stat->I_status = NAMELIST|FOPEN_FAILED; return(GP_Stat->I_status);
   }

/* Find number of characters in namelist file */
STATP_nlbuf = &STAT_nlbuf;
fstat(I_fildes_nl, STATP_nlbuf);
I_str_length = STATP_nlbuf->st_size;


/* Associate a stream with the fildes generated above */
  fp_namelist=fdopen(I_fildes_nl,"r");

/* Allocate memory for namelist buffer, */
  CP_buf = (char *) calloc(I_str_length+2, sizeof(char));

/* Read the namelist data */
  if(fread(CP_buf, sizeof(char), I_str_length, fp_namelist) == 0) {
  printf("parse_namelist: Could not read from file %s", CP_Filename);
  GP_Stat->I_status = NAMELIST|COULDNOTREAD; return(GP_Stat->I_status);
  }
  CP_buf[I_str_length] = '\0';
  
/* Close the stream */
  fclose(fp_namelist);
 
/* Look for the specified section */

   CP_Pointer = CP_buf;
   while(CP_Pointer <= CP_buf + I_str_length && !I_past_end && !I_found){
   CP_Pointer = strchr(CP_Pointer, '&');
   if (CP_Pointer + strlen(CP_SectionName) >= CP_buf + I_str_length) 
   	I_past_end = 1;
   else	if (strncmp(CP_SectionName, CP_Pointer+1,
   	    strlen(CP_SectionName)))
   	CP_Pointer += strlen("&");
   	else I_found = 1;
   }
   
  
   if(I_past_end)
   {GP_Stat->I_status = NAMELIST|COULDNOTREAD; return(GP_Stat->I_status);}

	CP_Pointer += strlen("&");
	CP_Pointer += strlen(CP_SectionName);

/* Match the string values read to the variable definitions and
   write the  values into the appropriate variables. */
	for(CP_Comma = strtok(CP_Pointer, ","); 
	CP_Comma != NULL && strchr(CP_Comma, '&') == 0;
	CP_Comma = strtok(NULL, ","))	{

	CP_Equals = strstr(CP_Comma, "=");
        if(CP_Equals)
	   *CP_Equals = '\0';
	sprintf(CP_NextWord, "%s", CP_Comma);
	sscanf(CP_NextWord, "%s", CP_NextWord);
	sscanf(CP_Equals+1, "%s", CP_Value);
	
	/* Allocate more memory for names and value */
	names =
	(M_NameVal*) realloc(names, (I_Numnames+1)*sizeof(M_NameVal));
	if(names==NULL)
	{printf("parse_namelist: could not allocate memory\n");
	 GP_Stat->I_status = NAMELIST|MALLOC_FAILED; return(GP_Stat->I_status);}
	sprintf(names[I_Numnames].CP_Name, "%s", CP_NextWord);
	sprintf(names[I_Numnames].CP_Value, "%s", CP_Value);
	
	I_Numnames++;			
	}

  if(CP_buf!=NULL) free(CP_buf);

  
if(GP_Stat->I_Verbose)	
printf("parse_namelist: Namelist File = %s\n", CP_Filename);

if(GP_Stat->I_Debug)
for(i=0;i<I_Numnames;i++)
printf("parse_namelist: %s\t%s\n",
names[i].CP_Name, names[i].CP_Value);


  return(GP_Stat->I_status);
}


/*********************************************** 
get_namelist_value

Purpose
Look up the value of the named variable in the 
namelist information.

Overview
Loop through the table of values produced by
parse_namelist, seeking the designatated name, and
overwrite the variable's value if the name is found.

Definitions of Variables

External:
See generic.h


Internal:


***********************************************/


int M_Namelist::get_namelist_value(const char* CP_Name, unsigned short& US_val)
{
int		i;
for(i=0;i<I_Numnames;i++)
if(strcmp(CP_Name, names[i].CP_Name) == 0)	{
sscanf(names[i].CP_Value, "%hd", &US_val);
if(0)printf("get_namelist_value: Name = %s\t New Value = %d\n", CP_Name, US_val);
}
  return(0);
}


int M_Namelist::get_namelist_value(const char* CP_Name, short& S_val)
{
int		i;
for(i=0;i<I_Numnames;i++)
if(strcmp(CP_Name, names[i].CP_Name) == 0)	{
sscanf(names[i].CP_Value, "%hd", &S_val);
if(0)printf("get_namelist_value: Name = %s\t New Value = %d\n", CP_Name, S_val);
}
  return(0);
}


int M_Namelist::get_namelist_value(const char* CP_Name, unsigned int& UI_val)
{
int		i;
for(i=0;i<I_Numnames;i++)
if(strcmp(CP_Name, names[i].CP_Name) == 0)	{
if(strcmp(names[i].CP_Value, "") == 0)
UI_val = 1 - UI_val;
else
sscanf(names[i].CP_Value, "%d", &UI_val);
if(0)printf("get_namelist_value: Name = %s\t New Value = %d\n", CP_Name, UI_val);
}
  return(0);
}


int M_Namelist::get_namelist_value(const char* CP_Name, int& I_val)
{
int		i;
for(i=0;i<I_Numnames;i++)
if(strcmp(CP_Name, names[i].CP_Name) == 0)	{
if(strcmp(names[i].CP_Value, "") == 0)
I_val = 1 - I_val;
else
sscanf(names[i].CP_Value, "%d", &I_val);
if(0)printf("get_namelist_value: Name = %s\t New Value = %d\n", CP_Name, I_val);
}
  return(0);
}

int M_Namelist::get_namelist_value(const char* CP_Name, long& L_val)
{
int		i;
for(i=0;i<I_Numnames;i++)
if(strcmp(CP_Name, names[i].CP_Name) == 0)	{
if(strcmp(names[i].CP_Value, "") == 0)
L_val = 1 - L_val;
else
sscanf(names[i].CP_Value, "%d", &L_val);
if(0)printf("get_namelist_value: Name = %s\t New Value = %d\n", CP_Name, L_val);
}
  return(0);
}



int M_Namelist::get_namelist_value(const char* CP_Name, float& F_val)
{
int		i;
for(i=0;i<I_Numnames;i++)
if(strcmp(CP_Name, names[i].CP_Name) == 0)	{
sscanf(names[i].CP_Value, "%f", &F_val);
if(0)printf("get_namelist_value: Name = %s\t New Value = %f\n", CP_Name, F_val);
}
  return(0);
}


int M_Namelist::get_namelist_value(const char* CP_Name, double& D_val)
{
int		i;
for(i=0;i<I_Numnames;i++)
if(strcmp(CP_Name, names[i].CP_Name) == 0)	{
sscanf(names[i].CP_Value, "%lf", &D_val);
if(0)printf("get_namelist_value: Name = %s\t New Value = %f\n", CP_Name, D_val);
}
  return(0);
}


int M_Namelist::get_namelist_value(const char* CP_Name, char*	CP_val)
{
int		i;
for(i=0;i<I_Numnames;i++)
if(strcmp(CP_Name, names[i].CP_Name) == 0)	{
sprintf(CP_val, "%s", names[i].CP_Value);
if(0)printf("get_namelist_value: Name = %s\t New Value = %s\n", CP_Name, CP_val);
}
  return(0);
}
