/* $Id: datascale_estimate.c,v 1.1 2001/05/14 23:55:07 laher Exp $ */


/*********************************************** 
datascale_estimate.c

Purpose

Implements robust estimation method of Rousse and Croux (1993)
for scale of a distribution of data.

Reference:

     Rousseeuw, P.J. and Croux, C. (1993), "Alternatives to the
        Median Absolute Deviation," Journal of the American
        Statistical Association, Vol. 88, 1273-1283.

Number of data points must be greater than or equal to 2 and
less than or equal to 100000.

************************************************/

#include <stdio.h>

/* Includes for mathc90 sort routine. */
/* path to mathc90 and fcrt: 
/stage/ssc-pipe/hemple/sw/downlink/libsrc/mathc90/
*/
#include "mathc90.h"
#include <stdlib.h>
#include <math.h>
#include "fcrt.h"


datascale_estimate(double x[], long n, double *Sn) {

   int	i, k, tryA, tryB, diff, Amin, Amax, even, half;
   int leftA, leftB, rightA, rightB, nA, nB, length;

   long nstart;

   double *y, *a2, medA, medB, cn, pull;

   y = (double *) calloc(n + 1000, sizeof(double));

   a2 = (double *) calloc(n + 1000, sizeof(double));

   nstart = 1;
   dsort(x, nstart, n);

   for (i = n; i >= 1 ; i--) {
      y[i] = x[i-1];
   }

   a2[1] = y[n/2+1] - y[1];

   for (i = 2; i <= (n + 1) / 2; i++) {

      nA = i - 1;
      nB = n - i;
      diff = nB - nA;
      leftA = 1;
      leftB = 1;
      rightA = nB;
      rightB = nB;
      Amin = diff / 2 + 1;
      Amax = diff / 2 + nA;

loopback:

      if (leftA < rightA) {
         length = rightA - leftA + 1;
         even = 1 - length % 2;
         half = (length - 1) / 2;
         tryA = leftA + half;
         tryB = leftB + half;
         if (tryA < Amin) {
            rightB = tryB;
            leftA = tryA + even;
         } else {
	    if (tryA > Amax) {
               rightA = tryA;
               leftB = tryB + even;           
	    } else {
               medA = y[i] - y[i - tryA + Amin - 1];
               medB = y[tryB + i] - y[i];
               if (medA >= medB) {
                  rightA = tryA;
                  leftB = tryB + even;
               } else {
                  rightB = tryB;
                  leftA = tryA + even;
               }
            }
         }
         goto loopback;
      }
      if (leftA > Amax) {
         a2[i] = y[leftB + i] - y[i];
      } else {
         medA = y[i] - y[i - leftA + Amin - 1];
         medB = y[leftB + i] - y[i];

         a2[i] = medA;
         if (medB < medA) a2[i] = medB;
      }
   }

   for (i = (n + 1) / 2 + 1; i <= n - 1; i++) {
      nA = n - i;
      nB = i - 1;
      diff = nB - nA;
      leftA = 1;
      leftB = 1;
      rightA = nB;
      rightB = nB;
      Amin = diff / 2 + 1;
      Amax = diff / 2 + nA;

loopback2:

      if (leftA < rightA) {
         length = rightA - leftA + 1;
         even = 1 - length % 2;
         half = (length - 1) / 2;
         tryA = leftA + half;
         tryB = leftB + half;
         if (tryA < Amin) {
            rightB = tryB;
            leftA = tryA + even;
         } else {
	    if (tryA > Amax) {
               rightA = tryA;
               leftB = tryB + even;           
	    } else {
               medA = y[i + tryA - Amin + 1] - y[i];
               medB = y[i] - y[i - tryB];
               if (medA >= medB) {
                  rightA = tryA;
                  leftB = tryB + even;
               } else {
                  rightB = tryB;
                  leftA = tryA + even;
               }
            }
         }
         goto loopback2;
      }
      if (leftA > Amax) {
         a2[i] = y[i] - y[i - leftB ];
      } else {
         medA = y[i + leftA - Amin + 1] - y[i];
         medB = y[i] - y[i - leftB];
         a2[i] = medA;
         if (medB < medA) a2[i] = medB;
      }
   }
   a2[n] = y[n] - y[(n + 1) / 2];
   cn = 1.0;
   if (n <= 9) {
      if (n == 2) cn = 0.743;
      if (n == 3) cn = 1.851;
      if (n == 4) cn = 0.954;
      if (n == 5) cn = 1.351;
      if (n == 6) cn = 0.993;
      if (n == 7) cn = 1.198;
      if (n == 8) cn = 1.005;
      if (n == 9) cn = 1.131;
   } else {
      if (n % 2 == 1) 
         cn = ((double) n) / ((double) n - 0.9);
   }

   free(y);

   k = (n + 1) / 2;
   pull_datascale_estimate(a2, n, k, &pull);

   *Sn = cn * 1.1926 * pull;

   free(a2);

}
