/*
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@@ PROGRAM:  integrate_area.c
@@
@@ PURPOSE:  Integrate the contents of an image within a quadrilateral
@@
@@ ROUTINES: integrate_quad(...) -- routine for external use.
@@
@@ USES:     poly_clip.c, integrate.h
@@
@@ TESTING:  integrate_test.c
@@
@@ PROGRAMMERS:
@@	CW  - Clare Waterson
@@
@@ MODIFICATION HISTORY:  (set INTEGRATE_AREA_VERSION in integrate.h)
@@	Ver.    Date      Prg.  What
@@ 	0.90  2000-08-31  CW    Initial Version
@@	0.91  2000-09-05  CW    Eliminate erroneous error messages found by
@@	                        Bob Narron; test more image boundary cases.
@@	                        Use "CAUTION" to avoid a "bad word"
@@	                        Change bool to boolean.
@@	0.92  2000-09-07  CW    Remove exit(1) conditions.
@@	1.00  2002-04-28  CW    Add distribute_quad() and distribute_quad_BB().
@@                              and integrate_quad_2() -- packaged for others.
@@	1.01  2002-04-28  CW    Update integrate_quad_2()
@@      1.02  2002-07-03  CW    Fixed bugs.  Added integrate_name() and
@@                              integrate_version()
@@      1.10  2002-08-08  CW    Added three new routines for David Makovoz:
@@                              integrate_two_quad()         CR3030
@@                              distribute_two_quad()        CR3031
@@                              unweighted_distribute_quad() CR3032
@@      1.11  2002-08-xx  CW    in unweighted_distribute_quad():
@@                                  REPLACED image[j] += pseudopixel_value;
@@                                  WITH     image[j]  = pseudopixel_value;
@@      1.20  2003-05-29  CW    Added one new routine for David Makovoz:
@@                                distribute_quad_n()        CR4270
@@      1.21  2003-09-23  CW    Added integrate_quad_3 for mask combining
@@      1.22  2003-09-23  CW    Updated integrate_quad_3
@@      1.30  2004-02-04  CW    Added two new routines for David Makovoz:
@@                                mask_distribute_quad_short CR5496
@@                                mask_distribute_quad_int
@@      1.31  2006-05-03  DGH   added nanvalue.h invoked NANVALUE for nan's
@@      1.32  2007-02-20  IRK   replace isnan vy functions in nanlib       
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>       /* Contains atof(), used to generate a NaN */
#include "integrate.h"
#include "nanvalue.h"     /* improved nan library routine, multiplatform */

/*===========================================================================
** Return the name and version of this module/library"
*/
char* integrate_name( void )
{
   return( INT_NAME ) ;
}
char* integrate_version( void )
{
   return( INT_VERSION) ;
}


/*===========================================================================
** Find the area of an N-point polygon.   (npts is 4 to 8 points)
** Use the formula from Math CRC 27th ed. p. 194:
**   Area of Polygon P[0]P[1]...P[N-1] =
**      1/2 ABS(   X[0]Y[1] + X[1]Y[2] + ... + X[N-2]Y[N-1] + X[N-1]Y[0]
**               - Y[0]X[1] - Y[1]X[2] - ... - Y[N-2]X[N-1] - Y[N-1]X[0] )
** NOTE: ABS() is necessary because the area is positive for 
**   CCW-ordered points, and negative for CW-ordered points.
**===========================================================================
*/
double area_poly(poly_type *poly)
{
int i,j,npts;
double area;

  npts = (poly->len) - 1; /* don't use the repeated last vertex */
  area = 0.0;
  for ( i=0; i<npts; i++ ){   
     j = (i+1) % npts;
     area += poly->p[i].x * poly->p[j].y;
     area -= poly->p[i].y * poly->p[j].x;
  }
  return( 0.5 * fabs(area) );
     
}     

/*===========================================================================
** Find the area of an N-point polygon.   (npts is 4 to 8 points)
** Use the formula from Math CRC 27th ed. p. 194:
**   Area of Polygon P[0]P[1]...P[N-1] =
**      1/2 ABS(   X[0]Y[1] + X[1]Y[2] + ... + X[N-2]Y[N-1] + X[N-1]Y[0]
**               - Y[0]X[1] - Y[1]X[2] - ... - Y[N-2]X[N-1] - Y[N-1]X[0] )
** NOTE: ABS() is necessary because the area is positive for 
**   CCW-ordered points, and negative for CW-ordered points.
**===========================================================================
*/
double area_poly1(point_type *pts, int npts)
{
int i,j;
double area;

  area = 0.0;
  for ( i=0; i<npts; i++ ){   
     j = (i+1) % npts;
     area += pts[i].x * pts[j].y;
     area -= pts[i].y * pts[j].x;
  }
  return( 0.5 * fabs(area) );
     
}     


/*****************************************************************************
** Stuff four vertices into a poly_type structure (five points to be closed)
*/
poly_type stuff_four( double rx1, double ry1,  double rx2, double ry2, 
                      double rx3, double ry3,  double rx4, double ry4  )
{
poly_type return_poly;

	return_poly.p[0].x = rx1; 
	return_poly.p[0].y = ry1;
	return_poly.p[1].x = rx2;
	return_poly.p[1].y = ry2;
	return_poly.p[2].x = rx3;
	return_poly.p[2].y = ry3;
	return_poly.p[3].x = rx4;
	return_poly.p[3].y = ry4;
	return_poly.p[4].x = rx1;
	return_poly.p[4].y = ry1;

        return_poly.len    = 5;
        
        return( return_poly );

}                      
                      
/*********************************************************************
** Print out the points contained in a polygon
*/
void print_poly(poly_type *poly)
{
int	i;

   printf("Closed Poly of %1d pts: ", poly->len);
   if ( ( (poly->len) < 0 ) || ( (poly->len) > MAX_POLY_LEN ) )  {
      printf("ERROR in print_poly: poly->len=%1d   Exiting routine.\n\n",poly->len);
      return;
   }

   for ( i=0; i<(poly->len); i++ ) {
      printf("[%5.3f %5.3f] ", poly->p[i].x, poly->p[i].y );
   }
   printf("\n");
   
   if ( (poly->len) == 0 ) return;
   
   if ( ( poly->p[0].x != poly->p[(poly->len)-1].x ) ||
        ( poly->p[0].y != poly->p[(poly->len)-1].y )    )  {
      printf("ERROR in print_poly: poly->len=%1d   \n\n",poly->len);
      printf("ERROR: POLY NOT CLOSED! Exiting routine.\n\n");
      return;
   }
}

/*****************************************************************************
** Implement the mathematical sign() function , used in verify_cyclical()
*/
int sign( double arg )
{
   if (arg==0.) return(  0 );
   if (arg>0)   return(  1 );
   else		return( -1 );
}   

/*****************************************************************************
** Verify that a quadrilateral is connected in a strictly cyclical manner
** REQUIRE that poly be CLOSED.
*/
int verify_cyclical( poly_type poly )
{
point_type	vectors[MAX_POLY_LEN];	/* Vectors: p[j+1]-p[j], in order */
point_type 	v, w;			/* two vectors */
int		i;
double		cp;			/* Z cross product of two vectors */
int 		sign_cp0;		/* Sign of the first nonzero cp */

   /* printf("Verify Cyclical:\n"); */
   if( poly.len < 3 ) return( TRUE );  /* Degenerate case: not a problem */

   for( i=0; i<poly.len-1; i++ ) {
      vectors[i].x = poly.p[i+1].x - poly.p[i].x; 
      vectors[i].y = poly.p[i+1].y - poly.p[i].y; 
   }

   sign_cp0 = 0;
   v = vectors[poly.len-2];  /* Set vector v to the last vector */
   for( i=0; i<poly.len-1; i++ ) {
      w  = vectors[i];
      cp = (v.x * w.y) - (v.y * w.x);  /* Z cross product of v and w */
      if (sign_cp0==0) { 
         sign_cp0 = sign(cp);
      }
      else {                            /* sign(cp) must be the same for all */
         if ( sign(cp) != sign_cp0 ) {  /*   cyclically-connected vectors    */
            if ( cp != 0 ) return( FALSE );  /* ignore vectors with cp==0 */
         }
      }
      v  = w;  /* ready for next pass of for loop */
   }
   return( TRUE );

}

/*****************************************************************************
** rmin4 -- Find the minimum of four real values, and take the (int)floor() of it
*/
int rmin4floor(double r1, double r2, double r3, double r4)
{
double rm;
  rm = ( (r1 < r2) ? r1 : r2 );
  rm = ( (rm < r3) ? rm : r3 );
  rm = ( (rm < r4) ? rm : r4 );
  return( (int)floor(rm) );
}

/*****************************************************************************
** rmax4 -- Find the minimum of four real values, and take the (int)ceil() of it
*/
int rmax4ceil(double r1, double r2, double r3, double r4)
{
double rm;
  rm = ( (r1 > r2) ? r1 : r2 );
  rm = ( (rm > r3) ? rm : r3 );
  rm = ( (rm > r4) ? rm : r4 );
  return( (int)ceil(rm) );
}


/************************************************************************
** Find the bounding rectangle for a quadrilateral.
** Each pixel [i,j] is defined from [i,j] <= pt < [i+1,j+1], so
** Define the bounding rectangle such that
**     for a minimum: [i <= rmin <  i+1] maps to i   so rmin -> floor(rmin)
** and for a maximum: [i <  rmax <= i+1] maps to i   so rmax ->  ceil(rmax)-1
**
**    ixmin,ixmax;	Return values: set to the bounding rectangle containing
**    iymin,iymax;  	the quadrilateral (int pixels)	 
**
**    ixdim,iydim;      Maximum dimensions of the full image:  0..ixdim-1
*/
void  bounding_rect( poly_type pp, int *ixmin, int *ixmax, int *iymin, int *iymax ,
                     int ixdim, int iydim )
{

   *ixmin = rmin4floor( pp.p[0].x, pp.p[1].x, pp.p[2].x, pp.p[3].x );
   *iymin = rmin4floor( pp.p[0].y, pp.p[1].y, pp.p[2].y, pp.p[3].y );
   *ixmax = rmax4ceil(  pp.p[0].x, pp.p[1].x, pp.p[2].x, pp.p[3].x )-1;
   *iymax = rmax4ceil(  pp.p[0].y, pp.p[1].y, pp.p[2].y, pp.p[3].y )-1;
   if ( *ixmin <  0     ) *ixmin = 0;       
   if ( *iymin <  0     ) *iymin = 0;       
   if ( *ixmax >= ixdim ) *ixmax = ixdim-1; 
   if ( *iymax >= iydim ) *iymax = iydim-1; 

}


/*========================= PRIMARY EXPORTED ROUTINES ======================*/


/*****************************************************************************
** Integrate Quad:  Integrate the contents of an image within a quadrilateral.
** See illustrated example below in "MAIN PROCESSING LOOP"
**  
** INPUTS:
**	rx1,ry1, rx2,ry2, 	Cyclical points around the quadrilateral
**      rx3,ry3, rx4,ry4	(marked with "r" to remind that they're real)
**      image[]			The image to be integrated
**	ixdim,iydim		Dimensions of the image
**
** OUTPUTS:
**	area			The area of the quadrilateral, in pixels.
**				Only the area within the image dimensions
**				0..xdim and 0..ydim will be returned.
**				All parts of the quad outside the image dims
**				are defined to have zero area.
**	RETURN VALUE		(integration) The integrated contents of the 
**				quadrilateral (within the image dims).
**
** DEFINE:
**	Pixels are defined from their index (i,j) to extend to 
**	strictly less than index (i+1,j+1).
**	The lowest index in the image is pixel (0,0).
**	The maximum real index values that may be addressed are 
**	x<=xdim and y<=ydim.  For the max index pixels ONLY, the
**	pixel is defined fully from (i,j) to (i+1,j+1) INLCUSIVE.
**	REQUIRES input quadrilateral to be CONVEX, defined cyclically.
**	The area of an unclipped pixel is 1, and the area of a 
**	clipped pixel is thus the fractional part of the full pixel
**	that is covered by the intersection of the pixel and the polygon.
**
**	Clip the Quadrilateral at the boundary of the image, and
**	return zero for any or all of the areas of the image that
**	were outside the image (Requested by Bob Narron).
**
**      Use the equivalence: 
**	   for an image im[] with dimensions (xdim,ydim), use
**	   im[x,y] == (y * xdim) + x
**
** PSEUDOCODE:
**
**	* Find the integer-valued bounding rectangle for the quadrilateral
**	* Initialize: integration=0 and sum_area=0
**	* For each pixel in the bounding rectangle
**	  * Clip each pixel against the quadrilateral: produces clipped_pixel
**	  * Calculate the area of clipped_pixel
**	  * Add this area to sum_area 
**	  * Add (area * image pixel value) to integration
**	* Calculate the area_quad of the quadrilateral directly
**	* Check area_quad against direct calculation of the area, area_calc
**	* Return values of integration and area_sum 
**
*/
double integrate_quad( double rx1, double ry1,  /* The four vertices of the */
                       double rx2, double ry2,	/*    quadrilateral to be   */
                       double rx3, double ry3,	/*    integrated within     */
                       double rx4, double ry4, 
                       double image[],		/* Image to be integrated   */
                       int ixdim, int iydim,	/* Dimensions of the image  */
                       double *area          )	/* Returned area of the     */
                       				/*    quadrilateral within  */
                                                /*    the bounds of the     */
                                                /*    image.                */
{
int		ix,iy, iy_offset;  /* Pixel counters within the image */
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
double		clipped_area;	/* Area of the clipped overlap region within */
				/*    both the pixel and the quadrilateral   */
double		integration;	/* Accumulates the desired integration value */
double		area_sum;	/* Accumulates the desired overlap area      */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
poly_type	quad; 		/* Holds a closed quadrilateral              */
poly_type	clipped_pixel;  /* Holds the vertices of the clipped overlap */
double		area_quad;	/* Quadrilateral area, computed directly     */
double		err_frac;	/* Check accuracy of area_sum vs. area_quad  */
boolean		pt_out;		/* A point in the quad is outside the image  */
int		pixel_count;    /* Number of pixels processed within a quad  */
int		pixels_skipped; /* Number of NaN pixels skipped within quad  */

   /* REMEMBER: for any edit changes, change ALL of:                      */
   /*    integrate_quad() AND integrate_quad_2() AND integrate_two_quad() */
   

   /************************************************************************
   ** Find the bounding rectangle for the quadrilateral.
   ** Each pixel [i,j] is defined from [i,j] <= pt < [i+1,j+1], so
   ** Define the bounding rectangle such that
   **     for a minimum: [i <= rmin <  i+1] maps to i   so rmin -> floor(rmin)
   ** and for a maximum: [i <  rmax <= i+1] maps to i   so rmax ->  ceil(rmax)-1
   */
   ixmin = rmin4floor( rx1, rx2, rx3, rx4 );
   iymin = rmin4floor( ry1, ry2, ry3, ry4 );
   ixmax = rmax4ceil(  rx1, rx2, rx3, rx4 )-1;
   iymax = rmax4ceil(  ry1, ry2, ry3, ry4 )-1;

   pt_out = FALSE;

   if ( ixmin <  0     ) { pt_out  =  TRUE;  ixmin = 0;       }
   if ( iymin <  0     ) { pt_out  =  TRUE;  iymin = 0;       }
   if ( ixmax >= ixdim ) { pt_out  =  TRUE;  ixmax = ixdim-1; }
   if ( iymax >= iydim ) { pt_out  =  TRUE;  iymax = iydim-1; }
   
   quad = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );


   /************************************************************************
   ** MAIN PROCESSING LOOP
   ** For each pixel in the bounding rectangle: clip the quadrilateral to
   ** the pixel boundaries, calculate its area, add to sums for total area 
   ** and integrated value.
   **
   ** Added support for NaNs:
   **
   **
   **   +------------+------------+
   **   | 10         |        NaN |  EXAMPLE:
   **   |           W|            |  Quadrilateral WXYZ overlaps pixels
   **   |          / \            |  A, B, C, and D, with
   **   |         /  |\           |  corresponding overlap areas
   **   |   A    /   | \     B    |  .3, .2, .1, and .3 .  Pixels   
   **   |       /    |  \         |  A, B, C, and D, have image values
   **   |      /     |   \        |  10, NaN, 20, and 30, respectively.
   **   |     /      |    \       |
   **   |    Z       |     \      |  integrate_quad calculates:      
   **   |     \    .3|.2    \     |
   **   +------\-----+-------\----+  area_sum = .3 + (skip)
   **   |       \  .1|.3      X   |           + .1 + .3            = .7
   **   |        \   |       /    |           
   **   |         \  |      /     |  integration = .3*10 + (skip)
   **   |          \ |     /      |              + .1*20 + .3*30   = 14
   **   |   C       \|    /  D    |
   **   |            \   /        |  so Averaged Value = integration/area_sum 
   **   |            |\ /         |                    = 20
   **   |            | Y          |
   **   | 20         |         30 |
   **   +------------+------------+
   **
   **         W       
   **        /#*      Effectively, the area and integral in the 
   **       /##|`     above quadrilateral is calculated as if the
   **      /###| `    region containing the NaN wasn't there.
   **     Z####|  `   
   **      \###|   `  
   **       \##*----*  
   **        \#######X 
   **         \#####/  
   **          \###/   
   **           \#/    
   **            Y     
   **                  
   **   EXCEPT:  In the special case that all pixel values overlapped by the 
   **            quadrilateral are NaNs, the returns are set to:
   **
   **                  area_sum = 0  and  integration = NaN 
   **
   */
   area_sum       = 0;
   integration    = 0;
   pixel_count    = 0;
   pixels_skipped = 0;
   for( iy=iymin,iy_offset=iymin*ixdim; iy<=iymax; iy++,iy_offset+=ixdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon (pixmin, pixmax, &quad, &clipped_pixel);
         clipped_area = area_poly( &clipped_pixel );
         if (clipped_area > 0) {
            pixel_count++;
            if ( !iznand( image[ix+iy_offset] ) ) {
               area_sum    += clipped_area;
               integration += clipped_area * image[ix+iy_offset];
            } else {
               pixels_skipped++;
            }
         }
      }
   }
   
   /* If no "good data" within quad (ALL NaNs), return a NaN (area_sum==0 already) */
   if ( (pixels_skipped == pixel_count) && (pixels_skipped > 0) ) {
     integration =NANVALUE;
   }

   /********************************************************************
   ** If polygon is entirely within the pixel, then the area_sum should
   ** be equal to the area of the polygon computed directly (area_quad)
   ** Done for paranoid verification purposes only (may eliminate in final code)
   ** Use the word "CAUTION" to avoid the official "bad words"
   **
   ** COMMENTED OUT, BECAUSE NaNS DESTROY THIS CHECK.
   **
   if (!pt_out) { 
      area_quad = area_poly( &quad );  
      err_frac  = fabs((area_quad - area_sum)/(area_quad));
      if ( err_frac > TOL ) {
         printf("integrate_quad: CAUTION: Area error fraction, %g,", err_frac );
         printf(" exceeded TOL, %g  (Please notify Clare.)\n", TOL );                
         printf("in integrate_area.c: VERSION %1d\n\n", INTEGRATE_AREA_VERSION );      }
   }
   */

   /********************************************************************
   ** Return final answers
   */
   *area = area_sum;
   return( integration );
   
}


/*****************************************************************************
** Integrate Quad "2":  Integrate the contents of an image within a 
** quadrilateral, AND the array of error values, except that instead of 
** summing the error values in pixels, the sums of squares of error values 
** and pixel overlap areas are used.
**
** Error values are assumed to be passed as sigma (not sigma squared)
**
** Made into an entirely separate routine from integrate_quad() for SPEED.
** Both the regular area AND the sum of squared area-weights (area_sq) 
** are provided.
*/
double integrate_quad_2( double rx1,     double ry1,
                         double rx2,     double ry2,
                         double rx3,     double ry3,
                         double rx4,     double ry4,
                         double image[], double sigmas[],	
                         int ixdim,      int iydim,	
                         double *area,   double *area_sq, double *sigma_result )
                                          
{
int		ix,iy, iy_offset;  /* Pixel counters within the image */
int		ixy;		/* Holds ix + iy_offset                      */
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
double		clipped_area;	/* Area of the clipped overlap region within */
				/*    both the pixel and the quadrilateral   */
double		integration1;	/* Accumulates the desired integration value */
double		integration2;	/* Accumulates the integration of squares    */
double		area_sum;	/* Accumulates the desired overlap area      */
double		area_sum_sq;	/* Accumulates the desired overlap area^2    */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
poly_type	quad; 		/* Holds a closed quadrilateral              */
poly_type	clipped_pixel;  /* Holds the vertices of the clipped overlap */
double		area_quad;	/* Quadrilateral area, computed directly     */
double		err_frac;	/* Check accuracy of area_sum vs. area_quad  */
boolean		pt_out;		/* A point in the quad is outside the image  */
int		pixel_count;    /* Number of pixels processed within a quad  */
int		pixels_skipped; /* Number of NaN pixels skipped within quad  */
double		square;		/* Square of clipped_area                    */

   /* REMEMBER: for any edit changes, change ALL of:                      */
   /*    integrate_quad() AND integrate_quad_2() AND integrate_two_quad() */


   /*2**********************************************************************
   ** Find the bounding rectangle for the quadrilateral.
   ** Each pixel [i,j] is defined from [i,j] <= pt < [i+1,j+1], so
   ** Define the bounding rectangle such that
   **     for a minimum: [i <= rmin <  i+1] maps to i   so rmin -> floor(rmin)
   ** and for a maximum: [i <  rmax <= i+1] maps to i   so rmax ->  ceil(rmax)-1
   */
   ixmin = rmin4floor( rx1, rx2, rx3, rx4 );
   iymin = rmin4floor( ry1, ry2, ry3, ry4 );
   ixmax = rmax4ceil(  rx1, rx2, rx3, rx4 )-1;
   iymax = rmax4ceil(  ry1, ry2, ry3, ry4 )-1;
   
   pt_out = FALSE;
   
   if ( ixmin <  0     ) { pt_out  =  TRUE;  ixmin = 0;       }
   if ( iymin <  0     ) { pt_out  =  TRUE;  iymin = 0;       }
   if ( ixmax >= ixdim ) { pt_out  =  TRUE;  ixmax = ixdim-1; }
   if ( iymax >= iydim ) { pt_out  =  TRUE;  iymax = iydim-1; }
   
   quad = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );


   /*2**********************************************************************
   ** MAIN PROCESSING LOOP
   ** For each pixel in the bounding rectangle: clip the quadrilateral to
   ** the pixel boundaries, calculate its area, add to sums for total area,
   ** total squared overlap areas and integrated squared values.
   **
   ** Added support for NaNs:
   **
   **
   **   +------------+------------+
   **   | 10         |        NaN |  EXAMPLE:
   **   |           W|            |  Quadrilateral WXYZ overlaps pixels
   **   |          / \            |  A, B, C, and D, with
   **   |         /  |\           |  corresponding overlap areas
   **   |   A    /   | \     B    |  .3, .2, .1, and .3 .  Pixels   
   **   |       /    |  \         |  A, B, C, and D, have image values
   **   |      /     |   \        |  10, NaN, 20, and 30, respectively.
   **   |     /      |    \       |
   **   |    Z       |     \      |  integrate_quad_2 calculates:      
   **   |     \    .3|.2    \     |
   **   +------\-----+-------\----+  area_sum    = .3 + (skip)
   **   |       \  .1|.3      X   |              + .1 + .3               = .7
   **   |        \   |       /    |           
   **   |         \  |      /     |  area_sum_sq = (.3)^2 + (skip)
   **   |          \ |     /      |              + (.1)^2 + (.3)^2       = .19
   **   |   C       \|    /  D    |
   **   |            \   /        |  integration = (.3*10)^2 + (skip)
   **   |            |\ /         |              + (.1*20)^2 + (.3*30)^2 = 20
   **   |            | Y          |
   **   | 20         |         30 |
   **   +------------+------------+
   **
   **         W       
   **        /#*      Effectively, the area and integral in the 
   **       /##|`     above quadrilateral is calculated as if the
   **      /###| `    region containing the NaN wasn't there.
   **     Z####|  `   
   **      \###|   `  
   **       \##*----*  
   **        \#######X 
   **         \#####/  
   **          \###/   
   **           \#/    
   **            Y     
   **                  
   **   EXCEPT:  In the special case that all pixel values overlapped by the 
   **            quadrilateral are NaNs, the returns are set to:
   **
   **                area_sum = area_sum_sq = 0  and  integration2 = NaN 
   **
   */
   area_sum       = 0;
   area_sum_sq    = 0;
   integration1   = 0;
   integration2   = 0;
   pixel_count    = 0;
   pixels_skipped = 0;
   for( iy=iymin,iy_offset=iymin*ixdim; iy<=iymax; iy++,iy_offset+=ixdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         ixy = ix+iy_offset;
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon (pixmin, pixmax, &quad, &clipped_pixel);
         clipped_area = area_poly( &clipped_pixel );

         if (clipped_area > 0) {
            pixel_count++;
            if ( !iznand( image[ixy] ) ) {
               square        = clipped_area * clipped_area;
               area_sum     += clipped_area;
               area_sum_sq  += square;
               integration1 += clipped_area * image[ixy];
               integration2 += square * sigmas[ixy] * sigmas[ixy];
            } else {
               pixels_skipped++;
            }
         }
         
      }
   }
   
   /* If no "good data" within quad (ALL NaNs), return a NaN (area_sum==0 already) */
   if ( (pixels_skipped == pixel_count) && (pixels_skipped > 0) ) {
     integration1 = NANVALUE;
   }

   /*2******************************************************************
   ** If polygon is entirely within the pixel, then the area_sum should
   ** be equal to the area of the polygon computed directly (area_quad)
   ** Done for paranoid verification purposes only (may eliminate in final code)
   ** Use the word "CAUTION" to avoid the official "bad words"
   **
   ** COMMENTED OUT, BECAUSE NaNS DESTROY THIS CHECK.
   **
   if (!pt_out) { 
      area_quad = area_poly( &quad );  
      err_frac  = fabs((area_quad - area_sum)/(area_quad));
      if ( err_frac > TOL ) {
         printf("integrate_quad: CAUTION: Area error fraction, %g,", err_frac );
         printf(" exceeded TOL, %g  (Please notify Clare.)\n", TOL );                
         printf("in integrate_area.c: VERSION %1d\n\n", INTEGRATE_AREA_VERSION );      }
   }
   */

   /*2******************************************************************
   ** Return final answers
   */
   *area         = area_sum;
   *area_sq      = area_sum_sq;
   *sigma_result = sqrt( integration2 );
   return( integration1 );
   
}



/*****************************************************************************
** Integrate Quad "3":  Integrate the contents of an image within a 
** quadrilateral, AND the array of error values, except that instead of 
** summing the error values in pixels, the sums of squares of error values 
** and pixel overlap areas are used.
**
** Error values are assumed to be passed as sigma (not sigma squared)
**
** Made into an entirely separate routine from integrate_quad() for SPEED.
** Both the regular area AND the sum of squared area-weights (area_sq) 
** are provided.
**
** This routine also ORs together mask values, and returns mask_result
*/
double integrate_quad_3( double rx1,           double ry1,
                         double rx2,           double ry2,
                         double rx3,           double ry3,
                         double rx4,           double ry4,
                         double image[],       double sigmas[],
                         short int mask[],     short int mask_fatal, /*(not used)*/
                         int    ixdim,         int iydim,
                         double *area,         double *area_sq,
                         double *sigma_result, short int *mask_result )
                                          
{
int		ix,iy, iy_offset;  /* Pixel counters within the image */
int		ixy;		/* Holds ix + iy_offset                      */
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
double		clipped_area;	/* Area of the clipped overlap region within */
				/*    both the pixel and the quadrilateral   */
double		integration1;	/* Accumulates the desired integration value */
double		integration2;	/* Accumulates the integration of squares    */
double		area_sum;	/* Accumulates the desired overlap area      */
double		area_sum_sq;	/* Accumulates the desired overlap area^2    */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
poly_type	quad; 		/* Holds a closed quadrilateral              */
poly_type	clipped_pixel;  /* Holds the vertices of the clipped overlap */
double		area_quad;	/* Quadrilateral area, computed directly     */
double		err_frac;	/* Check accuracy of area_sum vs. area_quad  */
boolean		pt_out;		/* A point in the quad is outside the image  */
int		pixel_count;    /* Number of pixels processed within a quad  */
int		pixels_skipped; /* Number of NaN pixels skipped within quad  */
double		square;		/* Square of clipped_area                    */
short int       masking;        /* OR'd together mask values                 */

   /* REMEMBER: for any edit changes, change ALL of:                      */
   /*    integrate_quad() AND integrate_quad_2() AND integrate_quad_3()   */
   /*    AND integrate_two_quad()                                         */


   /*3**********************************************************************
   ** Find the bounding rectangle for the quadrilateral.
   ** Each pixel [i,j] is defined from [i,j] <= pt < [i+1,j+1], so
   ** Define the bounding rectangle such that
   **     for a minimum: [i <= rmin <  i+1] maps to i   so rmin -> floor(rmin)
   ** and for a maximum: [i <  rmax <= i+1] maps to i   so rmax ->  ceil(rmax)-1
   */
   ixmin = rmin4floor( rx1, rx2, rx3, rx4 );
   iymin = rmin4floor( ry1, ry2, ry3, ry4 );
   ixmax = rmax4ceil(  rx1, rx2, rx3, rx4 )-1;
   iymax = rmax4ceil(  ry1, ry2, ry3, ry4 )-1;
   
   pt_out = FALSE;
   
   if ( ixmin <  0     ) { pt_out  =  TRUE;  ixmin = 0;       }
   if ( iymin <  0     ) { pt_out  =  TRUE;  iymin = 0;       }
   if ( ixmax >= ixdim ) { pt_out  =  TRUE;  ixmax = ixdim-1; }
   if ( iymax >= iydim ) { pt_out  =  TRUE;  iymax = iydim-1; }
   
   quad = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );


   /*3**********************************************************************
   ** MAIN PROCESSING LOOP
   ** For each pixel in the bounding rectangle: clip the quadrilateral to
   ** the pixel boundaries, calculate its area, add to sums for total area,
   ** total squared overlap areas and integrated squared values.
   **
   ** Added support for NaNs:
   **
   **
   **   +------------+------------+
   **   | 10         |        NaN |  EXAMPLE:
   **   |           W|            |  Quadrilateral WXYZ overlaps pixels
   **   |          / \            |  A, B, C, and D, with
   **   |         /  |\           |  corresponding overlap areas
   **   |   A    /   | \     B    |  .3, .2, .1, and .3 .  Pixels   
   **   |       /    |  \         |  A, B, C, and D, have image values
   **   |      /     |   \        |  10, NaN, 20, and 30, respectively.
   **   |     /      |    \       |
   **   |    Z       |     \      |  integrate_quad_2 calculates:      
   **   |     \    .3|.2    \     |
   **   +------\-----+-------\----+  area_sum    = .3 + (skip)
   **   |       \  .1|.3      X   |              + .1 + .3               = .7
   **   |        \   |       /    |           
   **   |         \  |      /     |  area_sum_sq = (.3)^2 + (skip)
   **   |          \ |     /      |              + (.1)^2 + (.3)^2       = .19
   **   |   C       \|    /  D    |
   **   |            \   /        |  integration = (.3*10)^2 + (skip)
   **   |            |\ /         |              + (.1*20)^2 + (.3*30)^2 = 20
   **   |            | Y          |
   **   | 20         |         30 |
   **   +------------+------------+
   **
   **         W       
   **        /#*      Effectively, the area and integral in the 
   **       /##|`     above quadrilateral is calculated as if the
   **      /###| `    region containing the NaN wasn't there.
   **     Z####|  `   
   **      \###|   `  
   **       \##*----*  
   **        \#######X 
   **         \#####/  
   **          \###/   
   **           \#/    
   **            Y     
   **                  
   **   EXCEPT:  In the special case that all pixel values overlapped by the 
   **            quadrilateral are NaNs, the returns are set to:
   **
   **                area_sum = area_sum_sq = 0  and  integration2 = NaN 
   **
   */
   area_sum       = 0;
   area_sum_sq    = 0;
   integration1   = 0;
   integration2   = 0;
   pixel_count    = 0;
   pixels_skipped = 0;
   masking        = 0;
   for( iy=iymin,iy_offset=iymin*ixdim; iy<=iymax; iy++,iy_offset+=ixdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         ixy = ix+iy_offset;
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon (pixmin, pixmax, &quad, &clipped_pixel);
         clipped_area = area_poly( &clipped_pixel );

         if (clipped_area > 0) {
            pixel_count++;
            masking |= mask[ixy];
            if ( !iznand( image[ixy] ) ) {
               square        = clipped_area * clipped_area;
               area_sum     += clipped_area;
               area_sum_sq  += square;
               integration1 += clipped_area * image[ixy];
               integration2 += square * sigmas[ixy] * sigmas[ixy];
            } else {
               pixels_skipped++;
            }
         }
         
      }
   }
   
   /* If no "good data" within quad (ALL NaNs), return a NaN (area_sum==0 already) */
   if ( (pixels_skipped == pixel_count) && (pixels_skipped > 0) ) {
     integration1 = NANVALUE;
   }

   /*3******************************************************************
   ** If polygon is entirely within the pixel, then the area_sum should
   ** be equal to the area of the polygon computed directly (area_quad)
   ** Done for paranoid verification purposes only (may eliminate in final code)
   ** Use the word "CAUTION" to avoid the official "bad words"
   **
   ** COMMENTED OUT, BECAUSE NaNS DESTROY THIS CHECK.
   **
   if (!pt_out) { 
      area_quad = area_poly( &quad );  
      err_frac  = fabs((area_quad - area_sum)/(area_quad));
      if ( err_frac > TOL ) {
         printf("integrate_quad: CAUTION: Area error fraction, %g,", err_frac );
         printf(" exceeded TOL, %g  (Please notify Clare.)\n", TOL );                
         printf("in integrate_area.c: VERSION %1d\n\n", INTEGRATE_AREA_VERSION );      }
   }
   */

   /*3******************************************************************
   ** Return final answers
   */
   *area         = area_sum;
   *area_sq      = area_sum_sq;
   *sigma_result = sqrt( integration2 );
   *mask_result  = masking;
   return( integration1 );
   
}



/*****************************************************************************
** Distribute Quad:  Distribe the contents of a quadrilateral appropriately
** to the pixels it overlaps.  
**
** THIS ROUTINE IS IDENTICAL TO distribute_quad_BB() except that it lacks the
** extra FF-specific processing to populate mask_image[] and order_image[].
**
** See illustrated example below in "MAIN PROCESSING LOOP"
**  
** INPUTS:
**	rx1,ry1, rx2,ry2, 	Cyclical points around the quadrilateral
**      rx3,ry3, rx4,ry4	(marked with "r" to remind that they're real)
**      image[]			The image to receive values ADDED IN to it
**                              (USER IS RESPONSIBLE FOR SETTING image[]==0
**                              IF IT IS DESIRED BEFORE PROCESSING)
**      pseudopixel_value       Pseudopixel (quadrilateral) value to be distributed
**	xdim,ydim		Dimensions of the image
**
** OUTPUTS:
**	image[] 		Appropriate values have been added.
**
** DEFINE:
**	Pixels are defined as in integrate_quad()
**	REQUIRES input quadrilateral to be CONVEX, defined cyclically.
**
**	Parts of the Quadrilateral that lie outside the boundary of the image
**      will be ignored.
**
**      Use the equivalence: 
**	   for an image im[] with dimensions (xdim,ydim), use
**	   im[x,y] == (y * xdim) + x
**
** PSEUDOCODE:
**
**      * Store the quadrilateral into pp (a poly_type)
**      * Calculate the bounding_rect() for that pp -- i.e. the smallest
**        rectangle within the image that entirely contains the pixels that 
**        lie within or overlap the pp.  These are the only pixels that need
**        to be examined for possible addition of full or partial pp_value.
**      * For each pixel in the bounding rectangle (pseduopixel):
**        * Clip the pseudopixel to the pixel boundaries.
**        * Add the appropriate fraction (the fractional area of the 
**          clipped_pixel) of the pseudopixel value to the FITS image pixel.
**      * End Loop
*/
void distribute_quad( double rx1, double ry1,   /* The four vertices of the */
                      double rx2, double ry2,   /*    quadrilateral to be   */
                      double rx3, double ry3,   /*    distributed           */
                      double rx4, double ry4,  
                      double pseudopixel_value, /* Value to be distributed  */
                      double main_image[],	/* Image to receive added   */
                                   		/*    distributed values    */
                      int xdim,   int ydim  )   /* Dimensions of the image  */
{
int		j;
poly_type	pp; 		/* PseudoPixel:   a closed quadrilateral     */
poly_type	clipped_pixel;
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
int		ix,iy, iy_offset;  /* Pixel counters within the pseudopixel  */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
double		clipped_area;	/* Area of clipped pixel within the pp       */

   /* REMEMBER: for any edit changes, change ALL of:                         */
   /*   distribute_quad() AND distribute_quad_BB() AND distribute_two_quad() */

   pp = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );
                    
   bounding_rect( pp, &ixmin, &ixmax, &iymin, &iymax, xdim, ydim );
            
   /***********************************************************************
   ** For each pixel in the bounding rectangle (pseduopixel): clip the 
   ** pseudopixel to the pixel boundaries.  Add the appropriate fraction
   ** (the fractional area of the clipped_pixel) of the pseudopixel value
   ** to the FITS image pixel.  
   */
   for( iy=iymin,iy_offset=iymin*xdim; iy<=iymax; iy++,iy_offset+=xdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon( pixmin, pixmax, &pp, &clipped_pixel );
         clipped_area = area_poly( &clipped_pixel );
         if (clipped_area != 0.) {
            j = ix+iy_offset;
            main_image[j] += clipped_area * pseudopixel_value;
         }
      }
   }

}



/*2***************************************************************************
** Distribute Quad 2:  Distribe the contents of a quadrilateral appropriately
** to the pixels it overlaps, and distribute the sigmas in a similar manner.  
**
** THIS ROUTINE IS IDENTICAL TO distribute_quad_BB() except that it lacks the
** extra FF-specific processing to populate mask_image[] and order_image[].
**
** See illustrated example below in "MAIN PROCESSING LOOP"
**  
** INPUTS:
**	rx1,ry1, rx2,ry2, 	Cyclical points around the quadrilateral
**      rx3,ry3, rx4,ry4	(marked with "r" to remind that they're real)
**      image[]			The image to receive values ADDED IN to it
**                              (USER IS RESPONSIBLE FOR SETTING image[]==0
**                              IF IT IS DESIRED BEFORE PROCESSING)
**      pseudopixel_value       Pseudopixel (quadrilateral) value to be distributed
**	xdim,ydim		Dimensions of the image
**
** OUTPUTS:
**	image[] 		Appropriate values have been added.
**
** DEFINE:
**	Pixels are defined as in integrate_quad()
**	REQUIRES input quadrilateral to be CONVEX, defined cyclically.
**
**	Parts of the Quadrilateral that lie outside the boundary of the image
**      will be ignored.
**
**      Use the equivalence: 
**	   for an image im[] with dimensions (xdim,ydim), use
**	   im[x,y] == (y * xdim) + x
**
** PSEUDOCODE:
**
**      * Store the quadrilateral into pp (a poly_type)
**      * Calculate the bounding_rect() for that pp -- i.e. the smallest
**        rectangle within the image that entirely contains the pixels that 
**        lie within or overlap the pp.  These are the only pixels that need
**        to be examined for possible addition of full or partial pp_value.
**      * For each pixel in the bounding rectangle (pseduopixel):
**        * Clip the pseudopixel to the pixel boundaries.
**        * Add the appropriate fraction (the fractional area of the 
**          clipped_pixel) of the pseudopixel value to the FITS image pixel.
**      * End Loop
*/
void distribute_quad_2( double rx1, double ry1,   /* The four vertices of the */
                        double rx2, double ry2,   /*    quadrilateral to be   */
                        double rx3, double ry3,   /*    distributed           */
                        double rx4, double ry4,  
                        double pseudopixel_value, /* Value to be distributed  */
                        double pseudopixel_sigma, /* Sigma to be distributed  */
                        double main_image[],      /* Image to receive values  */
                        double sigmas[],          /* Corresponding sigmas     */
                        int xdim,   int ydim  )   /* Dimensions of the image  */
{
int		j;
poly_type	pp; 		/* PseudoPixel:   a closed quadrilateral     */
poly_type	clipped_pixel;
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
int		ix,iy, iy_offset;  /* Pixel counters within the pseudopixel  */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
double		clipped_area;	/* Area of clipped pixel within the pp       */

   /* REMEMBER: for any edit changes, change ALL of:                         */
   /*   distribute_quad() AND distribute_quad_BB() AND distribute_two_quad() */

   pp = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );
                    
   bounding_rect( pp, &ixmin, &ixmax, &iymin, &iymax, xdim, ydim );
            
   /*2*********************************************************************
   ** For each pixel in the bounding rectangle (pseduopixel): clip the 
   ** pseudopixel to the pixel boundaries.  Add the appropriate fraction
   ** (the fractional area of the clipped_pixel) of the pseudopixel value
   ** to the FITS image pixel.  
   */
   for( iy=iymin,iy_offset=iymin*xdim; iy<=iymax; iy++,iy_offset+=xdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon( pixmin, pixmax, &pp, &clipped_pixel );
         clipped_area = area_poly( &clipped_pixel );
         if (clipped_area != 0.) {
            j = ix+iy_offset;
            main_image[j] += clipped_area * pseudopixel_value;
            sigmas[j]     += clipped_area * pseudopixel_sigma;
         }
      }
   }

}



/*N***************************************************************************
** Distribute Quad N:  Distribe the contents of a quadrilateral appropriately
** to the pixels it overlaps, and distribute the sigmas in a similar manner.  
**
** THIS ROUTINE IS IDENTICAL TO distribute_quad_2() except that takes a 
** variable number of pixel values passed in the array *pseudopixel_value and
** distributes each value to the corresponding image from the array **image:
**     
**      pseudopixel_value[i] -> image[i];
**
** The number of images is specified by the parameter n_images.  The sigmas
** in distribute_quad_2() aren't needed here, since the calling program will
** decide whether it is processing data values and/or sigmas, and the
** data values and sigmas are processed identically.
**
** David Makovoz requested this function in CR4270, and expects this function
** to speed up mosaic interpolation in the drizzle mode by a factor of 2.
**
** The array of images used in this code was created by code SIMILAR to the following:
**
**   image_array = (double **) malloc( (unsigned) MAX_IMAGES * sizeof(double *) );
**   for (ii=0; ii<MAX_IMAGES; ii++) {
**      image_array[ii] = (double *) malloc( (unsigned) xydim * sizeof(double) );
**   }
*/
void distribute_quad_n( double rx1, double ry1,    /* The four vertices of the */
                        double rx2, double ry2,    /*    quadrilateral to be   */
                        double rx3, double ry3,    /*    distributed           */
                        double rx4, double ry4,  
                        double *pseudopixel_values,/* Values to be distributed */
                        double **main_images,      /* Images to receive values */
                        int    n_images,           /* Number of images         */
                        int    xdim, int ydim  )   /* Dimensions of the image  */
{
int		j;              /* 1D index within a single image (2D)         */
int             ii;             /* Image index: which image is being processed.*/
poly_type	pp; 		/* PseudoPixel:   a closed quadrilateral       */
poly_type	clipped_pixel;
int		ixmin,ixmax;	/* Set to the bounding rectangle containing    */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	       */
int		ix,iy, iy_offset;  /* Pixel counters within the pseudopixel    */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel         */
double		clipped_area;	/* Area of clipped pixel within the pp         */

   /* REMEMBER: for any edit changes, change ALL of:                           */
   /*   distribute_quad() AND distribute_quad_BB() AND distribute_two_quad()   */

   pp = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );
                    
   bounding_rect( pp, &ixmin, &ixmax, &iymin, &iymax, xdim, ydim );
   
   /* Process n_images, as requested */
   for( ii=0; ii<n_images; ii++ ) {

      /*N*********************************************************************
      ** For each pixel in the bounding rectangle (pseduopixel): clip the 
      ** pseudopixel to the pixel boundaries.  Add the appropriate fraction
      ** (the fractional area of the clipped_pixel) of the pseudopixel value
      ** to the FITS image pixel.  
      */
      for( iy=iymin,iy_offset=iymin*xdim; iy<=iymax; iy++,iy_offset+=xdim  ) {
         for( ix=ixmin; ix<=ixmax; ix++ ) {
            pixmin.x = ix;   pixmin.y = iy;
            pixmax.x = ix+1; pixmax.y = iy+1;
            clip_polygon( pixmin, pixmax, &pp, &clipped_pixel );
            clipped_area = area_poly( &clipped_pixel );
            if (clipped_area != 0.) {
               j = ix+iy_offset;
               main_images[ii][j] += clipped_area * pseudopixel_values[ii];
            }
         }
      }

   }

}



/*N***************************************************************************
** Distribute Quad N:  Distribe the contents of a quadrilateral appropriately
** to the pixels it overlaps, and distribute the sigmas in a similar manner.
**
** THIS IS THE 1D version, assuming the big array of images was to be indexed
** as a 1D array.  THIS VERSION HAS BEEN REPLACED BY distribute_quad_n(), which
** addresses an array of (1D-indexed) images.  I've saved this code here, just
** for the hell of it.
**
** THIS ROUTINE IS IDENTICAL TO distribute_quad_2() except that takes a 
** variable number of pixel values passed in the array *pseudopixel_value and
** distributes each value to the corresponding image from the array **image:
**     
**      pseudopixel_value[i] -> image[i];
**
** The number of images is specified by the parameter n_images.  The sigmas
** in distribute_quad_2() aren't needed here, since the calling program will
** decide whether it is processing data values and/or sigmas, and the
** data values and sigmas are processed identically.
**
** David Makovoz requested this function in CR4270, and expects this function
** to speed up mosaic interpolation in the drizzle mode by a factor of 2.
*/
void distribute_quad_n_1D( double rx1, double ry1,    /* The four vertices of the */
                           double rx2, double ry2,    /*    quadrilateral to be   */
                           double rx3, double ry3,    /*    distributed           */
                           double rx4, double ry4,  
                           double *pseudopixel_values,/* Values to be distributed */
                           double *main_images,       /* Images to receive values */
                           int    n_images,           /* Number of images         */
                           int    xdim, int ydim  )   /* Dimensions of the image  */
{
int		j;              /* 1D index within a single image (2D)         */
int             ii_j;           /* 1D index to full image array (3D)           */
int             ii;             /* Image index: which image is being processed.*/
poly_type	pp; 		/* PseudoPixel:   a closed quadrilateral       */
poly_type	clipped_pixel;
int		ixmin,ixmax;	/* Set to the bounding rectangle containing    */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	       */
int		ix,iy, iy_offset;  /* Pixel counters within the pseudopixel    */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel         */
double		clipped_area;	/* Area of clipped pixel within the pp         */
int             xydim;          /* xdim * ydim */

   /* 3D INDEXING SCHEME:
   ** main_images[ii_j] == main_images[ii][j] 
   **                   == main_images[ii][ix][iy]
   **                   == main_images[ (ii*xydim) + j ] 
   **                      where j = (iymin*xdim)+ix 
   */
   xydim = xdim * ydim;
        
   /* REMEMBER: for any edit changes, change ALL of:                           */
   /*   distribute_quad() AND distribute_quad_BB() AND distribute_two_quad()   */

   pp = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );
                    
   bounding_rect( pp, &ixmin, &ixmax, &iymin, &iymax, xdim, ydim );
   
   /* Process n_images, as requested */
   for( ii=0; ii<n_images; ii++ ) {

      /*N*********************************************************************
      ** For each pixel in the bounding rectangle (pseduopixel): clip the 
      ** pseudopixel to the pixel boundaries.  Add the appropriate fraction
      ** (the fractional area of the clipped_pixel) of the pseudopixel value
      ** to the FITS image pixel.  
      */
      for( iy=iymin,iy_offset=iymin*xdim; iy<=iymax; iy++,iy_offset+=xdim  ) {
         for( ix=ixmin; ix<=ixmax; ix++ ) {
            pixmin.x = ix;   pixmin.y = iy;
            pixmax.x = ix+1; pixmax.y = iy+1;
            clip_polygon( pixmin, pixmax, &pp, &clipped_pixel );
            clipped_area = area_poly( &clipped_pixel );
            if (clipped_area != 0.) {
               j = ix+iy_offset;
               ii_j = (ii*xydim) + j;
               main_images[ii_j] += clipped_area * pseudopixel_values[ii];
            }
         }
      }

   }

}


/*BB**************************************************************************
** Distribute Quad (BB):  Distribe the contents of a quadrilateral 
** appropriately to the pixels it overlaps, and distribute the sigmas in a 
** similar manner.
**
** THIS ROUTINE IS IDENTICAL TO distribute_quad_2() except for the addition of
** extra processing to populate a mask_image[] and order_image[].
*/
void distribute_quad_BB( double rx1, double ry1,   /* The four vertices of the */
                         double rx2, double ry2,   /*    quadrilateral to be   */
                         double rx3, double ry3,   /*    distributed           */
                         double rx4, double ry4,  
                         double pseudopixel_value, /* Value to be distributed  */
                         double pseudopixel_sigma, /* Sigma to be distributed  */
                         double main_image[],      /* Image to receive values  */
                         double sigmas[],          /* Corresponding sigmas     */
                         int xdim,   int ydim,     /* Dimensions of the image  */
                         double mask_image[],
                         double order_image[],
                         int order_number        ) 
{
int		j;
poly_type	pp; 		/* PseudoPixel:   a closed quadrilateral     */
poly_type	clipped_pixel;
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
int		ix,iy, iy_offset;  /* Pixel counters within the pseudopixel  */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
double		clipped_area;	/* Area of clipped pixel within the pp       */


   /* REMEMBER: for any edit changes,                           */
   /*    change BOTH distribute_quad() AND distribute_quad_BB() */

   pp = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );
                    
   bounding_rect( pp, &ixmin, &ixmax, &iymin, &iymax, xdim, ydim );
            
   /*BB********************************************************************
   ** For each pixel in the bounding rectangle (pseduopixel): clip the 
   ** pseudopixel to the pixel boundaries.  Add the appropriate fraction
   ** (the fractional area of the clipped_pixel) of the pseudopixel value
   ** to the FITS image pixel.  Store the order number in the order_image.
   */
   for( iy=iymin,iy_offset=iymin*xdim; iy<=iymax; iy++,iy_offset+=xdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon( pixmin, pixmax, &pp, &clipped_pixel );
         clipped_area = area_poly( &clipped_pixel );
         if (clipped_area != 0.) {
            j = ix+iy_offset;
            main_image[j] += clipped_area * pseudopixel_value;
            sigmas[j]     += clipped_area * pseudopixel_sigma;
            mask_image[j] += clipped_area;
            order_image[j] = order_number;
         }
      }
   }

}


/*
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@@  THREE NEW ROUTINES for David Makovoz:
@@     integrate_two_quad()         = CR3030
@@     distribute_two_quad()        = CR3031
@@     unweighted_distribute_quad() = CR3032
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*/

/*****************************************************************************
** Integrate Quad:  Integrate the contents of an image within a quadrilateral.
** See illustrated example below in "MAIN PROCESSING LOOP"
**  
** INPUTS:
**	rx1,ry1, rx2,ry2, 	Cyclical points around the quadrilateral
**      rx3,ry3, rx4,ry4	(marked with "r" to remind that they're real)
**      image1[], image2[]	The images to be integrated
**	ixdim,iydim		Dimensions of the image
**
** OUTPUTS:
**	area1			The area of the quadrilateral, in pixels.
**	area2			Only the area within the image dimensions
**				0..xdim and 0..ydim will be returned.
**				All parts of the quad outside the image dims
**				are defined to have zero area.
**				These two values may be different, since areas
**				in the image containing NaNs don't contribute.
**	*integrated_value1	(integration) The integrated contents of the 
**	*integrated_value2	quadrilateral (within the image dims).
**
** [for the rest of the comments, see integrate_quad()]
*/
void integrate_two_quad(  double rx1, double ry1,  /* The four vertices of the */
                          double rx2, double ry2,  /*    quadrilateral to be   */
                          double rx3, double ry3,  /*    integrated within     */
                          double rx4, double ry4,
                          double image1[],         /* Images to be integrated  */
                          double image2[], 
                          double *integrated_value1, /* Returned int'd values  */
                          double *integrated_value2,
                          int ixdim, int iydim,    /* Dimensions of the image  */
                          double *area1,           /* Returned areas of the    */
                          double *area2          ) /*    quadrilateral within  */
                                                   /*    the bounds of the     */
                                                   /*    image.                */
                                                   

{
int		ix,iy, iy_offset;  /* Pixel counters within the image */
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
double		clipped_area;	/* Area of the clipped overlap region within */
				/*    both the pixel and the quadrilateral   */
double		integration;	/* Accumulates the desired integration value */
double		area_sum;	/* Accumulates the desired overlap area      */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
poly_type	quad; 		/* Holds a closed quadrilateral              */
poly_type	clipped_pixel;  /* Holds the vertices of the clipped overlap */
double		area_quad;	/* Quadrilateral area, computed directly     */
double		err_frac;	/* Check accuracy of area_sum vs. area_quad  */
boolean		pt_out;		/* A point in the quad is outside the image  */
int		pixel_count;    /* Number of pixels processed within a quad  */
int		pixels_skipped1;/* Number of NaN pixels skipped within quad  */
int		pixels_skipped2;

   /* REMEMBER: for any edit changes, change ALL of:                      */
   /*    integrate_quad() AND integrate_quad_2() AND integrate_two_quad() */
   

   /************************************************************************
   ** Find the bounding rectangle for the quadrilateral.
   ** Each pixel [i,j] is defined from [i,j] <= pt < [i+1,j+1], so
   ** Define the bounding rectangle such that
   **     for a minimum: [i <= rmin <  i+1] maps to i   so rmin -> floor(rmin)
   ** and for a maximum: [i <  rmax <= i+1] maps to i   so rmax ->  ceil(rmax)-1
   */
   ixmin = rmin4floor( rx1, rx2, rx3, rx4 );
   iymin = rmin4floor( ry1, ry2, ry3, ry4 );
   ixmax = rmax4ceil(  rx1, rx2, rx3, rx4 )-1;
   iymax = rmax4ceil(  ry1, ry2, ry3, ry4 )-1;

   pt_out = FALSE;

   if ( ixmin <  0     ) { pt_out  =  TRUE;  ixmin = 0;       }
   if ( iymin <  0     ) { pt_out  =  TRUE;  iymin = 0;       }
   if ( ixmax >= ixdim ) { pt_out  =  TRUE;  ixmax = ixdim-1; }
   if ( iymax >= iydim ) { pt_out  =  TRUE;  iymax = iydim-1; }
   
   quad = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );


   /************************************************************************
   ** MAIN PROCESSING LOOP -- see main routine integrate_quad() for details
   */
   *area1             = 0;   *area2             = 0;  /* Was area_sum */
   *integrated_value1 = 0;   *integrated_value2 = 0;  /* Was integration */
   pixels_skipped1    = 0;   pixels_skipped2    = 0;
   pixel_count        = 0;
   for( iy=iymin,iy_offset=iymin*ixdim; iy<=iymax; iy++,iy_offset+=ixdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon (pixmin, pixmax, &quad, &clipped_pixel);
         clipped_area = area_poly( &clipped_pixel );
         if (clipped_area > 0) {
            pixel_count++;
            if ( !iznand( image1[ix+iy_offset] ) ) {
               *area1             += clipped_area;
               *integrated_value1 += clipped_area * image1[ix+iy_offset];
            } else {
               pixels_skipped1++;
            }
            if ( !iznand( image2[ix+iy_offset] ) ) {
               *area2             += clipped_area;
               *integrated_value2 += clipped_area * image2[ix+iy_offset];
            } else {
               pixels_skipped2++;
            }
         }
      }
   }
   
   /* If no "good data" within quad (ALL NaNs), return a NaN (*area1==0 already) */
   if ( (pixels_skipped1 == pixel_count) && (pixels_skipped1 > 0) ) {
      *integrated_value1 = atof("NaN");
   }
   /* If no "good data" within quad (ALL NaNs), return a NaN (*area2==0 already) */
   if ( (pixels_skipped2 == pixel_count) && (pixels_skipped2 > 0) ) {
      *integrated_value2 = atof("NaN");
   }

   /********************************************************************
   ** Return final answers -- already stored in appropriate locations:
   **   *area1, *area2, *integrated_value1, *integrated_value2
   */
   
}
                          
                          
                          
/*****************************************************************************
** Distribute Quad:  Distribe the contents of a quadrilateral appropriately
** to the pixels it overlaps.  SEE ORIGINAL ROUTINE FOR MORE DETAILS
**
** INPUTS:
**	rx1,ry1, rx2,ry2, 	Cyclical points around the quadrilateral
**      rx3,ry3, rx4,ry4	(marked with "r" to remind that they're real)
**      image1[], image2[],	The images to receive values ADDED IN to it
**                              (USER IS RESPONSIBLE FOR SETTING image[]==0
**                              IF IT IS DESIRED BEFORE PROCESSING)
**      pseudopixel_value1      Pseudopixel (quadrilateral) values to be 
**      pseudopixel_value2        distributed to the corresponding images
**	xdim,ydim		Dimensions of both identically-sized images
**
** OUTPUTS:
**	image1[], image2[] 	Appropriate values have been added to each.
**
** SEE ORIGINAL ROUTINE distribute_quad() FOR MORE DETAILS
*/
void distribute_two_quad( double rx1, double ry1, /* The four vertices of the */
                          double rx2, double ry2, /*    quadrilateral to be   */
                          double rx3, double ry3, /*    distributed           */
                          double rx4, double ry4,
                          double pseudopixel_value1, /* Value1 to be dist'd     */
                          double image1[],           /* Image to receive value1 */
                          double pseudopixel_value2, /* Value2 to be dist'd     */
                          double image2[],           /* Image to receive value2 */
                          int xdim, int ydim )    /* Dimensions of both images  */
{
int		j;
poly_type	pp; 		/* PseudoPixel:   a closed quadrilateral     */
poly_type	clipped_pixel;
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
int		ix,iy, iy_offset;  /* Pixel counters within the pseudopixel  */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
double		clipped_area;	/* Area of clipped pixel within the pp       */

   /* REMEMBER: for any edit changes, change ALL of:                         */
   /*   distribute_quad() AND distribute_quad_BB() AND distribute_two_quad() */

   pp = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );
                    
   bounding_rect( pp, &ixmin, &ixmax, &iymin, &iymax, xdim, ydim );
            
   /***********************************************************************
   ** For each pixel in the bounding rectangle (pseduopixel): clip the 
   ** pseudopixel to the pixel boundaries.  Add the appropriate fraction
   ** (the fractional area of the clipped_pixel) of the pseudopixel value
   ** to the FITS image pixel.  
   */
   for( iy=iymin,iy_offset=iymin*xdim; iy<=iymax; iy++,iy_offset+=xdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon( pixmin, pixmax, &pp, &clipped_pixel );
         clipped_area = area_poly( &clipped_pixel );
         if (clipped_area != 0.) {
            j = ix+iy_offset;
            image1[j] += clipped_area * pseudopixel_value1;
            image2[j] += clipped_area * pseudopixel_value2;
         }
      }
   }

}


/****************************************************************************
** UNWEIGHTED Distribute Quad:  Distribe the contents of a quadrilateral 
** in an unweighted manner to the pixels it overlaps.  
**
** UNWEIGHTED means that if a the quadrilateral overlaps a pixel at all,
** the FULL VALUE of that pseudopixel_value is added, rather than weighting
** it by any pixel overlap area.
**
** SEE ORIGINAL ROUTINE FOR REMAINING DETAILS
**
** PSEUDOCODE:
**
**      * Store the quadrilateral into pp (a poly_type)
**      * Calculate the bounding_rect() for that pp -- i.e. the smallest
**        rectangle within the image that entirely contains the pixels that 
**        lie within or overlap the pp.  These are the only pixels that need
**        to be examined for possible addition of full or partial pp_value.
**      * For each pixel in the bounding rectangle (pseduopixel):
**        * Clip the pseudopixel to the pixel boundaries.
**        * If the clipped area is >0 then store the pseudopixel value 
**          in the FITS image pixel (don't add it, replace the value).
**      * End Loop
*/
void unweighted_distribute_quad(
                          double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          double pseudopixel_value,  double image[],
                          int xdim, int ydim                                 )
{
int		j;
poly_type	pp; 		/* PseudoPixel:   a closed quadrilateral     */
poly_type	clipped_pixel;
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
int		ix,iy, iy_offset;  /* Pixel counters within the pseudopixel  */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
double		clipped_area;	/* Area of clipped pixel within the pp       */

   /* REMEMBER: for any edit changes, change ALL of:                         */
   /*   distribute_quad() AND distribute_quad_BB() AND distribute_two_quad() */

   pp = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );
                    
   bounding_rect( pp, &ixmin, &ixmax, &iymin, &iymax, xdim, ydim );
            
   /***********************************************************************
   ** For each pixel in the bounding rectangle (pseduopixel): clip the 
   ** pseudopixel to the pixel boundaries.  If there is nonzero overlap
   ** with the pixel, then add the pseudopixel value to the FITS image pixel.  
   */
   for( iy=iymin,iy_offset=iymin*xdim; iy<=iymax; iy++,iy_offset+=xdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon( pixmin, pixmax, &pp, &clipped_pixel );
         clipped_area = area_poly( &clipped_pixel );
         if (clipped_area != 0.) {
            j = ix+iy_offset;
            image[j]  = pseudopixel_value;
         }
      }
   }

}

/****************************************************************************
** OR'd Mask Distribute Quad (SHORT):  Distribe the contents of a quadrilateral 
** in by or'ing together the pixels it overlaps.  Ignore the amount of the
** overlap area.
**
** SEE ORIGINAL ROUTINE FOR REMAINING DETAILS
*/
void mask_distribute_quad_short(
                          double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          short pseudopixel_value,   short image[],
                          int xdim, int ydim                                 )
{
int		j;
poly_type	pp; 		/* PseudoPixel:   a closed quadrilateral     */
poly_type	clipped_pixel;
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
int		ix,iy, iy_offset;  /* Pixel counters within the pseudopixel  */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
double		clipped_area;	/* Area of clipped pixel within the pp       */

   /* REMEMBER: for any edit changes, change ALL of:                         */
   /*   distribute_quad() AND distribute_quad_BB() AND distribute_two_quad() */

   pp = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );
                    
   bounding_rect( pp, &ixmin, &ixmax, &iymin, &iymax, xdim, ydim );
            
   /***********************************************************************
   ** For each pixel in the bounding rectangle (pseduopixel): clip the 
   ** pseudopixel to the pixel boundaries.  If there is nonzero overlap
   ** with the pixel, then add the pseudopixel value to the FITS image pixel.  
   */
   for( iy=iymin,iy_offset=iymin*xdim; iy<=iymax; iy++,iy_offset+=xdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon( pixmin, pixmax, &pp, &clipped_pixel );
         clipped_area = area_poly( &clipped_pixel );
         if (clipped_area != 0.) {
            j = ix+iy_offset;
            image[j] |= pseudopixel_value;
         }
      }
   }

}


/****************************************************************************
** OR'd Mask Distribute Quad (INT):  Distribe the contents of a quadrilateral 
** in by or'ing together the pixels it overlaps.  Ignore the amount of the
** overlap area.
**
** SEE ORIGINAL ROUTINE FOR REMAINING DETAILS
*/
void mask_distribute_quad_int(
                          double rx1, double ry1,    double rx2, double ry2,
                          double rx3, double ry3,    double rx4, double ry4,
                          int    pseudopixel_value,  int image[],
                          int    xdim, int ydim                                 )
{
int		j;
poly_type	pp; 		/* PseudoPixel:   a closed quadrilateral     */
poly_type	clipped_pixel;
int		ixmin,ixmax;	/* Set to the bounding rectangle containing  */
int		iymin,iymax;  	/*    the quadrilateral (int pixels)	     */
int		ix,iy, iy_offset;  /* Pixel counters within the pseudopixel  */
pixel_type	pixmin, pixmax;	/* Window set to the size of one pixel       */
double		clipped_area;	/* Area of clipped pixel within the pp       */

   /* REMEMBER: for any edit changes, change ALL of:                         */
   /*   distribute_quad() AND distribute_quad_BB() AND distribute_two_quad() */

   pp = stuff_four( rx1,ry1, rx2,ry2, rx3,ry3, rx4,ry4 );
                    
   bounding_rect( pp, &ixmin, &ixmax, &iymin, &iymax, xdim, ydim );
            
   /***********************************************************************
   ** For each pixel in the bounding rectangle (pseduopixel): clip the 
   ** pseudopixel to the pixel boundaries.  If there is nonzero overlap
   ** with the pixel, then add the pseudopixel value to the FITS image pixel.  
   */
   for( iy=iymin,iy_offset=iymin*xdim; iy<=iymax; iy++,iy_offset+=xdim  ) {
      for( ix=ixmin; ix<=ixmax; ix++ ) {
         pixmin.x = ix;   pixmin.y = iy;
         pixmax.x = ix+1; pixmax.y = iy+1;
         clip_polygon( pixmin, pixmax, &pp, &clipped_pixel );
         clipped_area = area_poly( &clipped_pixel );
         if (clipped_area != 0.) {
            j = ix+iy_offset;
            image[j] |= pseudopixel_value;
         }
      }
   }

}

