/*
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
@@
@@ PROGRAM:  poly_clip.c
@@
@@ PURPOSE:  Performs "Sutherland-Hodgeman Polygon Clipping"
@@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*/

/*******************************************************************************
**  Sutherland-Hodgeman Polygon Clipping
**  - clip a polygon against a rectangular window
**
**  taken from COMPUTER GRAPHICS, C VERSION, 2nd ed.
**  by Donald Hearn and M. Pauline Baker
**  Prentice Hall, 1997, Pages 239-242
**
**  Source code at: http://www.ncsa.uiuc.edu/Vis/Graphics/code.html
**  This code: http://www.ncsa.uiuc.edu/Vis/Graphics/src/clipSuthHodge.c
**  (modified from the original)
**
**  Modified AND DEBUGGED by Clare Waterson (original code was BUGGY!!!)
**  * Replaced: point *first[]
**        with: point first[] and boolean first_set[] 
**    these changes were propagated throughout the code from the book.
**
********************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>  /* Has prototype for "system()" and "exit()" */
#include <math.h>
#include "integrate.h"


/***** CODE adapted FROM BOOK STARTS HERE *****/

typedef enum { Left, Right, Bottom, Top } Edge;
#define N_EDGE 4

/*******************************************************************************
**  Is a point "inside" wrt an Edge
*/
int inside (point_type p, Edge b, pixel_type wMin, pixel_type wMax)
{
   switch (b) {
      case Left:   if (p.x < wMin.x) return (FALSE); break;
      case Right:  if (p.x > wMax.x) return (FALSE); break;
      case Bottom: if (p.y < wMin.y) return (FALSE); break;
      case Top:    if (p.y > wMax.y) return (FALSE); break;
   }
   return (TRUE);
}

/*******************************************************************************
**  Cross()
*/
int cross( point_type p1, point_type p2, Edge b, 
           pixel_type wMin, pixel_type wMax      )
{
   return(  !( inside(p1,b,wMin,wMax) == inside(p2,b,wMin,wMax) )  );
}

/*******************************************************************************
**  Intersect()
*/
point_type intersect( point_type p1, point_type p2, Edge b, 
                      pixel_type wMin, pixel_type wMax      )
{
point_type iPt;
double m;

   if (p1.x != p2.x) {
      m = (p1.y - p2.y) / (p1.x - p2.x);
   }
   switch (b) {
      case Left:
         iPt.x = wMin.x;
         iPt.y = p2.y + (wMin.x - p2.x) * m;
      break;
      case Right:
         iPt.x = wMax.x;
         iPt.y = p2.y + (wMax.x - p2.x) * m;
      break;
      case Bottom:
         iPt.y = wMin.y;
         if (p1.x != p2.x) iPt.x = p2.x + (wMin.y - p2.y) / m;
         else iPt.x = p2.x;
      break;
      case Top:
         iPt.y = wMax.y;
         if (p1.x != p2.x) iPt.x = p2.x + (wMax.y - p2.y) / m;
         else iPt.x = p2.x;
      break;
   }
   return (iPt); 
}


/*******************************************************************************
**  Clip Point (a recursive routine -- calls itself from two places)
*/
void clipPoint (point_type p, Edge b, pixel_type wMin, pixel_type wMax,
                point_type *pOut, int *cnt, 
                boolean *first_set, point_type *first, point_type *s)
{
point_type iPt;

   /* If no previous point exists for this edge, save this point. */
   if (!first_set[b]) {
     first[b]     = p;      /* WAS *first[b]=&p in the book!  VERY WRONG!!! */
     first_set[b] = TRUE;   /*     I replaced it with these two lines.      */
   }
   else {
      /* Previous point exists.  If 'p' and previous point cross edge,
         find intersection.  Clip against next boundary, if any.  If
         no more edges, add intersection to output list. */
      if (cross (p, s[b], b, wMin, wMax)) {
         iPt = intersect (p, s[b], b, wMin, wMax);
         if (b < Top) {
            clipPoint (iPt, b+1, wMin, wMax, pOut, cnt, first_set, first, s);
         }
         else {
            if ( *cnt < MAX_WORKING_PTS ) {
               pOut[*cnt] = iPt;  (*cnt)++;
            }
            else{
               printf("ERROR in clipPoint: *cnt=%1d Exceeded MAX_WORKING_PTS=%1d\n",
                       *cnt, MAX_WORKING_PTS);
               printf("ERROR values: p    = [%g,%g]\n", p.x, p.y);
               printf("ERROR values: wMin = [%4d,%4d]\n", wMin.x, wMin.y);
               printf("ERROR values: wMax = [%4d,%4d]\n", wMax.x, wMax.y);
               *cnt = 0;
               return;
            }
         }
      }  
   }

   s[b] = p;           /* Save 'p' as most recent point for this edge */

   /* For all, if point is 'inside' proceed to next clip edge, if any */
   if (inside (p, b, wMin, wMax)) {
      if (b < Top){
         clipPoint (p, b+1, wMin, wMax, pOut, cnt, first_set, first, s);
      }
      else {
         if ( *cnt < MAX_WORKING_PTS ) {
            pOut[*cnt] = p;  (*cnt)++;
         }
         else{
            printf("ERROR in clipPoint: *cnt=%1d Exceeded MAX_WORKING_PTS=%1d\n",
                    *cnt, MAX_WORKING_PTS);
            printf("ERROR values: p    = [%g,%g]\n", p.x, p.y);
            printf("ERROR values: wMin = [%4d,%4d]\n", wMin.x, wMin.y);
            printf("ERROR values: wMax = [%4d,%4d]\n", wMax.x, wMax.y);
            *cnt = 0;
            return;
         }
      }
   }
}


/*******************************************************************************
**  Close Clip
*/
void closeClip( pixel_type wMin, pixel_type wMax, point_type *pOut,
                int *cnt, boolean *first_set, 
                point_type *first, point_type *s)
{
point_type	iPt;
Edge		b;

   for (b = Left; b <= Top; b++) {
      if (cross (s[b], first[b], b, wMin, wMax)) {
         iPt = intersect (s[b], first[b], b, wMin, wMax);
         if (b < Top) {
            clipPoint (iPt, b+1, wMin, wMax, pOut, cnt, first_set, first, s);
         }
         else {
            if ( *cnt < MAX_WORKING_PTS ) {
               pOut[*cnt] = iPt;  (*cnt)++;
            }
            else {
               printf("ERROR in closeClip: *cnt=%1d Exceeded MAX_WORKING_PTS=%1d\n",
                       *cnt, MAX_WORKING_PTS);
               printf("ERROR values: iPt  = [%g,%g]\n",  iPt.x, iPt.y);
               printf("ERROR values: wMin = [%4d,%4d]\n", wMin.x, wMin.y);
               printf("ERROR values: wMax = [%4d,%4d]\n", wMax.x, wMax.y);
               *cnt = 0;
               return;
            }
         }
      }
   }
}


/*********************************************************************
** clip_polygon() is the MAIN POLYGON CLIPPING ROUTINE
**
*/
void clip_polygon( pixel_type wMin, pixel_type wMax, 
                   poly_type *polyIn,  poly_type *polyOut  )
{
boolean	    first_set[N_EDGE] = { FALSE, FALSE, FALSE, FALSE };
point_type  first[N_EDGE];
	       /* Holds pointer to first point processed against a clip edge. */
point_type  s[N_EDGE];  /* Holds most recent point processed against an edge. */
point_type  pOut[MAX_WORKING_PTS];
int i;
int nPts = 0;  /* Counts the number of clipped points */

   for (i=0; i<(polyIn->len); i++) {
      clipPoint( polyIn->p[i], Left, wMin, wMax, pOut, &nPts, first_set, first, s);
   }

   closeClip (wMin, wMax, pOut, &nPts, first_set, first, s);

   /* Polygon Bookkeeping: close pOut (repeat first vertex) and set length */
   if (nPts>0) {
      polyOut->p[0].x = pOut[0].x; 
      polyOut->p[0].y = pOut[0].y; 
      for( i=1; i<nPts; i++ ) {
         polyOut->p[i].x = pOut[i].x;
         polyOut->p[i].y = pOut[i].y;
      }
      polyOut->p[nPts].x = pOut[0].x;
      polyOut->p[nPts].y = pOut[0].y;
      polyOut->len       = nPts+1;
   }
   else polyOut->len = 0;
   
}
/***** CODE adapted FROM BOOK ENDS HERE *****/


