/****************************************************************************
**                                                                         **
** Routine Name:                isis_colstr.c                              **
**                                                                         **
** Description:                 This routine retrieves address of a column.**
**                                                                         **
** Input Parameters:            ifile   ____ input file number             **
**                              icol    ____ column number                 **
**                              *stat   ____ status                        **
**                                                                         **
** Output Parameters:           None                                       **
**                                                                         **
** Function Return Values:      stat > 0,               column pointer     **
**                              stat = TBL_ERROR  (-1), no such column     **
**                              stat = TBL_NOFILE (-3), no such file       **
**                              stat = TBL_NOTOPN (-6), file not open      **
**                                                                         **
** Routines Be Called:          None                                       **
**                                                                         **
** Revision History:                                                       **
**                                                                         **
**   19??-??    John Good       initial author				   **
**   1998-03    Angela Zhang    Add header and comments                    **
**   			        Modified code                              **
**   1998-11    John Good       Fixed bug with endcol                      **
**   1998-12    Angela Zhang    When column number is larger than 0        **
**                                  change                                 **
**                                     bgncol =                            ** 
**                                     tbl[ipac]->colinfo[icol-1]->endcol  **
**                                  to                                     ** 
**                                     bgncol =                            **
**                                     tbl[ipac]->colinfo[icol-1]->endcol+1**
**                                                                         **
**   1999-02    Angela Zhang    change endcol and bgncol to match IPAC     **
**                              conversion.
**                                                                         **
** Version:                     TBL0.0                                     **
**                                                                         **
**                                                                         **
** Infrared Science Information System                                     **
** California Institute of Technology                                      **
**                                                                         **
** Copyright (c) 1998, California Institute of Technology.                 **
** U. S. Government Sponsorship under NASA Contract                        **
** NAS7-???? is acknowledged.                                              **
**                                                                         **
*****************************************************************************/


/***********************
* define Include Files *
************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <tbl.h>


/**************************
* Define Global Variables *
***************************/
int      maxfiles;               /* maximun number of files are allocated */
long int key_bytes[TBL_MAXFILES];/* keyword bytes in a file */
long int isis_add[TBL_MAXFILES]; /* current row address */
char	 *tbl_ival;	         /* column content for ISIS file */
char	 fix[TBL_MAXFILES];      /* FIX ISIS file */


char *isis_colstr(ifile, icol, stat)
int 	ifile;			/* input file */	 
int 	icol;			/* column index  */	 
int	*stat;			/* status */

{
  /*************************
  * Define Local Variables *
  **************************/
  char	*rec;			/* current column record */
  char	*ptr;			/* column pointer */ 
  int	bgncol;			/* beginning column number of a record */
  int	endcol;			/* ending column number of a record */
  int	i,j, ipac;		/* index */
  FILE  *tf;			/* file pointer */


  /************************
  * Check if a file exist *
  *************************/
  *stat = 0;
  ipac = tbl_finfo[ifile]->file_handle;

  if (ipac <0 || ipac>= maxfiles)
  {
      sprintf(tbl_ival, "NO such file");	/* print error messge */
      *stat = TBL_NOFILE;
      return(tbl_ival);	
  }

  /****************************
  * Initialize Local Pointers *
  *****************************/
  tf 	= tbl[ipac]->tfile;
  if (tf == (FILE *) NULL)
  {
      sprintf(tbl_ival, "File not open");	    /* print error messge */
      *stat = TBL_NOTOPN;
      return(tbl_ival);	
  }
  rec 	= tbl[ipac]->rec;

  /*********************
  * Check Column index *
  **********************/
  if (icol<0 || icol>= tbl[ipac]->ncols)	    /* column out of boundary */
  {
      sprintf(tbl_ival, "NO such column.");	    /* print error messge */
      *stat = TBL_ERROR;
      return(tbl_ival);	
  }

  /**************************
  * Assign beginning column * 
  ***************************/
  if (icol != tbl[ipac]->ncols-1)
      endcol= tbl[ipac]->colinfo[icol]->endcol-1;   /* changed to IPAC convension */
  else
      endcol= tbl[ipac]->colinfo[icol]->endcol;     /* changed to IPAC convension */

  if (icol == 0) 
      bgncol = 0;
  else
      bgncol = tbl[ipac]->colinfo[icol-1]->endcol;  /* changed for IPAC convension */

  /******************
  * Assign tbl_ival * 
  *******************/
  if (fix[ipac] == 'T')
  {
      *stat = fseek(tf, isis_add[ipac]+bgncol, SEEK_SET);
      if (fgets(tbl_ival, endcol-bgncol+2, tf) == NULL) 
      {
	  *stat = TBL_EOF;
	  sprintf(tbl_ival,"TBL_EOF");
	  return(tbl_ival);
      }
  }
  else
  {
      j = 0;				/* initialize j index */
      for (i=bgncol; i<=endcol; ++i)		
      { 
           tbl_ival[j] = rec[i]; 	/* copy the string */
           ++j;
      }
      tbl_ival[j] = '\0';		/* end of string */
  }

  /*******************
  * Get rid of space *
  ********************/
  tbl_ival[strlen(tbl_ival)] = '\0';
  for (j=strlen(tbl_ival)-1; j>=0; --j)
  {
       if (tbl_ival[j] != ' ' ) break;
       tbl_ival[j] = '\0';
  }

  ptr = tbl_ival;
  while (*ptr == ' ') ++ptr;
  strcpy(tbl_ival, ptr);
  ptr = tbl_ival;
  return(ptr);
}

