/****************************************************************************
** Routine Name:                isis_header.c                              **
**                                                                         **
** Description:                 This routine finds number of columns in a  **
**                              record using vertical bar or using the     **
**			 	first line of a record.  It reallocates    **
**				column memory if more is needed.	   **
**                                                                         **
** Input Parameters:            ifile   ____ input file number             **
**                                                                         **
** Output Parameters:           numcol  ____ number of columns             **
**                                                                         **
** Function Return Values:      stat = TBL_OK    ( 0), success		   **
** 			 	stat = TBL_ERROR (-1), error     	   **
** 			 	stat = TBL_NOEME (-2), no memory	   **
** 			 	stat = TBL_EOF   (-4), end of file	   **
** 			 	stat = TBL_RDERR (-5), read error	   **
**                                                                         **
** Routines Be Called:		isis_gets.c				   **
** 				isis_barparse.c				   **
** 				isis_colrallc.c				   **
**                                                                         **
** Revision History:                                                       **
**                                                                         **
**   19??-??    John Good       Initial author                             **
**   1998-03    Angela Zhang    (1) Seperate it from original prototype    **
**   				    tbl_open.c    			   **
**                              (2) Modify the code                        **
**                                                                         **
** Version:                     ISIS0.0                                    **
**                                                                         **
**                                                                         **
** Infrared Science Information System                                     **
** California Institute of Technology                                      **
**                                                                         **
** Copyright (c) 1998, California Institute of Technology.                 **
** U. S. Government Sponsorship under NASA Contract                        **
** NAS7-???? is acknowledged.                                              **
**                                                                         **
*****************************************************************************/


/***********************
* Define Include Files *
************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <tbl.h>


/**************************
* Declear global variables *
***************************/
long int  key_bytes[TBL_MAXFILES];	/* total keyword bytes */
long int  rec_bytes[TBL_MAXFILES];	/* total bytes in a record */
char      fix[TBL_MAXFILES];		/* fixed ISIS_ASCII file */


isis_header(ifile, numcol)
int 	ifile;	 			/* file number */
int 	*numcol;			/* total column number of a line */ 

{
   /**************************
   * Declear local variables *
   ***************************/
   int	i, ch;				/* loop index */
   int	k;				/* string counter */
   int	icol;				/* column index */
   int	irow;				/* row index  */
   int	col_len;			/* column length */
   int	rec_len;		        /* string length */
   int	headlen;		        /* record length */
   int	got_head;		      	/* find/not find (1/0) vertical bar */
   int	ncols;				/* column numbers in a line */
   int	stat;				/* return status */
   char	*rec;				/* pointer to rec */
   FILE *tf;				/* file point */


   /*************************
   * initialize the pointer *
   * and variables          *
   **************************/
   rec 		= tbl[ifile]->rec;
   ncols	= 0;
   k		= 0;
   got_head	= 0;
   stat         = 0;
   tbl[ifile]->keynum = 0;

   /**************
   * get keyword *
   ***************/
   fix[ifile]       = 'F';
   key_bytes[ifile] = 0;
   while (1)
   {
     headlen = isis_gets(ifile);	/* get status or record length */
     if (headlen < 0 && headlen !=TBL_EOF) return(TBL_ERROR);/* return, if any error occurs */
     if (headlen ==TBL_EOF) /* EOF */
     {
       tbl[ifile]->ncols	= 0;
       *numcol 		= 0;
       tbl[ifile]->header	= got_head;
       return(0);
     }
     if (rec[0] =='\\') 
     {
         key_bytes[ifile] = key_bytes[ifile] + headlen + 1;	 
	 stat = isis_keyword(ifile, headlen);
     }
     else
	 break;		    	   	/* exit loop at right place */ 
   }
   rec_bytes[ifile] = headlen + 1;

   /******************
   * Find the Column *
   *******************/
   for (i=0; i<headlen; ++i)
   {
     if (rec[i] == '|' )		/* search for vertical bar */
     {  
	got_head =  1;  		/* find a column */
	break;				/* out of searching loop */
     }
   }

   /*********************************
   * Use a vertical bars to delimit *
   * a column. Reallocate Column    *
   * Memory, If Needed.	         *
   **********************************/
   if (got_head)				/* find a column already */
   {
      for (i=0; i<headlen; ++i)	 		/* search a striing */	
      {
	  if (rec[i] == '\\')  break;
	  if (rec[i] == '\n')  break;		/* end of a line */
          else if (rec[i] == '|')		/* end of a column */ 
	  {
	    tbl[ifile]->colinfo[ncols]->endcol = i;
 	    ++ ncols;

	    if (ncols == tbl[ifile]->maxcols) 	/* enough memory ? */
		stat = isis_colrallc(ifile);   	/* reallocate memory */
	    if (stat < 0) return(stat);

	    k = 0;			      	/* initialize counter */ 
	    if (i == 0) --ncols;      	/* exclude the first ver. bar */
          }
	  else if ( rec[i] != ' ')
	    stat = isis_barparse(ifile, ncols, i, &k); 
      }					      	/* end of for loop */
   }					      	/* end of if (got_head) */


   /**********************
   * Use first data line *
   * to delimit column.  *		
   ***********************/
   else 	 
   {
     for (i=0; i<headlen; ++i)
     {
	  if (rec[i] != ' ')
	  {
	      if (ncols > 0) 
		  tbl[ifile]->colinfo[ncols-1]->endcol = i-1;
              ++ncols;

	      if (ncols == tbl[ifile]->maxcols)  /* enough memory ? */
		  stat=isis_colrallc(ifile,ncols);/* reallocate memory */
              if (stat < 0) return (stat);

   	      while (i<headlen && rec[i] != ' ') ++i;
          }
     }
     tbl[ifile]->colinfo[ncols]->endcol = headlen;
     --ncols;
   }


   /***************
   * Assign width *
   ****************/
   tbl[ifile]->colinfo[0]->width= tbl[ifile]->colinfo[0]->endcol + 1;
   for (i=1; i<ncols; ++i)
   {
        tbl[ifile]->colinfo[i]->width
		  		= tbl[ifile]->colinfo[i]->endcol 
		  		- tbl[ifile]->colinfo[i-1]->endcol;
   }
   
   for (i=0; i<ncols; ++i)
        tbl[ifile]->colinfo[i]->delcol = 0;

   for (i=0; i<ncols; ++i)
   {
	if (tbl[ifile]->colinfo[i]->name[0] == '\0')
	    sprintf(tbl[ifile]->colinfo[i]->name, "field.%d", i);

   }
   tbl[ifile]->ncols	= ncols;
   *numcol 		= ncols;
   tbl[ifile]->header	= got_head;

   return(0);
}
