/**************************************************************************** 
**                                                                         **
** Routine Name:                tbl_colval.c                               **
**                                                                         **
**                                                                         **
** Description:                 This routine returns column value in real  **
**				data type. If scaling factor or offset     **
**				is wrong, error status is returned.        **
**                                                                         **
**                                                                         **
** Input Parameters:            ifile   ____ input file number             **
**                              erow    ____ read row erow                 **
**                              col     ____ return value on column col.   ** 
**                              type    ____ data type of a return value   **
**                              **ptr   ____ vlid  data pointer            **
**                                                                         **
**                                                                         **
** Output Parameters:           None   				 	   **
**                                                                         **
**                                                                         **
** Function Return Values:      stat = TBL_OK    (  0) success             **
**                              stat = TBL_WARNING( 4) can't be scaled     **
**                              stat = TBL_ERROR ( -1) error               **
**                              stat = TBL_NOOPN ( -6) file does not open  **
**                              stat = TBL_NOCOL ( -7) no such column      **
**                              stat = TBL_NOTPY ( -9) no such data type   **
**                              stat = TBL_CNVERR(-10) convertion error    **
**                              stat = TBL_WSCAL (-11) zero scaling factor **
**                                                                         **
**                                                                         **
** Usage:			tbl_colval(int ifile, int erow, int col,   **
**			                   int type, void **ptr)           **
**                                                                         **
**                                                                         **
**                                                                         **
**				ISIS library calls:		   	   **
**				     tbl_type.c				   **
**				     tbl_scale.c			   **
**				     tbl_zero.c				   **
**				     tbl_convert.c			   **
**				     tbl_colvals.c			   **
**				     isis_scaling.c			   **
**                                                                         **
**                                                                         **
** Revision History:                                                       **
**                                                                         **
**   1998-03    Angela Zhang    code first written	                   ** 
**   2007-02    Iffat Khan      check for null before conversion           ** 
**                                                                         **
** Version:                     TBL0.0                                     **
**                                                                         **
**                                                                         **
** Infrared Science Information System                                     **
** California Institute of Technology                                      **
**                                                                         **
** Copyright (c) 1998, California Institute of Technology.                 **
** U. S. Government Sponsorship under NASA Contract                        **
** NAS7-???? is acknowledged.                                              **
**                                                                         **
*****************************************************************************/

/***********************
* define Include Files *
************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <tbl.h>


/**************************
* define global variables *
***************************/
char 	*tbl_ival;		  /* string for ISIS column content */
char    **tform[TBL_MAXFILES];    /* data format */
int     tabletype;                /* FITS_ASCII, FITS_BINARY, or ISIS_ASCII */


tbl_colval(ifile, erow, col, type, ptr)
int	ifile;			  /* file index */
int	erow;			  /* row number */
int	col;			  /* col number */
int	type;			  /* data type  */
void    **ptr;			  /* data pointer */

{
    /*************************
    * define local variables *
    **************************/
    float   		fx;	   /* dummy variable */
    double  		fd;	   /* dummy variable */
    double  		*fdd;	   /* dummy variable */
    double  		sfd;	   /* dummy variable */
    double  		*dptr;	   /* double data pointer */
    double		scale;     /* real scale value */
    double		offset;    /* real offset value */

    int	    		stat;	   /* return status */
    int	    		tfields;   /* number of column */
    int	    		ii, i;	   /* loop index */
    int     		anynull;   /* null value */	
    int     		file_no;   /* file index */	

    short int		fi;	   /* dummy variable */
    long int		fj;	   /* dummy variable */
    long int		ioffset;   /* integer offset */	
    long int		iscale;	   /* integer scale */
    long int		sfj;	   /* dummy variable */
    long int		*iptr;	   /* integer data pointer */
    long long int	lli;	   /* dummy variable */

    unsigned short int	fui;	   /* dummy variable */
    unsigned long int	fuj;	   /* dummy variable */

    char		**cptr;     /* character data pointer */
    char		cscale;     /* dummy variable */
    char		coffset;    /* dummy variable */
    char		**cval;	    /* dummy variable */
    unsigned char	fb;	    /* dummy variable */

    char* tbl_colstr(int, int, int, int*);	


    stat = 0;
    file_no    = tbl_finfo[ifile]->file_handle;
    tabletype  = tbl_finfo[ifile]->table_type;
    *ptr = 0;

    /*******************
    * ISIS ASCII files *
    ********************/
    if (tabletype == ISIS_ASCII)
    {
	/* check for column boundary */
        if (col+1 < 1 || col+1 > tbl[file_no]->ncols)       
		    return (TBL_NOCOL);      

	/* get character string from isis table */
        tbl_ival   = tbl_colstr(ifile, erow, col, &stat);
        if (stat < 0) return (stat);
        if (!strcmp(tbl_ival, tbl[file_no]->colinfo[col]->null))
          return(TBL_OK);

	switch (type)
	{
	    /*-----------------------------*/
	    /* get scaled value for double */
	    /*-----------------------------*/
	    case TBL_DOUBLE:

                 stat = tbl_scale(ifile, col, type, &scale);
		 if (stat < 0) return (stat);

                 stat = tbl_zero (ifile, col, type, &offset);
		 if (stat < 0) return (stat);

 	         stat  = isis_convert(tbl_ival, type, &fd);
		 dptr  = (double *) ptr;
		 if (stat != 0) 
                 {
                     *dptr = 0.0;
		      return (stat);
                 }

		 stat  = isis_scaling(type, &scale, &offset, &fd, &sfd);
		 if (stat < 0) return (stat);

		 *dptr = sfd;
	         return(TBL_OK);

	    /*------------------------------*/
	    /* get scaled value for integer */
	    /*------------------------------*/
	    case TBL_INTEGER:

                 stat = tbl_scale(ifile, col, type, &iscale);
		 if (stat < 0) return (stat);

                 stat = tbl_zero (ifile, col, type, &ioffset);
		 if (stat < 0) return (stat);

 	         stat  = isis_convert(tbl_ival, type, &fj);
		 iptr  = (long int *) ptr;
		 if (stat != 0) 
                 {
                     *iptr = 0;
		      return (stat);
                 }

		 stat  = isis_scaling(type, &iscale, &ioffset, &fj, &sfj);
		 if (stat < 0) return (stat);

		 *iptr = sfj;
	         return(TBL_OK);

	    /*--------------------------------*/
	    /* In fact scaling a character    */
	    /* string doesn't make sense. So  */
	    /* original string is retured.    */
	    /*--------------------------------*/
	    case TBL_CHARACTER:

		 cptr   = (char **) ptr;
		 *cptr  = tbl_ival;
		 return(TBL_OK);
	}
    }
    return(TBL_NOTYP);
}    
