/****************************************************************************
**                                                                         **
** Routine Name:                tbl_open.c                                 **
**                                                                         **
**                                                                         **
** Description:                 This routine opens an input file which     **
**                              can be an ISIS ASCII file or a FITS bianry **
**				and ASCII file. It calls either ISIS       **
**                              library or FITS library depends on 	   **
**				file format.                               **
**                                                                         **
**                                                                         **
** Input Parameters:            *fname   ____ input file name              **
**                                                                         **
**                                                                         **
** Output Parameters:           passinfo ---- file structure which holds   **
**					      information		   **
**                                                                         **
**                                                                         **
** Function Return Values:      istat = file number(>0), suceses           **
**                              istat = TBL_ERROR  (-1), error             **
**                              istat = TBL_NOMEM  (-2), no memory left    **
**                              istat = TBL_RDERR  (-5), read rec. error   **
**                              istat = TBL_NOTOPN (-6), can't open file   **
**                              istat = TBL_TYPERR (-8), not a proper file **
**                                                                         **
**                                                                         **
** Usage:			tbl_open(char* fname, 			   **
**					 struct tbl_filinfo **passinfo)	   **
**									   **
**									   **
** Routines Be Called:          FITS library calls:			   **
**				     ffopen.c                              **
**	 			     ffghpr.c 				   **
**				     ffmahd.c				   **
**				     ffmrhd.c				   **
**				     ffgkyn.c				   **
**				     ffghtb.c				   **
**				     ffghbn.c				   **
**				     					   **
**				ISIS library calls:			   **
**				     isis_open.c			   **
**				     isis_read.c			   **
**									   **
**									   **
** Revision History:                                                       **
**                                                                         **
**   1998-03    Angela Zhang    code first written		           **
**   2004-08    Iffat Khan      modified for appending in input file       **
**   2005-10    Iffat Khan      initialized link_handle                    **
**   2007-02    Iffat Khan      don't initialize 'Fix"                     **
**                                                                         **
**                                                                         **
** Version:                     TBL0.0                                     **
**                                                                         **
**                                                                         **
** Infrared Science Information System                                     **
** California Institute of Technology                                      **
**                                                                         **
** Copyright (c) 1998, California Institute of Technology.                 **
** U. S. Government Sponsorship under NASA Contract                        **
** NAS7-???? is acknowledged.                                              **
**                                                                         **
*****************************************************************************/

/*
##########################
# define including files #
##########################
*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <tbl.h>
#include <sys/stat.h>
#include <sys/types.h>


/*
###########################
# define global variables #
###########################
*/
double      *tscal[TBL_MAXFILES];       /* scale factor */ 
double      *tzero[TBL_MAXFILES];       /* data offset */

long int    key_bytes[TBL_MAXFILES];    /* total keyword bytes */
long int    rec_bytes[TBL_MAXFILES];    /* total bytes in a record */
long int    tbcol[TBL_MAXFILES][200];   /* ASCII table ending col. byte num. */
long int    buffer_size[TBL_MAXFILES];  /* size of a table file */
int         tabletype;                  /* FITS_ASCII, FITS_BINARY, or ISIS_ASCII */
int         keynum[TBL_MAXFILES];       /* key number */
int         tmp_output[TBL_MAXFILES];   /* option to indicate empty output  */

char        fix[TBL_MAXFILES];          /* fixed ISIS_ASCII file */
char        mode;                       /* read/write */
char        **ttype[TBL_MAXFILES];      /* data type */
char        **tform[TBL_MAXFILES];      /* data format */
char        **tunit[TBL_MAXFILES];      /* data unit */
char        **tdisp[TBL_MAXFILES];      /* display format */
char	    *tbl_fval;		        /* string for FITS column content */
char	    *tbl_ival;		        /* string for ISIS column content */


tbl_open(fname, passinfo)
char		    *fname;	/* input file name */
struct tbl_filinfo  **passinfo;	/* file information structure */

{
    /*
    ##########################
    # define local variables #
    ##########################
    */

    int		istat;		/* return status of a function */
    int		ifile;		/* file index */
    int		jfile;		/* file index */
    int		keyindex;	/* keyword index */
    int         simple;		/* simple=1 indicates standard fits file */
    int		tfields;	/* column number  */
    int	        bitpix;		/* bits per pixel */
    int		naxis;		/* number of dimensions in a FITS array */
    int		extend; 	/* indicates FITS extension files */
    int		hdutype;	/* HDU type */
    int		ii,i,lbk;  	/* loop index */	

    long int	rowlen;		/* length of table row in bytes */ 
    long int	nrows;		/* number of rows in a table */
    long int	naxes[3];	/* define size of each dimension */
    long int	pcount;		/* size of binary heap */
    long int	gcount;		/* number of groups in primary array */

    long int    nulval;    /* mumerical value to represent undefined pixel */
    long int    repeat;    /* length of column vector */
    char        typechar;  /* symbolic code of the table column */
	    
    struct stat    buf;		/* buffer structure of the system */

    istat    = 0; 

    /************************
    * allocate memory space *
    * for tbl_finfo         *
    *************************/
    if (tbl_finfo == (struct tbl_filinfo **) NULL)
    {
	tbl_finfo = (struct tbl_filinfo **) 
		 malloc(TBL_MAXFILES * sizeof(struct tbl_filinfo *));
	if (tbl_finfo == (struct tbl_filinfo **)NULL) return(TBL_NOMEM);
    
	for (jfile = 0; jfile < TBL_MAXFILES; ++jfile)
        {
             tbl_finfo[jfile] = (struct tbl_filinfo *) 
				 malloc( sizeof(struct tbl_filinfo));
	     if (tbl_finfo[jfile] == (struct tbl_filinfo *) NULL) 
				  return(TBL_NOMEM);
	     tbl_finfo[jfile]->fileptr = (FILE *) NULL; 
	}
    }

    /*******************
    * find file number *
    ********************/

    for (jfile = 0; jfile < TBL_MAXFILES; ++jfile)
    {
         if (tbl_finfo[jfile]->fileptr == (FILE *) NULL)
	     break;
    }

        mode = 'r';
        ifile	   = isis_open(fname,  &tfields);
        if (ifile < 0) return(TBL_ERROR);
        tmp_output[ifile] = 0;

	/************************
	* save file information *
	*************************/
	strcpy(tbl_finfo[jfile]->file_name, fname);
	tbl_finfo[jfile]->fileptr      = tbl[ifile]->tfile;
	tbl_finfo[jfile]->table_type   = ISIS_ASCII;
	tbl_finfo[jfile]->file_handle  = ifile;
	tbl_finfo[jfile]->ncols        = tfields;
	tbl_finfo[jfile]->nkeywords    = tbl[ifile]->keynum;
	tbl_finfo[jfile]->link_handle      = -1;

	if (fix[ifile] == 'T')
	{
            stat(fname, &buf);
	    buffer_size[ifile]       = buf.st_size;
	    tbl_finfo[jfile]->nrows  = (buffer_size[ifile] - key_bytes[ifile])
				       / rec_bytes[ifile];
        }
        else
        {
	    tbl_finfo[jfile]->nrows    = 0;     
        }
	*passinfo 	               = tbl_finfo[jfile];
	passinfo 	               = &tbl_finfo[jfile];
	return(jfile);
}
