/****************************************************************************
**                                                                         **
** Routine Name:                tbl_zero.c                                 **
**                                                                         **
**                                                                         **
** Description:                 This routine returns offset factor from    **
**			        a table. If offset factor is not defined   **
**			        for a FITS file, double precision 0 is     **
**			        returned, regardless of data type. If a    **
**			        offset factor is not defined for an ISIS   **
**			        table file, value of 0 is return in double **
**			        floating, long integer, and character,     **
**			        according to the data type.                **
** 									   ** 
** 									   ** 
** Input Parameters:            col    ____ column number                  **
**                              ifile  ____ global file number             **
**                              type   ____ converted data type            **
**                                                                         **
**                                                                         **
** Input/Output Parameters:     **ptr  ---- void pointer                   **
**                                                                         **
**                                                                         **
** Function Return Values:      stat = TBL_OK    (  0) ok                  **
** 				stat = TBL_ERROR ( -1) error               **
** 				stat = TBL_NOTOPN( -6) file not open       **
** 				stat = TBL_NOTYP ( -9) no such data type   **
**                              stat = TBL_CNVERR(-10) convertion error    **
**                              stat = TBL_WARNING( 4) warning             **
**                                                                         **
**                                                                         **
** Routines Be Called:          FITS library calls:                        **
**                                   ffgacl.c                              **
**                                   ffgbcl.c                              **
**                                                                         **
**                              ISIS library calls:                        **
**                                   isis_convert.c                        **
**                                                                         **
** Revision History:                                                       **
**                                                                         **
**   1998-03    John Good       original written                           **
**   1998-03    Angela Zhang    integrate FITS and ISIS                    **
**                                                                         **
**                                                                         **
** Version:                     TBL0.0                                     **
**                                                                         **
**                                                                         **
** Infrared Science Information System                                     **
** California Institute of Technology                                      **
**                                                                         **
** Copyright (c) 1998, California Institute of Technology.                 **
** U. S. Government Sponsorship under NASA Contract                        **
** NAS7-???? is acknowledged.                                              **
**                                                                         **
*****************************************************************************/

/*
########################
# define Include Files #
########################
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <tbl.h>

/**************************
* Define Global Variables *
***************************/
double      *tscal[TBL_MAXFILES];       /* scale factor */
double      *tzero[TBL_MAXFILES];       /* data offset */

long int    tbcol[TBL_MAXFILES][200];   /* ASCII table ending col. byte num. */
int         tabletype;          /* FITS_ASCII, FITS_BINARY, or ISIS_ASCII */
int         keynum[TBL_MAXFILES];       /* key number */

char        **ttype[TBL_MAXFILES];      /* data type */
char        **tform[TBL_MAXFILES];      /* data format */
char        **tunit[TBL_MAXFILES];      /* data unit */
char        **tdisp[TBL_MAXFILES];      /* display format */


tbl_zero(int ifile, int icol, int type,  void** ptr)

/*
##########################
# Define Local Variables # 
##########################
*/
{
int		stat;	   /* return status */
int		file_no;   /* file index */
long int	nulval;    /* mumerical value to represent undefined pixel */
long int	repeat;    /* length of column vector */
char		typechar;  /* symbolic code of the table column */
char		nulstr[40];/* char. string to represent ASCII undefined pixel */

long int	*iptr;	   /* integer pointer */
double		*dptr;	   /* double pointer */
char		*cptr;     /* char pointer */

/*
######################
# get scaling factor # 
######################
*/

    file_no   = tbl_finfo[ifile]->file_handle;
    tabletype = tbl_finfo[ifile]->table_type;

    /******************
    * ISIS table file *
    *******************/
    if (tabletype == ISIS_ASCII)
    {
	/*-------------------------*/
	/* offset factor defined   */
	/* by the input table file */
	/*-------------------------*/
	if (tbl[file_no]->colinfo[icol]->zero[0] != '\0')
	{
	    stat=isis_convert(tbl[file_no]->colinfo[icol]->zero,
			      type, ptr);
	    return(stat);
        }
	else
	{
	    /*----------------------*/
	    /* return default value */
	    /*----------------------*/
	    switch(type)
	    {
		case TBL_DOUBLE:
	            dptr = (double *)ptr;
	            *dptr = 0.0;
		    return(TBL_OK);

		case TBL_INTEGER:
	            iptr = (long int *)ptr;
	            *iptr = 0;
		    return(TBL_OK);

		case TBL_CHARACTER:
	            cptr = (char *)ptr;
	            *cptr = '0';
		    return(TBL_OK);
	    } /* end of switch */
        }
    } /* end of ISIS table file */

    return(TBL_NOTYP);
}

