#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "banded.h"

/* when set, forces double precision space to be allocated
   in a separate call from the row pointer space
    (more conservative a design)

   when reset, assume some things about the alignment of
   the computer and allocate both row pointer space and
   double precision space with the same call.
    (somewhat faster design)
*/

#define separate_alloc 1

/* factor A = LU by a band factorization process with no pivoting
      A = square matrix, not necessarily symmetric
      p lower band width
      q upper band width

follows Golub & van Loan: Matrix Computations, Algorithm 4.3.1

returns 0 for OK completion, -1 for divide by eps error

 */
int bandlufact(double **a,int n,int p, int q,float eps) {
  int i,j,k,kpn,kqn;
  double piv;
  for(k=1;k<n;k++) {
    kpn=k+p;
    if(kpn>n) 
      kpn=n;
    piv=a[k-1][k-1];
    if(fabs(piv)<eps)
      return -1;
    for(i=k+1;i<=kpn;i++) {
      a[i-1][k-1]=a[i-1][k-1]/piv;
    }
    kqn=k+q;
    if(kqn>n)
      kqn=n;
    for(j=k+1;j<=kqn;j++) {
      for(i=k+1;i<=kpn;i++) {
	a[i-1][j-1]=a[i-1][j-1]-a[i-1][k-1]*a[k-1][j-1];
      }
    }
  }
  return 0;
}
/* solves factored matrix for forward soln
   Lower triangular matrix L has an implicit 1 for the diagonal entries.
   for loops are 1 origin, subscripts subtract 1 for zero origin
 */
int bandluforsubst(double **L,int n,int p, int q,double *b) {
  int i,j,jpn;
  for(j=1;j<=n;j++) {
    jpn=j+p;
    if(jpn>n)
      jpn=n;
    for(i=j+1;i<=jpn;i++) {
      b[i-1]=b[i-1]-L[i-1][j-1]*b[j-1];
    }
  }
  return 0;
}
/* solves factored matrix for backward soln
   Upper triangular matrix U has arbitrary diagonal entries.
   for loops are 1 origin, subscripts subtract 1 for zero origin
 */
int bandlubaksubst(double **U,int n,int p, int q,double *b) {
  int i,j,jq1;
  for(j=n;j>=1;j--) {
    jq1=j-q;
    if(jq1<1)
      jq1=1;
    b[j-1]=b[j-1]/U[j-1][j-1];
    for(i=jq1;i<j;i++) 
      b[i-1]=b[i-1]-U[i-1][j-1]*b[j-1];
  }
  return 0;
}
/* allocate double_count for a banded array
physical double_count - 
a   - list of pointers for each row
b   - doubles (p+q+1) for each row
to release- do a memory operation free(of result of this function)
*/
double **   bandalloc(int n,int p, int q) {
  double** row_pointers,**row_pointers_end,*double_mass;
  int double_count,i,pointer_count,pointer_sizeof,chars_to_alloc;
  double piv;

  pointer_sizeof=sizeof(double*);
  if(pointer_sizeof<sizeof(double))
    pointer_sizeof=sizeof(double);

  double_count=n*(p+q+1)+1;
  pointer_count=n+1;

 if(separate_alloc) {
  row_pointers=(double**)calloc(1,pointer_sizeof*pointer_count );
  double_mass=(double*)calloc(1,sizeof(double)*double_count);
  row_pointers[n]=double_mass;
  }
  else {
  chars_to_alloc = sizeof(double)*double_count + pointer_sizeof*pointer_count;
  row_pointers=(double**)calloc(1,chars_to_alloc );

  /* ensure that row_pointers end is on a proper alignment for double type */
  row_pointers_end=row_pointers+pointer_count;

  double_mass=(double*)row_pointers_end;

  /* n+1th pointer points to space allocated for both pointers and doubles */
  row_pointers[n]=(double*)row_pointers;
  }

  /* row_pointers points to double_count actually used */
  for(i=0;i<n;i++) {
    row_pointers[i]=double_mass+i*(p+q+1) ;
  }

  /* offset space actually used */
  for(i=0;i<n;i++) {
    row_pointers[i]=row_pointers[i]-(i-p);
  }

  /* touch each row to check pointer arithmetic */
  piv=0;
  for(i=0;i<n;i++) {
    piv+=row_pointers[i][i];
  }
  return row_pointers;
}
/* allocate space for a full array of dimension n x m
         physical space - 
             a   - list of n pointers for each row
             b   - double vector of length m for each row
             
    to release- do a memory operation free(of result of this function)
    */
double **   matalloc(int n,int m) {
  double** row_pointers,**row_pointers_end,*double_mass;
  int double_count,i,pointer_count,pointer_sizeof,chars_to_alloc;
  double piv;

  pointer_sizeof=sizeof(double*);
  if(pointer_sizeof<sizeof(double))
    pointer_sizeof=sizeof(double);

  double_count=n*m;
  pointer_count=n+1;
  if(separate_alloc) {
  row_pointers=(double**)calloc(1,pointer_sizeof*pointer_count );
  double_mass=(double*)calloc(1,sizeof(double)*double_count);
  row_pointers[n]=double_mass;
  }
  else {
  chars_to_alloc = sizeof(double)*double_count + pointer_sizeof*pointer_count;
  row_pointers=(double**)calloc(1,chars_to_alloc );

  /* ensure that row_pointers end is on a proper alignment for double type */
  row_pointers_end=row_pointers+pointer_count;

  double_mass=(double*)row_pointers_end;

  /* n+1th pointer points to space allocated for doubles */
  row_pointers[n]=(double*)row_pointers;
  }

  /* row_pointers points to space actually used */
  for(i=0;i<n;i++) {
    row_pointers[i]=double_mass+i*m ;
  }
  /* touch each row to check pointer arithmetic */
  piv=0;
  for(i=0;i<n;i++) {
    piv+=row_pointers[i][i];
  }

  return row_pointers;
}
int bandshow(FILE *out,double** a,int n,int p, int q,char*msg){
  int i,jlo,jhi,j;
  double z;
  static char fmt[]="%14.3lf,";

  fprintf(out,"banded matrix n=%d,p=%d,q=%d: %s\n",n,p,q,msg);
  z=0;
  for(i=0;i<n;i++) {
    jlo=i-p;
    jhi=i+q;
    if(jlo<0) jlo=0;
    if(jhi>(n-1))jhi=n-1;
    fprintf(out,"%3d,",i);
    for(j=0;j<jlo;j++)
      fprintf(out,fmt,z);
    for(j=jlo;j<=jhi;j++)
      fprintf(out,fmt,a[i][j]);
    for(j=jhi+1;j<n;j++)
      fprintf(out,fmt,z);
    fprintf(out,"\n");
  }
  return 0;
}
int bandunit(FILE *out,int n) {
  int i,j,k,notok;
  double **L,**U,**LU,sum,*x,*b,*y,*yfound,*xfound;
  double LUexpected[3],LUfound[3];


  L=matalloc(n,n);
  U=matalloc(n,n);
  LU=matalloc(n,n);
  b=(double*)malloc(n*sizeof(*b));
  x=(double*)malloc(n*sizeof(*x));
  y=(double*)malloc(n*sizeof(*y));
  xfound=(double*)malloc(n*sizeof(*y));
  yfound=(double*)malloc(n*sizeof(*y));
  notok=0;
  /* create synthetic matrices L and U */
  for(i=0;i<n;i++) {
    for(j=0;j<n;j++) {
      L[i][j]=0;
      U[i][j]=0;
      LU[i][j]=0;
    }
    L[i][i]=1;
    U[i][i]=1.5+i;
    if(i>0) {
      L[i][i-1]=0.125+i;
      U[i-1][i]=-1-i;
    }
    x[i]=i+0.25;
  }
  /* multiply L and U to create LU */
  for(i=0;i<n;i++)
    for(j=0;j<n;j++) {
      sum=0;
      for(k=0;k<n;k++)
        sum=sum+L[i][k]*U[k][j];
      LU[i][j]=sum;
    }
  /* y= U x */
  for(i=0;i<n;i++) {
    sum=0;
    for(k=0;k<n;k++)
      sum=sum+U[i][k]*x[k];
    y[i]=sum;
  }
  /* b = L y = L U x */
  for(i=0;i<n;i++) {
    sum=0;
    for(k=0;k<n;k++)
      sum=sum+L[i][k]*y[k];
    b[i]=sum;
  }
  /* factor LU using bandlufact */
  bandlufact(LU,n,1,1,1E-6);
  /* compare against known inputs */
  for(i=0;i<n;i++) {
    if(i==0) {
      LUexpected[0]=0;
      LUfound[0]=0;
    }
    else {
      LUexpected[0]=L[i][i-1];
      LUfound[0]=LU[i][i-1];
    }
    LUexpected[1]=U[i][i];
    LUfound[1]=LU[i][i];
    if(i==(n-1)) {
      LUexpected[2]=0;
      LUfound[2]=0;
    }
    else {
      LUexpected[2]=U[i][i+1];
      LUfound[2]=LU[i][i+1];
    }
    for(k=0;k<3;k++)
      if(fabs(LUexpected[k]-LUfound[k])>1E-5) {
        notok++;
        if(out!=0)
          fprintf(out,"row %d failed unit test, expected=(%lf,%lf,%lf), found=(%lf,%lf,%lf)\nn=%d",
		  LUexpected[0],LUexpected[1],LUexpected[2],LUfound[0],LUfound[1],LUfound[2],n);
        break;
      }
  }
  /* check out forward and back solution: get x back from b */
  if(notok==0) {
    for(i=0;i<n;i++) yfound[i]=b[i];
    k= bandluforsubst(LU,n,1,1,yfound) ;
    for(i=0;i<n;i++) xfound[i]=yfound[i];    
    k= bandlubaksubst(LU,n,1,1,xfound) ;
    for(i=0;i<n;i++) 
      if(fabs(xfound[i]-x[i])>1E-5) {
	notok++;
	break;
      }
    if(notok)
      if(out!=0) {
	fprintf(out,"failed forward & backward solution test \n y = Ux\n b =Ly =LU x\ni,b[i],x[i],y[i],xfound[i],yfound[i]\n");
	for(k=0;k<n;k++)
	  fprintf(out,"%d,%lf,%lf,%lf,%lf,%lf\n",k,b[k],x[k],y[k],xfound[k],yfound[k]);
        
      }
  }
  if(notok) {
    bandshow(out,L,n,n,n,"L");
    bandshow(out,U,n,n,n,"U");
    bandshow(out,LU,n,n,n,"LU factored");
  }
  if(out!=0)
    if(notok==0)
      fprintf(out,"bandunit passed unit test OK, n=%d\n",n);
  free(L);
  free(U);
  free(LU);
  free(x);
  free(y);
  free(b);
  free(xfound);
  free(yfound);
  return notok;
}
void  bandfree(double ** rowpoint,int n,int p, int q) {
  if(separate_alloc) {
    free(rowpoint[n]);
    free(rowpoint);
  }
  else {
    free(rowpoint);
  }
}
void   matfree(double ** rowpoint,int n,int m) {
  if(separate_alloc) {
    free(rowpoint[n]);
    free(rowpoint);
  }
  else {
    free(rowpoint);
  }
}
