/* Givens and Fast Givens QR transform,
     follows Matrix Computations by Golub & van Loan, 3rd edition
*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "qrgivens.h"
#include "banded.h"

/* given a matrix A of dimension m x n, where m >= n
    calculate a matrices M, D, Q, R, T such that
        D is diagonal, T is  upper  trianguar, M   is orthogonal
        transpose(M) M = D
        Q = M  inverse(sqrt(D))
        transpose(Q)  A  = inverse(sqrt(D)) T = R
        
   In other words, transpose(A) * A = transpose(R) R = transpose(T)   inverse(D) T
     (following paragraph 5.2.5 in Matrix Computations)
    
    D is returned as a vector,
    A is overwritten by T,
    M is not returned
    
    Currently, the D matrix is unity; it is specified in the interface
    to provide for "fast" givens rotations. The D vector will get
    used whn fast Givens rotators are implemented. Scaling
    is required for this. 
*/
void qrg_plain(int m,int n,double **a, double *d){
  int i,j,k;
  double q00,q01,q10,q11, a0, a1;
  double tau,c,s;
  for(i=0;i<m;i++)
    d[i]=1;
  for(j=0;j<n;j++) {
    for(i=m-1;i>=j+1;-1) {
      a0=a[i-1][j];
      a1=a[i][j];
      if(a1==0) {
	c=1;
	s=0;
      }
      else if(fabs(a1)>fabs(a0)) {
        tau=-a0/a1; 
        s=1./sqrt(1+tau*tau);
	c=s*tau;
      }
      else {
	tau=-a1/a0;
	c=1./sqrt(1+tau*tau);
	s=c*tau;
      }
      q00=c;
      q01=s;
      q10=-s;
      q11=c;
      for(k=j;k<n;k++) {
	a0=a[i-1][k];
	a1=a[i][k];
	/* multiply by the tranpose of Q */
	a[i-1][k]= q00 * a0 + q10 *a1;
	a[i][k]    = q01 * a0 + q11* a1;
      }
    }
  }
}
void qrg_fast(int m,	int n,double **a, double *d){
  /* two constants picked out such that:
     ascale  *  inverse(dscale) * ascale = 1
     Chosen to work on base 2 and base 16 floating point
  */
  static double dscale=256;
  static double ascale=16;
  int i,j,k,type;
  double a0, a1;
  double alpha,beta,gamma,tau;
  for(i=0;i<m;i++)
    d[i]=1;
  for(j=0;j<n;j++) {
    for(i=m-1;i>=j+1;-1) {
      a0=a[i-1][j];
      a1=a[i][j];
      if(a1==0) {
	beta=0;
	alpha=0;
	type=2;
      }
      else {
	alpha=-a0/a1;
	beta=-alpha*d[i]/d[i-1];
	gamma =-alpha*beta;
	if(gamma<=1) {
	  tau=d[i-1];
	  d[i-1]=(1+gamma)* d[i];
	  d[i]=(1+gamma)*tau;
	  type=1;
	}   
	else {
	  alpha=1.0/alpha;
	  beta=1.0/beta;
	  gamma=1.0/gamma;
	  d[i-1]*=(1+gamma);
	  d[i]*=(1+gamma);
	  type= 2;
	}
      }
      if(type==1)
	for(k=j;k<n;k++) {
	  a0=a[i-1][k];
	  a1=a[i][k];
	  /* multiply by the tranpose of Q =
	     |    beta       1              |
	     |    1          alpha     |
	  */
	  a[i-1][k]=beta* a0 +  a1;
	  a[i][k]    =a0     +  alpha* a1;
	}
      else
	for(k=j;k<n;k++) {
	  a0=a[i-1][k];
	  a1=a[i][k];
	  /* multiply by the tranpose of Q =
	     |     1             alpha     |
	     |     beta          1         |
	  */
	  a[i-1][k]= a0          +    beta  *a1;
	  a[i][k]    =alpha * a0 +    a1;
	}
    }
    if(d[i-1]>dscale) {
      d[i-1]/=dscale;
      for(k=j;j<n;k++)
	a[i-1][k]*=ascale;
    }
    if(d[i]>dscale) {
      d[i]/=dscale;
      for(k=j;k<n;k++) {
	a[i][k]*=ascale;
      }
    }
  }
}
void qrg_unit(char* msg,int testcase,void*qrg(int m,int n,double **a, double *d)){
  int m,n,i,j,k,pythag_ndx,xsign_ndx,ysign_ndx,swap_ndx;
  double *D,**M,**R,**T,**A,**Q,c,s,m00,m11,m10,m01;
  static npythag=3;
  static double xpythag[3]={3,5,13};
  static double ypythag[3]={4,12,84};
  static double zpythag[3]={5,13,85};

  m=5;
  n=3;
  Q=matalloc(m,m);
  M=matalloc(m,m);
  R=matalloc(m,n);
  T=matalloc(m,n);
  A=matalloc(m,n);
  D=(double*)malloc(m*sizeof(*D));

  /* zero matrices */
  for(i=0;i<m;i++) {
    D[i]=0;
    for(j=0;j<n;j++) {
      T[i][j]=0;
      R[i][j]=0;
      A[i][j]=0;
    }
    for(j=0;j<m;j++) {
      M[i][j]=0;
      Q[i][j]=0;
    }
    /* synthetic data */
    M[i][i]=1;
    for(j=i;j<n;j++)
      R[i][j]=1+0.125*i+0.5*(j-i);
  }
  /* add some rotations to M */
  for(i=0;i<n;i++) {
    for(j=i;j<n;j++) {
      pythag_ndx=i*n+j;
      swap_ndx=1&pythag_ndx;
      pythag_ndx=pythag_ndx>>1;
      xsign_ndx=1&pythag_ndx;
      pythag_ndx=pythag_ndx>>1;
      ysign_ndx=1&pythag_ndx;
      pythag_ndx=pythag_ndx>>1;
      pythag_ndx=pythag_ndx%npythag;
      c=xpythag[pythag_ndx];
      s=ypythag[pythag_ndx];
      if(swap_ndx) {
	s=xpythag[pythag_ndx];
	c=ypythag[pythag_ndx];
      }
      c*=1-2*xsign_ndx;
      s*=1-2*ysign_ndx;
      m00=M[i][i];
      m01=M[i][j];
      m10=M[j][i];
      m11=M[j][j];
      /* multiply by rotation */
      M[i][i]=m00*c+m10*s;
      M[i][j]=m01*c+m11*s;
      M[j][i]=m00*(-s)+m10*c;
      M[j][j]=m01*(-s)+m11*c;
    }
  }
}
   

