/* to make this unit test:
cc -fast -xarch=v8plusa -fsimple=0 -fns=no -o splitest banded.c splinefunc.c ugran.c splitest.c qrgivens.c -lm
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "banded.h"
#include "spline.h"
#include "ugran.h"

/* spline interpolation 
	1. Calculate tridiagonal R and S matrices
           These matrices are stored in banded form.
           Equation 1:  R y" = S y
	2. Solve for coefs of ypp_j and ypp_jp1, y" = R^-1 S y
	3. Calculate A,B,C,D
	4. Calculate interpolated yhat
	5. Calculate uncertainty on yhat using coefs
	      rms(ytab[i])    given as input
              rms(y"[i])      calculated as an order n^3 process

   General usage pattern:

      A. Generate tables by invoking splinetab
        (returns -1 if error computing tables)
      B. Call splineinterp multiple times to perform 
        interplocation
      C. Free up memory by calling splinefree
     */

int splinetab(int n,double xtab[],double ytab[],double ytab_sigma[],splinetab_type **table_p){
  FILE *out;
  int i,k,ifwd,ibak;
  splinetab_type *table;
  double xdif_ip1,xdif_i;
  double *scratch,term,sumsq,**R,**S;

  /* no debug output */
  out=0;

  /* null result pointer to table just in case malloc fails */
  *table_p=0;
  /* allocate scratch storage for a scratch matrix */
  scratch=0;
  R=0;
  S=0;
  scratch=(double*)malloc(sizeof(*scratch)*n);
  R=matalloc(n,n);
  S=matalloc(n,n);
  
  if(scratch==0|| R==0 || S==0){
    free(scratch);
    free(R);
    free(S);
    return -1;
  }
    
    
  table=(splinetab_type *)malloc(sizeof(*table));
  if(table==NULL)
    return -1;
  *table_p=table;
  table->n=n;
  table->xtab=(double*)malloc(sizeof(*table->xtab)*n);
  table->ytab=(double*)malloc(sizeof(*table->xtab)*n);
  table->ypp=(double*)malloc(sizeof(*table->xtab)*n);
  table->ytab_sigma=(double*)malloc(sizeof(*table->xtab)*n);
  table->ypp_sigma=(double*)malloc(sizeof(*table->xtab)*n);
  if(table->xtab==0 || table->ytab==0 || table->ypp==0 || table-> ypp_sigma==0 )
    return -1;

  /* move tables to dedicated data structure */
  for(i=0;i<n;i++) {
    table->xtab[i]=xtab[i];
    table->ytab[i]=ytab[i];
  }
  if(ytab_sigma!=0)
  for(i=0;i<n;i++) 
    table->ytab_sigma[i]=ytab_sigma[i];
else
  for(i=0;i<n;i++) 
    table->ytab_sigma[i]=0;

  /* Calculate R and S matrices */
  R[0][0]=1;
  R[0][1]=0;
  S[0][0]=0;
  S[0][1]=0;
  for( i=1;i<(n-1);i++)
    { 
      xdif_i=table->xtab[i]-table->xtab[i-1];
      xdif_ip1=table->xtab[i+1]-table->xtab[i];

      R[i][i-1]=xdif_i/6.0;
      R[i][i+1]=xdif_ip1/6.0;
      R[i][i]  = (table->xtab[i+1]-table->xtab[i-1])/3.0;
	
      S[i][i-1] = 1.0/xdif_i;
      S[i][i]= - 1.0/xdif_ip1 -1.0/xdif_i;
      S[i][i+1]=  1.0/xdif_ip1;
    }
  R[n-1][n-2]=0;
  R[n-1][n-1]=1;
  S[n-1][n-2]=0;
  S[n-1][n-1]=0;

  if(out!=0)
    bandshow(out,R, n,1,1,"R matrix");
  if(out!=0)
    bandshow(out,S, n,1,1,"S matrix");
  /*Factor the R matrix using banded LU factorization */
  k= bandlufact(R,table-> n,1, 1,1E-6);
  if(out!=0)
    bandshow(out,R, n,1,1,"R factored matrix");
  if(k<0)
    return -1;
  /* solve for ypp vector: ypp = inv(R)*S*y */
  i=0;
  table->ypp[i]=  +S[i][i  ]*table->ytab[i  ]
    +S[i][i+1]*table->ytab[i+1];
  for(i=1;i<(n-1);i++)
    table->ypp[i]= S[i][i-1]*table->ytab[i-1]
      +S[i][i  ]*table->ytab[i  ]
      +S[i][i+1]*table->ytab[i+1];
  i=n-1;
  table->ypp[i]= S[i][i-1]*table->ytab[i-1]
    +S[i][i  ]*table->ytab[i  ];
			  

  ifwd= bandluforsubst(R, n,1,1,table->ypp) ;
  ibak= bandlubaksubst(R, n,1,1,table->ypp) ;

 
  /* now calculate each row of inv(R)*S and use it quadruature
     all the ytab_sigma entryies to calculate the array called
     ypp_sigma*/
  for(k=0;k<n;k++) {
    sumsq=0;
    if(ytab_sigma!=0) {
    for(i=0;i<n;i++)
      scratch[i]=0;
    scratch[k]=1;
    ifwd= bandluforsubst(R, n,1,1,scratch) ;
    ibak= bandlubaksubst(R, n,1,1,scratch) ;
    sumsq=0;
    for(i=0;i<n;i++) {
      term=scratch[i]*table->ytab_sigma[i];
      sumsq+=term*term;
    }   
    }
    table->ypp_sigma[k]=sqrt(sumsq);
  }
  free(scratch);   
  free(R);
  free(S);
  return 0;
}
int splinetab_float(int n,float xtab[],float ytab[],float ytab_sigma[],splinetab_type **table_p){
  FILE *out;
  int i,k,ifwd,ibak;
  splinetab_type *table;
  double xdif_ip1,xdif_i;
  double *scratch,term,sumsq,**R,**S;

  /* no debug output */
  out=0;

  /* null result pointer to table just in case malloc fails */
  *table_p=0;
  /* allocate scratch storage for a scratch matrix */
  scratch=0;
  R=0;
  S=0;
  scratch=(double*)malloc(sizeof(*scratch)*n);
  R=matalloc(n,n);
  S=matalloc(n,n);
  
  if(scratch==0|| R==0 || S==0){
    free(scratch);
    free(R);
    free(S);
    return -1;
  }
    
    
  table=(splinetab_type *)malloc(sizeof(*table));
  if(table==NULL)
    return -1;
  *table_p=table;
  table->n=n;
  table->xtab=(double*)malloc(sizeof(*table->xtab)*n);
  table->ytab=(double*)malloc(sizeof(*table->xtab)*n);
  table->ypp=(double*)malloc(sizeof(*table->xtab)*n);
  table->ytab_sigma=(double*)malloc(sizeof(*table->xtab)*n);
  table->ypp_sigma=(double*)malloc(sizeof(*table->xtab)*n);
  if(table->xtab==0 || table->ytab==0 || table->ypp==0 || table-> ypp_sigma==0 )
    return -1;

  /* move tables to dedicated data structure */
  for(i=0;i<n;i++) {
    table->xtab[i]=xtab[i];
    table->ytab[i]=ytab[i];
  }
  if(ytab_sigma!=0)
  for(i=0;i<n;i++) 
    table->ytab_sigma[i]=ytab_sigma[i];
else
  for(i=0;i<n;i++) 
    table->ytab_sigma[i]=0;

  /* Calculate R and S matrices */
  R[0][0]=1;
  R[0][1]=0;
  S[0][0]=0;
  S[0][1]=0;
  for( i=1;i<(n-1);i++)
    { 
      xdif_i=table->xtab[i]-table->xtab[i-1];
      xdif_ip1=table->xtab[i+1]-table->xtab[i];

      R[i][i-1]=xdif_i/6.0;
      R[i][i+1]=xdif_ip1/6.0;
      R[i][i]  = (table->xtab[i+1]-table->xtab[i-1])/3.0;
	
      S[i][i-1] = 1.0/xdif_i;
      S[i][i]= - 1.0/xdif_ip1 -1.0/xdif_i;
      S[i][i+1]=  1.0/xdif_ip1;
    }
  R[n-1][n-2]=0;
  R[n-1][n-1]=1;
  S[n-1][n-2]=0;
  S[n-1][n-1]=0;

  if(out!=0)
    bandshow(out,R, n,1,1,"R matrix");
  if(out!=0)
    bandshow(out,S, n,1,1,"S matrix");
  /*Factor the R matrix using banded LU factorization */
  k= bandlufact(R,table-> n,1, 1,1E-6);
  if(out!=0)
    bandshow(out,R, n,1,1,"R factored matrix");
  if(k<0)
    return -1;
  /* solve for ypp vector: ypp = inv(R)*S*y */
  i=0;
  table->ypp[i]=  +S[i][i  ]*table->ytab[i  ]
    +S[i][i+1]*table->ytab[i+1];
  for(i=1;i<(n-1);i++)
    table->ypp[i]= S[i][i-1]*table->ytab[i-1]
      +S[i][i  ]*table->ytab[i  ]
      +S[i][i+1]*table->ytab[i+1];
  i=n-1;
  table->ypp[i]= S[i][i-1]*table->ytab[i-1]
    +S[i][i  ]*table->ytab[i  ];
			  

  ifwd= bandluforsubst(R, n,1,1,table->ypp) ;
  ibak= bandlubaksubst(R, n,1,1,table->ypp) ;

 
  /* now calculate each row of inv(R)*S and use it quadruature
     all the ytab_sigma entryies to calculate the array called
     ypp_sigma*/
  for(k=0;k<n;k++) {
    sumsq=0;
    if(ytab_sigma!=0) {
    for(i=0;i<n;i++)
      scratch[i]=0;
    scratch[k]=1;
    ifwd= bandluforsubst(R, n,1,1,scratch) ;
    ibak= bandlubaksubst(R, n,1,1,scratch) ;
    sumsq=0;
    for(i=0;i<n;i++) {
      term=scratch[i]*table->ytab_sigma[i];
      sumsq+=term*term;
    }   
    }
    table->ypp_sigma[k]=sqrt(sumsq);
  }
  free(scratch);   
  free(R);
  free(S);
  return 0;
}
int splineinterp(double x,double *y, double *ysigma,splinetab_type *table){
  double A,B,C,D,xdif_pos;
  double AA,BB,CC,DD,u,usq;
  int pos,hi,lo;

  /* binary search for table location  */
  lo=-1;
  hi=table->n;
  while ((hi-lo)>1)
    {pos=(lo+hi)>>1;
    if(table->xtab[pos]<=x)
      lo=pos;
    else
      hi=pos;
    }
  pos=lo;
  if(pos==(table->n-1))
    pos=table->n-2;

  /* calculate A,B,C,D */
  xdif_pos=table->xtab[pos+1]-table->xtab[pos];
  A=(table->xtab[pos+1]-x)/xdif_pos;
  B=1.0-A;
  C=(A*A*A-A)*xdif_pos*xdif_pos/6.0;
  D=(B*B*B-B)*xdif_pos*xdif_pos/6.0;

  /* calculate function value */
  AA= A*table->ytab[pos];
  BB= B*table->ytab[pos+1] ;
  CC= C*table->ypp[pos] ;
  DD= D*table->ypp[pos+1];
  u=AA+BB+CC+DD;
  AA= A*table->ytab_sigma[pos];
  BB= B*table->ytab_sigma[pos+1] ;
  CC= C*table->ypp_sigma[pos] ;
  DD= D*table->ypp_sigma[pos+1];
  usq=AA*AA+BB*BB+CC*CC+DD*DD;
  *y=u;
  *ysigma=sqrt(usq);
  return 0;
}
void splinefree(splinetab_type *table){
  if(table==0)
    return;
  free(table->xtab);
  free(table->ytab);
  free(table->ytab_sigma);
  free(table->ypp);
  free(table->ypp_sigma);
  free(table);
}

static double roots[5]={0,1,2,0,2};

void splitest_showit(FILE* outfile,int n, splinetab_type *sptab,char* lbl, double sigmay,double* ypp_calc) {
  int i,j,k,m;
  double x,y,yinterp,yinterp_sigma,xmin,xmax,dx;
  fprintf(outfile,"%s %lf\n",lbl,sigmay);
  fprintf(outfile,"sptab, n=%d\ni,x[i],y[i],y_sigma[i],ypp[i],ypp_sigma[i],ypp_calc[i]\n",sptab->n);
  for(i=0;i<sptab->n;i++)
    fprintf(outfile,"%d,%lf,%lf,%lf,%lf,%lf,%lf\n",i,sptab->xtab[i],sptab->ytab[i],sptab->ytab_sigma[i],sptab->ypp[i],sptab->ypp_sigma[i],ypp_calc[i]);
  xmax=sptab->xtab[n-1];
  xmin=sptab->xtab[0];
  dx=(xmax-xmin)/(2*n);
  fprintf(outfile,"i,x,y,yinterp,yinterp_sigma\n");
  for(i=0;i<2*n-1;i++) {
    k=i>>1;
    x=xmin+dx*(i);
    if(i&1) {
      y=(sptab->ytab[k]+sptab->ytab[k+1])*0.5;
      x=(sptab->xtab[k]+sptab->xtab[k+1])*0.5;
    }
    else {
      y=sptab->ytab[k];
      x=sptab->xtab[k];
    }
    j=splineinterp(x,&yinterp, &yinterp_sigma,sptab);
    fprintf(outfile,"%d,%lf,%lf,%lf,%lf\n",i,x,y,yinterp,yinterp_sigma);
  }
}
void spline_test(FILE* outfile,int order,int n, double sigmay, double sigmax) {
  double xtab[100],ytab[100],yhat[100],ysigma[100],zero[100],xoffset,dy[100],d2y[100];
  double dx,x,y,xexact,yexact,rootmin,rootmax,xpwr,ysum,dysum,d2ysum;
  int i,j,k,m;
  splinetab_type *st_exact,*st_noisy;
  double polycoef[5],dpolycoef[5],d2polycoef[5],timesx[15];
  static int jamit=0;


  rootmin=rootmax=roots[0];
  for(i=0;i<order;i++){
    if(rootmin>roots[i])
      rootmin=roots[i];
    if(rootmax<roots[i])
      rootmax=roots[i];
  }
  dx=(rootmax-rootmin)/(n-1);

  /* calculate polynomial coefs */
  polycoef[0]=1;
  for(k=0;k<order;k++) {
    for(j=0;j<=k;j++)
      timesx[j+1]=polycoef[j];
    timesx[0]=0;
    polycoef[k+1]=0;
    for(j=0;j<=(k+1);j++)
      polycoef[j]=-roots[k]*polycoef[j]+timesx[j];
  }
  /* calculate first derivative */
  for(k=1;k<=order;k++)
    dpolycoef[k-1]=k*polycoef[k];
  dpolycoef[order]=0;
  /* calculate second derivative */
  for(k=1;k<=order;k++)
    d2polycoef[k-1]=k*dpolycoef[k];
  dpolycoef[order]=0;

  for(i=0;i<n;i++) {
    zero[i]=0;
    xexact=rootmin+dx*i;
    xoffset=2*(uniform()-1)*sigmax;
    if(i==0)
      xoffset=-fabs(xoffset);
    else if(i==(n-1))
      xoffset=fabs(xoffset);
    x=xexact+xoffset;
    y=1;
    xpwr=1;
    ysum=0;
    dysum=0;
    d2ysum=0;
    for(j=0;j<=order;j++) {
      ysum+=xpwr*polycoef[j];
      dysum+=xpwr*dpolycoef[j];
      d2ysum+=xpwr*d2polycoef[j];
      xpwr*=x;
    }
    ytab[i]=ysum;
    yhat[i]=ytab[i]+gasdev()*sigmay;
    xtab[i]=x;
    dy[i]=dysum;
    d2y[i]=d2ysum;
    ysigma[i]=sigmay;
  }
  i= splinetab(n,xtab,ytab,zero,  &st_exact);
  i= splinetab(n,xtab,yhat,ysigma,&st_noisy);
  if(jamit) {
    for(i=0;i<n;i++)
      st_exact->ypp[i]=d2y[i];
  }
  fprintf(outfile,"n,order,sigmay,sigmax\n");
  fprintf(outfile,"%d,%d,%lf,%lf\n",n,order,sigmay,sigmax);
  splitest_showit(outfile,n,st_exact,"exact y values, sigma=",0,d2y);
  splitest_showit(outfile,n,st_noisy,"noisy y values, sigma=",sigmay,d2y);

}
