/* $Id: tridiagonal.c,v 1.6 2001/11/16 22:25:25 davidh Exp $ */

/* tridiagonal matrices 
     stored as 3 bands: left of diagonal, diagonal, right of diagonal

     for a mnemonic of usage:
       the three used elements of row i are leftdiag[i], diagonal[i], rightdiag[i]

*/
#include <math.h>
#include "tridiagonal.h"
/* solves a triadiagonal matrix equation 

	A r = u
	
   where:        (following Numberical Recipes section 2.4)
        A =tridiagonal matrix
        r =vector of unknowns
        u =right hand side
        gam= scratch array


   Returns:
     -1 if diagonal[0] is zero
     -2 if another singularity detected
     0 for solution completed, u has solution
     
   Notes:
   	All arrays are zero origin, gam is a scratch array
   	0th element of the c array is not used
   	scratch array is passed in, is of length n
   	singularity criteria is passed in
*/
int tridiagsolve(float leftdiag[], float diagonal[], float rightdiag[], float r[], float u[], float gam[],
	float eps, unsigned long n)
{
	long j;
	float bet;

	if (fabs(diagonal[0]) <= eps) 
	     return -1;
	u[0]=r[0]/(bet=diagonal[0]);
	/* decomposition and forward substitution */
	for (j=1;j<n;j++) {
		gam[j]=rightdiag[j-1]/bet;
		bet=diagonal[j]-leftdiag[j]*gam[j];
		if (fabs(bet) <=eps)	
		     return -2;
		u[j]=(r[j]-leftdiag[j]*u[j-1])/bet;
	}
	for (j=(n-2);j>=0;j--) {
		u[j] -= gam[j+1]*u[j+1];
	}
	return 0;
}

/* Given X and Y tridiagonal matrices,
   Exit: Y=inverse(X)*Y without any interchanges, X destroyed
   
   Warning:
     This routine can fail if interchanges are required.
     
   To actually compute the inverse of X, set Y to be an identity matrix
*/
int triadiaginvert(float X_leftdiag[],float X_diagonal[],float X_rightdiag[],float Y_leftdiag[],float Y_diagonal[],float Y_rightdiag[],float eps, int n)
{int k;
 double piv,fac;
 
/* forward sweep, setting sub-diagonal elements of X to zero */
 for(k=0;k<n;k++)
  {if(fabs(X_diagonal[k])<eps)
     return -1;
   piv=1.0/X_diagonal[k];
   /* multiply k'th row of X and Y by the pivot */
   Y_diagonal[k]*=piv;
   Y_diagonal[k]*=piv;
   if(k>0)
     {X_leftdiag[k]*=piv;
      Y_leftdiag[k]*=piv;
      }
   if(k<(n-1))
     {X_rightdiag[k]*=piv;
      Y_rightdiag[k+1]*=piv;
      }
   /* wipe out elements of X below X[k][k] */
   if(k<(n-1))
    {fac=X_leftdiag[k+1];
     /* multiply row k by fac, subtract it from row k+1 */
     X_leftdiag[k+1]=X_leftdiag[k+1]-fac*X_diagonal[k];
     X_diagonal[k+1]=X_diagonal[k+1]-fac*X_rightdiag[k];
     Y_leftdiag[k+1]=Y_leftdiag[k+1]-fac*Y_diagonal[k];
     Y_diagonal[k+1]=Y_diagonal[k+1]-fac*Y_rightdiag[k];
     }
   }
 /* backward sweep, setting super-diagonal elements of X to zero */
 for(k=n-1;k>=0;k--)
  {if(fabs(X_diagonal[k])<eps)
     return -1;
   piv=1.0/X_diagonal[k];
   /* multiply k'th row of X and Y by the pivot */
   Y_diagonal[k]*=piv;
   Y_diagonal[k]*=piv;
   if(k>0)
     {X_leftdiag[k]*=piv;
      Y_leftdiag[k]*=piv;
      }
   if(k<(n-1))
     {X_rightdiag[k]*=piv;
      Y_rightdiag[k]*=piv;
      }
   /* wipe out elements of X above X[k][k] */
   if(k>0)
     {fac=X_rightdiag[k-1];
      /* multiply row k by fac, subtract it from row k-1 */
      X_rightdiag[k-1]=X_rightdiag[k-1]-fac*X_diagonal[k];
      X_diagonal[k-1]=X_diagonal[k-1]-fac*X_leftdiag[k];
      Y_rightdiag[k-1]=Y_rightdiag[k-1]-fac*Y_diagonal[k];
      Y_diagonal[k-1]=Y_rightdiag[k-1]-fac*Y_diagonal[k];
      }
      
   }
  return 0;
}


/* extract the ith row from a tridiagonal matrix representation */
void trigetrow(float leftdiag[], float diagonal[], float rightdiag[], float r[],int i, unsigned long n)
{   int j;

    for(j=0;j<n;j++) r[j]=0;
    
    if(i>0)
       r[i-1]=leftdiag[i];
       
    r[i]=diagonal[i];
    
    if(i<(n-1))
       r[i+1]=rightdiag[i];
       
    return;
    
    
}
/* insert the ith row into a tridiagonal matrix representation */
void triputrow(float leftdiag[], float diagonal[], float rightdiag[], float r[],int i, unsigned long n)
{   
    
    if(i>0)
       leftdiag[i]=r[i-1];
       
    diagonal[i]=r[i];
    
    if(i<(n-1))
       rightdiag[i]=r[i+1];
       
    return;        
}
/* multiply the triangular matrix A times vector x, creating vector rslt */
void trimulvec(float leftdiag[], float diagonal[], float rightdiag[], float x[], float rslt[], unsigned long n)
{ int j;
  if(n>1)
   {/* zeroth element */
    rslt[0]= diagonal[0]*x[0] + rightdiag[0]*rslt[1];
    /* middle elements */
    for( j=1;j<n-1;j++)
      rslt[j]= leftdiag[j]*x[j-1] + diagonal[j]*x[j] +rightdiag[j]*x[j+1];
    /* final element */
    rslt[n-1]=leftdiag[n-1]*x[n-2]+diagonal[n-1]*x[n-1];
    }
  else if(n>0)
    { 
      rslt[0]=diagonal[0]*x[0];
    }
}

/* double precision version follows */
int dtridiagsolve(double leftdiag[], double diagonal[], double rightdiag[], double r[], double u[], double gam[],
	double eps, unsigned long n)
{
	long j;
	double bet;

	if (fabs(diagonal[0]) <= eps) 
	     return -1;
	u[0]=r[0]/(bet=diagonal[0]);
	/* decomposition and forward substitution */
	for (j=1;j<n;j++) {
		gam[j]=rightdiag[j-1]/bet;
		bet=diagonal[j]-leftdiag[j]*gam[j];
		if (fabs(bet) <=eps)	
		     return -2;
		u[j]=(r[j]-leftdiag[j]*u[j-1])/bet;
	}
	for (j=(n-2);j>=0;j--) {
		u[j] -= gam[j+1]*u[j+1];
	}
	return 0;
}

/* extract the ith row from a tridiagonal matrix representation */
void dtrigetrow(double leftdiag[], double diagonal[], double rightdiag[], double r[],int i, unsigned long n)
{   int j;

    for(j=0;j<n;j++) r[j]=0;
    
    if(i>0)
       r[i-1]=leftdiag[i];
       
    r[i]=diagonal[i];
    
    if(i<(n-1))
       r[i+1]=rightdiag[i];
       
    return;
    
    
}
/* insert the ith row into a tridiagonal matrix representation */
void dtriputrow(double leftdiag[], double diagonal[], double rightdiag[], double r[],int i, unsigned long n)
{   
    
    if(i>0)
       leftdiag[i]=r[i-1];
       
    diagonal[i]=r[i];
    
    if(i<(n-1))
       rightdiag[i]=r[i+1];
       
    return;        
}
/* multiply the triangular matrix A times vector x, creating vector rslt */
void dtrimulvec(double leftdiag[], double diagonal[], double rightdiag[], double x[], double rslt[], unsigned long n)
{ int j;
  if(n>1)
   {/* zeroth element */
    rslt[0]= diagonal[0]*x[0] + rightdiag[0]*rslt[1];
    /* middle elements */
    for( j=1;j<n-1;j++)
      rslt[j]= leftdiag[j]*x[j-1] + diagonal[j]*x[j] +rightdiag[j]*x[j+1];
    /* final element */
    rslt[n-1]=leftdiag[n-1]*x[n-2]+diagonal[n-1]*x[n-1];
    }
  else if(n>0)
    { 
      rslt[0]=diagonal[0]*x[0];
    }
}
